Package: rxref
Title: Tidy Utilities for RxNorm and NDC Resolution
Version: 0.4.0
Authors@R: 
	c(person(
	given = "Steven", 
	family = "Smith", 
	email = "ssmith@cop.ufl.edu", 
	role = c("aut", "cre")))
Description: Provides a tidy, vectorized interface to the 'RxNorm' / 'RxNav'
    API for resolving drug names, RxCUIs, National Drug Codes (NDCs), and 
    related drug concept metadata. The package supports workflows for mapping 
    between drug names, RxCUIs, NDCs, ingredients, products, drug classes, and 
    related concepts using data from the National Library of Medicine's 'RxNav' 
    services <https://lhncbc.nlm.nih.gov/RxNav/APIs/> and 'RxNorm'
    <https://www.nlm.nih.gov/research/umls/rxnorm/>.
License: MIT + file LICENSE
URL: https://github.com/ssmithm/rxref,
        https://www.stevenmsmith.org/rxref/
BugReports: https://github.com/ssmithm/rxref/issues
Depends: R (>= 4.1)
Imports: cachem, cli, digest, dplyr, httr2 (>= 1.0.0), jsonlite,
        lifecycle, memoise, purrr, rlang, stringr, tibble, tidyr, vctrs
Suggests: curl, knitr, rmarkdown, testthat (>= 3.0.0), withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
RoxygenNote: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-19 05:18:07 UTC; ssmithm
Author: Steven Smith [aut, cre]
Maintainer: Steven Smith <ssmith@cop.ufl.edu>
Repository: CRAN
Date/Publication: 2026-05-27 08:50:02 UTC
Built: R 4.5.2; ; 2026-05-27 11:37:34 UTC; unix

Package: rxref
Title: Tidy Utilities for RxNorm and NDC Resolution
Version: 0.4.0
Authors@R: 
	c(person(
	given = "Steven", 
	family = "Smith", 
	email = "ssmith@cop.ufl.edu", 
	role = c("aut", "cre")))
Description: Provides a tidy, vectorized interface to the 'RxNorm' / 'RxNav'
    API for resolving drug names, RxCUIs, National Drug Codes (NDCs), and 
    related drug concept metadata. The package supports workflows for mapping 
    between drug names, RxCUIs, NDCs, ingredients, products, drug classes, and 
    related concepts using data from the National Library of Medicine's 'RxNav' 
    services <https://lhncbc.nlm.nih.gov/RxNav/APIs/> and 'RxNorm'
    <https://www.nlm.nih.gov/research/umls/rxnorm/>.
License: MIT + file LICENSE
URL: https://github.com/ssmithm/rxref,
        https://www.stevenmsmith.org/rxref/
BugReports: https://github.com/ssmithm/rxref/issues
Depends: R (>= 4.1)
Imports: cachem, cli, digest, dplyr, httr2 (>= 1.0.0), jsonlite,
        lifecycle, memoise, purrr, rlang, stringr, tibble, tidyr, vctrs
Suggests: curl, knitr, rmarkdown, testthat (>= 3.0.0), withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
RoxygenNote: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-19 05:18:07 UTC; ssmithm
Author: Steven Smith [aut, cre]
Maintainer: Steven Smith <ssmith@cop.ufl.edu>
Repository: CRAN
Date/Publication: 2026-05-27 08:50:02 UTC
Built: R 4.5.2; ; 2026-05-27 11:21:23 UTC; unix

# Generated by roxygen2: do not edit by hand

export(default_product_ttys)
export(exclude_products_with_ingredients)
export(extended_product_ttys)
export(extended_ttys)
export(filter_products_by_route)
export(find_classes)
export(find_ingredients)
export(get_atc)
export(get_atcprod)
export(get_chemical_structure)
export(get_class_members)
export(get_classes)
export(get_clinical_attributes)
export(get_contraindications)
export(get_drug_classes)
export(get_epc)
export(get_may_treat)
export(get_medrt)
export(get_medrt_moa)
export(get_medrt_pe)
export(get_properties)
export(get_va)
export(ingredients_for_rxcui)
export(map_ndc_to_rxcui)
export(map_rxcui_to_ndc)
export(map_to)
export(ndc_status)
export(product_ttys)
export(products_for_ingredients)
export(resolve)
export(rxclass_rela_sources)
export(rxclass_relas)
export(rxclass_relationships)
export(rxref_conf)
export(search_drug)
export(tty_catalogue)
importFrom(digest,digest)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,relocate)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(jsonlite,fromJSON)
importFrom(memoise,memoise)
importFrom(purrr,compact)
importFrom(purrr,imap)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,map_chr)
importFrom(purrr,map_int)
importFrom(purrr,map_lgl)
importFrom(purrr,pluck)
importFrom(purrr,possibly)
importFrom(purrr,quietly)
importFrom(rlang,.data)
importFrom(rlang,.env)
importFrom(stringr,str_detect)
importFrom(stringr,str_remove_all)
importFrom(stringr,str_replace_all)
importFrom(stringr,str_squish)
importFrom(stringr,str_trim)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,replace_na)
importFrom(tidyr,unnest)
importFrom(tidyr,unnest_longer)
importFrom(tidyr,unnest_wider)
importFrom(vctrs,vec_c)
importFrom(vctrs,vec_cast)
importFrom(vctrs,vec_rbind)
importFrom(vctrs,vec_recycle)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# rxref

<!-- badges: start -->

[![R-CMD-check](https://github.com/ssmithm/rxref/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ssmithm/rxref/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

**rxref** is an R package for working with medication identifiers from
the National Library of Medicine’s RxNorm and RxClass APIs. It provides
tidy, vectorized helpers for resolving medication names to RxCUIs,
expanding ingredients to product concepts, mapping between RxCUIs and
NDCs, retrieving clinical product attributes, and querying selected drug
class relationships.

The package is designed for pharmacoepidemiologic and health services
research workflows where medication exposure definitions often need to
move between ingredient names, RxNorm concepts, National Drug Codes
(NDCs), routes, dose forms, and therapeutic classes.

**rxref** was originally built to support pharmacoepidemiologic research
in the [UF CVmedLab](https://cvmedlab.org/), where we often need
reproducible, transparent medication definitions for prescriptions,
pharmacy dispensings, and claims-based drug exposure measures.

## What rxref can do

Common workflows include:

- resolve free-text medication names to RxNorm ingredient concepts with
  `find_ingredients()` and `resolve()`;
- expand ingredient RxCUIs to product RxCUIs with
  `products_for_ingredients()`;
- include active-only or active plus historical RxNorm concepts using
  `concept_status`;
- map between RxCUIs and NDCs with `map_to()`, `map_rxcui_to_ndc()`, and
  `map_ndc_to_rxcui()`;
- retrieve product metadata with `get_properties()` and
  `get_clinical_attributes()`;
- filter products by route with `filter_products_by_route()`;
- query selected RxClass relationships with helpers such as `get_atc()`,
  `get_epc()`, and `get_medrt()`;
- use predefined RxNorm term type sets with `product_ttys()` and
  `tty_catalogue()`.

## Installation

You can install the development version of `rxref` from
[GitHub](https://github.com/) with:

``` r
# install.packages("pak")
# pak::pak("ssmithm/rxref")
```

## Basic Workflow: ingredients to products

This is a basic example for querying product RxCUIs associated with the
ingredient lisinopril (a blood pressure-lowering drug):

``` r
library(rxref)
library(dplyr)
#> 
#> Attaching package: 'dplyr'
#> The following objects are masked from 'package:stats':
#> 
#>     filter, lag
#> The following objects are masked from 'package:base':
#> 
#>     intersect, setdiff, setequal, union

lisinopril_ing <- find_ingredients("lisinopril")

lisinopril_products <- products_for_ingredients(
  lisinopril_ing$rxcui,
  ttys = product_ttys("default"),
  include_combos = TRUE
)

# Resulting table 
lisinopril_products |> 
  dplyr::mutate(name = paste0("`", name, "`")) |> 
  knitr::kable()
```

| ingredient_rxcui | product_rxcui | name | tty | n_ingredients |
|:---|:---|:---|:---|---:|
| 29046 | 104375 | `lisinopril 2.5 MG Oral Tablet [Zestril]` | SBD | 1 |
| 29046 | 104376 | `lisinopril 5 MG Oral Tablet [Zestril]` | SBD | 1 |
| 29046 | 104377 | `lisinopril 10 MG Oral Tablet [Zestril]` | SBD | 1 |
| 29046 | 104378 | `lisinopril 20 MG Oral Tablet [Zestril]` | SBD | 1 |
| 29046 | 1806890 | `lisinopril 1 MG/ML Oral Solution [Qbrelis]` | SBD | 1 |
| 29046 | 206765 | `lisinopril 10 MG Oral Tablet [Prinivil]` | SBD | 1 |
| 29046 | 206766 | `lisinopril 20 MG Oral Tablet [Prinivil]` | SBD | 1 |
| 29046 | 206771 | `lisinopril 40 MG Oral Tablet [Zestril]` | SBD | 1 |
| 29046 | 213482 | `lisinopril 30 MG Oral Tablet [Zestril]` | SBD | 1 |
| 29046 | 823971 | `hydrochlorothiazide 25 MG / lisinopril 20 MG Oral Tablet [Zestoretic]` | SBD | 2 |
| 29046 | 823982 | `hydrochlorothiazide 12.5 MG / lisinopril 20 MG Oral Tablet [Zestoretic]` | SBD | 2 |
| 29046 | 823986 | `hydrochlorothiazide 12.5 MG / lisinopril 10 MG Oral Tablet [Zestoretic]` | SBD | 2 |
| 29046 | 1806884 | `lisinopril 1 MG/ML Oral Solution` | SCD | 1 |
| 29046 | 197884 | `lisinopril 40 MG Oral Tablet` | SCD | 1 |
| 29046 | 197885 | `hydrochlorothiazide 12.5 MG / lisinopril 10 MG Oral Tablet` | SCD | 2 |
| 29046 | 197886 | `hydrochlorothiazide 12.5 MG / lisinopril 20 MG Oral Tablet` | SCD | 2 |
| 29046 | 197887 | `hydrochlorothiazide 25 MG / lisinopril 20 MG Oral Tablet` | SCD | 2 |
| 29046 | 205326 | `lisinopril 30 MG Oral Tablet` | SCD | 1 |
| 29046 | 311353 | `lisinopril 2.5 MG Oral Tablet` | SCD | 1 |
| 29046 | 311354 | `lisinopril 5 MG Oral Tablet` | SCD | 1 |
| 29046 | 314076 | `lisinopril 10 MG Oral Tablet` | SCD | 1 |
| 29046 | 314077 | `lisinopril 20 MG Oral Tablet` | SCD | 1 |

``` r


# Count of TTYs in resulting table
lisinopril_products |> 
  count(tty, n_ingredients, sort=TRUE)
#> # A tibble: 4 × 3
#>   tty   n_ingredients     n
#>   <chr>         <int> <int>
#> 1 SBD               1     9
#> 2 SCD               1     7
#> 3 SBD               2     3
#> 4 SCD               2     3
```

## Active and historical RxNorm concepts

By default, `products_for_ingredients()` searches active RxNorm
concepts:

``` r
products_for_ingredients(
  "29046",
  ttys = product_ttys("extended_product"),
  concept_status = "active"
) |> knitr::kable()
```

| ingredient_rxcui | product_rxcui | name | tty | n_ingredients |
|:---|:---|:---|:---|---:|
| 29046 | 1806883 | lisinopril Oral Solution | SCDF | 1 |
| 29046 | 372613 | hydrochlorothiazide / lisinopril Oral Tablet | SCDF | 2 |
| 29046 | 372614 | lisinopril Oral Tablet | SCDF | 1 |
| 29046 | 1162124 | hydrochlorothiazide / lisinopril Oral Product | SCDG | 2 |
| 29046 | 1162125 | hydrochlorothiazide / lisinopril Pill | SCDG | 2 |
| 29046 | 1164689 | lisinopril Oral Product | SCDG | 1 |
| 29046 | 1164690 | lisinopril Pill | SCDG | 1 |
| 29046 | 1806882 | lisinopril Oral Liquid Product | SCDG | 1 |
| 29046 | 104375 | lisinopril 2.5 MG Oral Tablet \[Zestril\] | SBD | 1 |
| 29046 | 104376 | lisinopril 5 MG Oral Tablet \[Zestril\] | SBD | 1 |
| 29046 | 104377 | lisinopril 10 MG Oral Tablet \[Zestril\] | SBD | 1 |
| 29046 | 104378 | lisinopril 20 MG Oral Tablet \[Zestril\] | SBD | 1 |
| 29046 | 1806890 | lisinopril 1 MG/ML Oral Solution \[Qbrelis\] | SBD | 1 |
| 29046 | 206765 | lisinopril 10 MG Oral Tablet \[Prinivil\] | SBD | 1 |
| 29046 | 206766 | lisinopril 20 MG Oral Tablet \[Prinivil\] | SBD | 1 |
| 29046 | 206771 | lisinopril 40 MG Oral Tablet \[Zestril\] | SBD | 1 |
| 29046 | 213482 | lisinopril 30 MG Oral Tablet \[Zestril\] | SBD | 1 |
| 29046 | 823971 | hydrochlorothiazide 25 MG / lisinopril 20 MG Oral Tablet \[Zestoretic\] | SBD | 2 |
| 29046 | 823982 | hydrochlorothiazide 12.5 MG / lisinopril 20 MG Oral Tablet \[Zestoretic\] | SBD | 2 |
| 29046 | 823986 | hydrochlorothiazide 12.5 MG / lisinopril 10 MG Oral Tablet \[Zestoretic\] | SBD | 2 |
| 29046 | 1806887 | lisinopril Oral Solution \[Qbrelis\] | SBDF | 1 |
| 29046 | 368248 | lisinopril Oral Tablet \[Zestril\] | SBDF | 1 |
| 29046 | 368865 | lisinopril Oral Tablet \[Prinivil\] | SBDF | 1 |
| 29046 | 823970 | hydrochlorothiazide / lisinopril Oral Tablet \[Zestoretic\] | SBDF | 2 |
| 29046 | 1178275 | Prinivil Oral Product | SBDG | 1 |
| 29046 | 1178276 | Prinivil Pill | SBDG | 1 |
| 29046 | 1187277 | Zestoretic Oral Product | SBDG | 2 |
| 29046 | 1187278 | Zestoretic Pill | SBDG | 2 |
| 29046 | 1187283 | Zestril Oral Product | SBDG | 1 |
| 29046 | 1187284 | Zestril Pill | SBDG | 1 |
| 29046 | 1806888 | Qbrelis Oral Liquid Product | SBDG | 1 |
| 29046 | 1806889 | Qbrelis Oral Product | SBDG | 1 |
| 29046 | 1806884 | lisinopril 1 MG/ML Oral Solution | SCD | 1 |
| 29046 | 197884 | lisinopril 40 MG Oral Tablet | SCD | 1 |
| 29046 | 197885 | hydrochlorothiazide 12.5 MG / lisinopril 10 MG Oral Tablet | SCD | 2 |
| 29046 | 197886 | hydrochlorothiazide 12.5 MG / lisinopril 20 MG Oral Tablet | SCD | 2 |
| 29046 | 197887 | hydrochlorothiazide 25 MG / lisinopril 20 MG Oral Tablet | SCD | 2 |
| 29046 | 205326 | lisinopril 30 MG Oral Tablet | SCD | 1 |
| 29046 | 311353 | lisinopril 2.5 MG Oral Tablet | SCD | 1 |
| 29046 | 311354 | lisinopril 5 MG Oral Tablet | SCD | 1 |
| 29046 | 314076 | lisinopril 10 MG Oral Tablet | SCD | 1 |
| 29046 | 314077 | lisinopril 20 MG Oral Tablet | SCD | 1 |

For older prescribing (or dispensing) data, users may need products that
are no longer active in the current RxNorm release. Set
`concept_status = "active_and_historical"` to include active concepts
plus selected historical statuses:

``` r
products_for_ingredients(
  "1369",
  ttys = "SCD",
  include_combos = FALSE,
  concept_status = "active_and_historical"
) |> knitr::kable()
```

| ingredient_rxcui | product_rxcui | name | tty | n_ingredients | concept_status | active_start_date | active_end_date | release_start_date | release_end_date |
|:---|:---|:---|:---|---:|:---|:---|:---|:---|:---|
| 1369 | 197394 | bendroflumethiazide 5 MG Oral Tablet | SCD | 1 | Active | NA | NA | NA | NA |
| 1369 | 308614 | bendroflumethiazide 2.5 MG Oral Tablet | SCD | 1 | Active | NA | NA | NA | NA |
| 1369 | 197393 | bendroflumethiazide 10 MG Oral Tablet | SCD | 1 | Obsolete | 042005 | 062014 | 042005 |  |
| 1369 | 308617 | bendroflumethiazide 5 MG Extended Release Oral Tablet | SCD | 1 | Obsolete | 042005 | 092010 | 042005 |  |
| 1369 | 429502 | bendroflumethiazide 1.25 MG Oral Tablet | SCD | 1 | Obsolete | 042005 | 072005 | 042005 |  |

Historical searches return additional status and date fields, including
concept_status, active_start_date, active_end_date, release_start_date,
and release_end_date.

**Note**: Historical searches are slower than active-only searches
because they require broader RxNorm status lookups. `rxref` caches
repeated API calls where possible, but active-only queries should be
preferred when historical concepts are not needed.

## Clinical product attributes and route filtering

`get_clinical_attributes()` retrieves product-level attributes such as
ingredients, dose form, dose form group, route, strength, and ingredient
count.

``` r
attrs <- get_clinical_attributes(lisinopril_products$product_rxcui)

attrs |>
  select(rxcui, related_rxcui, name, strength, dose_form, route, tty)
#> # A tibble: 22 × 7
#>    rxcui   related_rxcui name                     strength dose_form route tty  
#>    <chr>   <chr>         <chr>                    <chr>    <chr>     <chr> <chr>
#>  1 104375  104375        lisinopril 2.5 MG Oral … 2.5 MG   Oral Tab… ORAL  SBD  
#>  2 104376  104376        lisinopril 5 MG Oral Ta… 5 MG     Oral Tab… ORAL  SBD  
#>  3 104377  104377        lisinopril 10 MG Oral T… 10 MG    Oral Tab… ORAL  SBD  
#>  4 104378  104378        lisinopril 20 MG Oral T… 20 MG    Oral Tab… ORAL  SBD  
#>  5 1806890 1806890       lisinopril 1 MG/ML Oral… 1 MG/ML  Oral Sol… ORAL  SBD  
#>  6 206765  206765        lisinopril 10 MG Oral T… 10 MG    Oral Tab… ORAL  SBD  
#>  7 206766  206766        lisinopril 20 MG Oral T… 20 MG    Oral Tab… ORAL  SBD  
#>  8 206771  206771        lisinopril 40 MG Oral T… 40 MG    Oral Tab… ORAL  SBD  
#>  9 213482  213482        lisinopril 30 MG Oral T… 30 MG    Oral Tab… ORAL  SBD  
#> 10 823971  823971        hydrochlorothiazide 25 … 25 MG /… Oral Tab… ORAL  SBD  
#> # ℹ 12 more rows
```

Products can also be filtered by route:

``` r
oral_lisinopril <- filter_products_by_route(
  lisinopril_products,
  route = "ORAL"
)

oral_lisinopril 
#> # A tibble: 22 × 8
#>    ingredient_rxcui product_rxcui name       tty   n_ingredients route dose_form
#>    <chr>            <chr>         <chr>      <chr>         <int> <chr> <chr>    
#>  1 29046            104375        lisinopri… SBD               1 ORAL  Oral Tab…
#>  2 29046            104376        lisinopri… SBD               1 ORAL  Oral Tab…
#>  3 29046            104377        lisinopri… SBD               1 ORAL  Oral Tab…
#>  4 29046            104378        lisinopri… SBD               1 ORAL  Oral Tab…
#>  5 29046            1806890       lisinopri… SBD               1 ORAL  Oral Sol…
#>  6 29046            206765        lisinopri… SBD               1 ORAL  Oral Tab…
#>  7 29046            206766        lisinopri… SBD               1 ORAL  Oral Tab…
#>  8 29046            206771        lisinopri… SBD               1 ORAL  Oral Tab…
#>  9 29046            213482        lisinopri… SBD               1 ORAL  Oral Tab…
#> 10 29046            823971        hydrochlo… SBD               2 ORAL  Oral Tab…
#> # ℹ 12 more rows
#> # ℹ 1 more variable: dose_form_group <chr>
```

## Mapping between RxCUIs and NDCs

RxCUIs can be mapped to NDCs (and vice-versa):

``` r
ndcs <- map_rxcui_to_ndc("314077")

ndcs
#> # A tibble: 249 × 3
#>    rxcui  ndc11       ndc_status
#>    <chr>  <chr>       <chr>     
#>  1 314077 00143971301 ACTIVE    
#>  2 314077 00143971310 ACTIVE    
#>  3 314077 00185062001 ACTIVE    
#>  4 314077 00185062010 ACTIVE    
#>  5 314077 00378207501 ACTIVE    
#>  6 314077 00378207510 ACTIVE    
#>  7 314077 00591040801 ACTIVE    
#>  8 314077 00591040810 ACTIVE    
#>  9 314077 00615825505 ACTIVE    
#> 10 314077 00615825530 ACTIVE    
#> # ℹ 239 more rows
```

The direction-agnostic `map_to()` function can be used when you want to
map between identifier types with one interface:

``` r
map_to("314077", to = "ndc") 
#> # A tibble: 249 × 3
#>    rxcui  ndc11       ndc_status
#>    <chr>  <chr>       <chr>     
#>  1 314077 00143971301 ACTIVE    
#>  2 314077 00143971310 ACTIVE    
#>  3 314077 00185062001 ACTIVE    
#>  4 314077 00185062010 ACTIVE    
#>  5 314077 00378207501 ACTIVE    
#>  6 314077 00378207510 ACTIVE    
#>  7 314077 00591040801 ACTIVE    
#>  8 314077 00591040810 ACTIVE    
#>  9 314077 00615825505 ACTIVE    
#> 10 314077 00615825530 ACTIVE    
#> # ℹ 239 more rows
```

## Drug class relationships

`rxref` includes selected helpers for querying RxClass relationships,
including ATC, EPC, and MED-RT classes.

``` r
get_atc("29046")
#> # A tibble: 1 × 10
#>   input rxcui drug_name  drug_tty class_id class_name class_type class_url rela 
#>   <chr> <chr> <chr>      <chr>    <chr>    <chr>      <chr>      <chr>     <chr>
#> 1 29046 29046 lisinopril IN       C09AA    ACE inhib… ATC1-4     <NA>      ""   
#> # ℹ 1 more variable: rela_source <chr>
get_epc("29046")
#> # A tibble: 2 × 10
#>   input rxcui  drug_name drug_tty class_id class_name class_type class_url rela 
#>   <chr> <chr>  <chr>     <chr>    <chr>    <chr>      <chr>      <chr>     <chr>
#> 1 29046 15460… lisinopr… PIN      N000017… Angiotens… EPC        <NA>      has_…
#> 2 29046 15460… lisinopr… PIN      N000017… Angiotens… EPC        <NA>      has_…
#> # ℹ 1 more variable: rela_source <chr>
get_medrt("29046")
#> # A tibble: 43 × 10
#>    input rxcui drug_name drug_tty class_id class_name class_type class_url rela 
#>    <chr> <chr> <chr>     <chr>    <chr>    <chr>      <chr>      <chr>     <chr>
#>  1 29046 1546… lisinopr… PIN      D000799  Angioedema DISEASE    <NA>      ci_w…
#>  2 29046 1546… lisinopr… PIN      D003928  Diabetic … DISEASE    <NA>      may_…
#>  3 29046 1546… lisinopr… PIN      D003928  Diabetic … DISEASE    <NA>      may_…
#>  4 29046 1546… lisinopr… PIN      D004342  Drug Hype… DISEASE    <NA>      ci_w…
#>  5 29046 1546… lisinopr… PIN      D006333  Heart Fai… DISEASE    <NA>      may_…
#>  6 29046 1546… lisinopr… PIN      D006929  Hyperaldo… DISEASE    <NA>      ci_w…
#>  7 29046 1546… lisinopr… PIN      D006973  Hypertens… DISEASE    <NA>      may_…
#>  8 29046 1546… lisinopr… PIN      D007022  Hypotensi… DISEASE    <NA>      ci_w…
#>  9 29046 1546… lisinopr… PIN      D009203  Myocardia… DISEASE    <NA>      may_…
#> 10 29046 1546… lisinopr… PIN      D011247  Pregnancy  DISEASE    <NA>      ci_w…
#> # ℹ 33 more rows
#> # ℹ 1 more variable: rela_source <chr>
```

## API use and caching

`rxref` queries public RxNorm and RxClass APIs. The package includes
configurable rate-delay and caching behavior to reduce repeated API
requests and to support longer medication-definition workflows. Note
that RxNorm indicates a rate of [no more than 20 API requests per second
per IP](https://lhncbc.nlm.nih.gov/RxNav/TermsofService.html).

``` r
rxref_conf()
#> $base_url
#> [1] "https://rxnav.nlm.nih.gov/REST"
#> 
#> $rate_delay
#> [1] 0.1
#> 
#> $cache
#> <cache_mem> <cachem>
#>   Methods:
#>     get(key, missing = missing_)
#>     set(key, value)
#>     exists(key)
#>     keys()
#>     remove(key)
#>     reset()
#>     prune()
#>     size()
#>     info()
rxref_conf(rate_delay = 0.1)
#> $base_url
#> [1] "https://rxnav.nlm.nih.gov/REST"
#> 
#> $rate_delay
#> [1] 0.1
#> 
#> $cache
#> <cache_mem> <cachem>
#>   Methods:
#>     get(key, missing = missing_)
#>     set(key, value)
#>     exists(key)
#>     keys()
#>     remove(key)
#>     reset()
#>     prune()
#>     size()
#>     info()
```

## Notes

RxNorm concepts and NDC mappings can change over time. For reproducible
research, consider recording the date of API access and retaining
intermediate drug-definition outputs used in analyses.

`rxref` is intended to support research workflows and
medication-definition development. Users should review outputs carefully
before using them in any high-stakes regulatory, clinical, or production
settings.

## ----include = FALSE----------------------------------------------------------
# Vignettes use precomputed example data by default.
# To rebuild examples with live RxNorm/RxClass API calls, set:
# Sys.setenv(RXREF_BUILD_VIGNETTES_ONLINE = "true")

online_env <- identical(
  tolower(Sys.getenv("RXREF_BUILD_VIGNETTES_ONLINE")),
  "true"
)

has_net <- tryCatch({
  requireNamespace("curl", quietly = TRUE) && curl::has_internet()
}, error = function(e) FALSE)

run_live <- online_env && has_net

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(rxref)
library(dplyr)

read_rxref_example <- function(file) {
  path <- system.file("extdata", file, package = "rxref")
  if (!nzchar(path)) {
    stop(
      "The example data file '", file, "' was not found. ",
      "Reinstall rxref or rebuild the vignette with ",
      "RXREF_BUILD_VIGNETTES_ONLINE=true."
    )
  }
  readRDS(path)
}

## ----eval = FALSE-------------------------------------------------------------
# Sys.setenv(RXREF_BUILD_VIGNETTES_ONLINE = "true")

## ----define-ingredients-------------------------------------------------------
glp1.names <- c(
  "semaglutide",
  "exenatide",
  "liraglutide",
  "lixisenatide",
  "dulaglutide",
  "albiglutide",
  "tirzepatide"
)

## ----ings---------------------------------------------------------------------
if (run_live) {
  glp1.ings <- find_ingredients(glp1.names) |>
    filter(tty == "IN") |>
    distinct(
      input,
      ingredient_rxcui = rxcui,
      ingredient_name = name,
      ingredient_tty = tty
    )
} else {
  glp1.ings <- read_rxref_example("glp1_ings.rds")
}

glp1.ings

## ----tty-sets-----------------------------------------------------------------
product_ttys("default")
product_ttys("extended_product")

## ----prods--------------------------------------------------------------------
if (run_live) {
  glp1.prods <- products_for_ingredients(
    glp1.ings$ingredient_rxcui,
    ttys = product_ttys("default"),
    include_combos = TRUE,
    concept_status = "active"
  )
} else {
  glp1.prods <- read_rxref_example("glp1_prods.rds")
}

glp1.prods |>
  head(30)

## ----historical-products, eval = FALSE----------------------------------------
# glp1.prods_historical <- products_for_ingredients(
#   glp1.ings$ingredient_rxcui,
#   ttys = product_ttys("default"),
#   include_combos = TRUE,
#   concept_status = "active_and_historical"
# )

## ----ndcs---------------------------------------------------------------------
if (run_live) {
  glp1.ndc.map <- map_rxcui_to_ndc(
    unique(glp1.prods$product_rxcui),
    status = "ACTIVE"
  )
} else {
  glp1.ndc.map <- read_rxref_example("glp1_ndc_map.rds")
}

glp1.ndcs <- glp1.ndc.map |>
  left_join(
    glp1.prods,
    by = c("rxcui" = "product_rxcui")
  ) |>
  left_join(
    glp1.ings |>
      select(ingredient_rxcui, ingredient_name),
    by = "ingredient_rxcui"
  ) |>
  distinct(
    ingredient_rxcui,
    ingredient_name,
    product_rxcui = rxcui,
    ndc11,
    ndc_status,
    name,
    tty
  ) |>
  arrange(ingredient_name, product_rxcui, ndc11)

glp1.ndcs |>
  head(30)

## ----search-------------------------------------------------------------------
if (run_live) {
  alt.glp1.ndcs <- search_drug(
    term = glp1.names,
    return = "ndc",
    concept_status = "active",
    ndc_status = c("ACTIVE", "OBSOLETE", "UNSPECIFIED")
  )
} else {
  alt.glp1.ndcs <- read_rxref_example("alt_glp1_ndc.rds")
}

alt.glp1.ndcs |>
  arrange(ingredient_name, product_rxcui, ndc11) |>
  head(30)

## ----search-historical, eval = FALSE------------------------------------------
# search_drug(
#   term = glp1.names,
#   return = "ndc",
#   concept_status = "active_and_historical",
#   ndc_status = c("ACTIVE", "OBSOLETE", "UNSPECIFIED")
# )

## ----check--------------------------------------------------------------------
glp1.ndcs |>
  filter(!is.na(ndc11)) |>
  arrange(ingredient_name, product_rxcui, ndc11) |>
  head(30)

## ----status-example, eval = FALSE---------------------------------------------
# search_drug(
#   term = "semaglutide",
#   return = "ndc",
#   concept_status = "active_and_historical",
#   ndc_status = c("ACTIVE", "OBSOLETE", "UNSPECIFIED")
# )

## ----setup, include = FALSE---------------------------------------------------
# Vignettes use precomputed example data by default.
# To rebuild examples with live RxNorm/RxClass API calls, set:
# Sys.setenv(RXREF_BUILD_VIGNETTES_ONLINE = "true")

online_env <- identical(
  tolower(Sys.getenv("RXREF_BUILD_VIGNETTES_ONLINE")),
  "true"
)

has_net <- tryCatch({
  requireNamespace("curl", quietly = TRUE) && curl::has_internet()
}, error = function(e) FALSE)

run_live <- online_env && has_net

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(rxref)
library(dplyr)

read_rxref_example <- function(file) {
  path <- system.file("extdata", file, package = "rxref")
  if (!nzchar(path)) {
    stop(
      "The example data file '", file, "' was not found. ",
      "Reinstall rxref or rebuild the vignette with ",
      "RXREF_BUILD_VIGNETTES_ONLINE=true."
    )
  }
  readRDS(path)
}

## ----eval = FALSE-------------------------------------------------------------
# Sys.setenv(RXREF_BUILD_VIGNETTES_ONLINE = "true")

## ----define-------------------------------------------------------------------
beta_blocker_names <- c(
  "acebutolol",
  "atenolol",
  "betaxolol",
  "bisoprolol",
  "carvedilol",
  "labetalol",
  "metoprolol",
  "nadolol",
  "nebivolol",
  "penbutolol",
  "pindolol",
  "propranolol",
  "sotalol",
  "timolol"
)

## ----resolve------------------------------------------------------------------
if (run_live) {
  bb_ingredients <- find_ingredients(beta_blocker_names) |>
    filter(tty == "IN") |>
    distinct(
      input,
      ingredient_rxcui = rxcui,
      ingredient_name = name,
      ingredient_tty = tty
    )
} else {
  bb_ingredients <- read_rxref_example("bb_ingredients.rds")
}

bb_ingredients

## ----expand-------------------------------------------------------------------
if (run_live) {
  bb_products <- products_for_ingredients(
    bb_ingredients$ingredient_rxcui,
    ttys = product_ttys("default"),
    include_combos = TRUE,
    concept_status = "active"
  ) |>
    left_join(bb_ingredients, by = "ingredient_rxcui")
} else {
  bb_products <- read_rxref_example("bb_products.rds")
}

bb_products

## ----exclude-option-----------------------------------------------------------
if (run_live) {
  bb_single_ingredient_products <- products_for_ingredients(
    bb_ingredients$ingredient_rxcui,
    ttys = product_ttys("default"),
    include_combos = FALSE,
    concept_status = "active"
  ) |>
    left_join(bb_ingredients, by = "ingredient_rxcui")
} else {
  bb_single_ingredient_products <- read_rxref_example(
    "bb_single_ingredient_products.rds"
  )
}

bb_single_ingredient_products

## ----historical-products, eval = FALSE----------------------------------------
# bb_products_historical <- products_for_ingredients(
#   bb_ingredients$ingredient_rxcui,
#   ttys = product_ttys("default"),
#   include_combos = TRUE,
#   concept_status = "active_and_historical"
# )

## ----filtering-matters--------------------------------------------------------
if (run_live) {
  bb_attrs <- get_clinical_attributes(
    unique(bb_products$product_rxcui)
  ) |>
    rename(product_rxcui = rxcui)
} else {
  bb_attrs <- read_rxref_example("bb_attrs.rds")
}

bb_attrs |>
  count(route, dose_form_group, sort = TRUE)

## ----filter-------------------------------------------------------------------
if (run_live) {
  bb_oral_products <- bb_products |>
    filter_products_by_route(route = "ORAL")
} else {
  bb_oral_products <- read_rxref_example("bb_oral_products.rds")
}

bb_oral_products

## ----filter-summary-----------------------------------------------------------
bb_oral_products |>
  count(route, dose_form_group, sort = TRUE)

## ----map----------------------------------------------------------------------
if (run_live) {
  bb_oral_ndc_map <- map_rxcui_to_ndc(
    unique(bb_oral_products$product_rxcui),
    status = "ACTIVE"
  )

  bb_oral_ndcs <- bb_oral_ndc_map |>
    left_join(
      bb_oral_products,
      by = c("rxcui" = "product_rxcui")
    ) |>
    rename(
      product_rxcui = rxcui,
    ) |>
    distinct()
} else {
  bb_oral_ndcs <- read_rxref_example("bb_oral_ndcs.rds")
}

bb_oral_ndcs

## ----map-select---------------------------------------------------------------
bb_oral_ndcs |>
  select(
    ingredient_name,
    product_rxcui,
    name,
    tty,
    route,
    dose_form,
    ndc11,
    ndc_status
  )

## ----summarize-products-------------------------------------------------------
bb_oral_products |>
  count(ingredient_name, sort = TRUE, name = "n_product_rxcuis")

## ----summarize-ndcs-----------------------------------------------------------
bb_oral_ndcs |>
  count(ingredient_name, sort = TRUE, name = "n_active_ndcs")

## ----summarize-combinations---------------------------------------------------
bb_oral_products |>
  filter(n_ingredients > 1) |>
  select(
    ingredient_name,
    product_rxcui,
    name,
    tty,
    n_ingredients,
    route,
    dose_form
  )

## ----alt-search-rxcui---------------------------------------------------------
if (run_live) {
  bb_oral_rxcuis <- search_drug(
    beta_blocker_names,
    return = "rxcui",
    route = "ORAL",
    include_combos = TRUE,
    concept_status = "active"
  )
} else {
  bb_oral_rxcuis <- read_rxref_example("bb_oral_rxcuis_search.rds")
}

bb_oral_rxcuis

## ----alt-search-ndc-----------------------------------------------------------
if (run_live) {
  bb_oral_ndcs_search_raw <- search_drug(
    beta_blocker_names,
    return = "ndc",
    route = "ORAL",
    ndc_status = "ACTIVE",
    include_combos = TRUE,
    concept_status = "active"
  )

  bb_oral_ndcs_search <- bb_oral_ndcs_search_raw

  if (!"ingredient_name" %in% names(bb_oral_ndcs_search)) {
    bb_oral_ndcs_search <- bb_oral_ndcs_search |>
      left_join(
        bb_ingredients |>
          select(ingredient_rxcui, ingredient_name),
        by = "ingredient_rxcui"
      )
  }

  product_cols <- c(
    "name",
    "tty",
    "route",
    "dose_form",
    "dose_form_group"
  )

  missing_product_cols <- setdiff(product_cols, names(bb_oral_ndcs_search))

  if (length(missing_product_cols) > 0) {
    bb_oral_ndcs_search <- bb_oral_ndcs_search |>
      left_join(
        bb_oral_products |>
          select(
            ingredient_rxcui,
            product_rxcui,
            all_of(missing_product_cols)
          ),
        by = c("ingredient_rxcui", "product_rxcui")
      )
  }

  bb_oral_ndcs_search <- bb_oral_ndcs_search |>
    distinct()
} else {
  bb_oral_ndcs_search <- read_rxref_example("bb_oral_ndcs_search.rds")
}

bb_oral_ndcs_search

## ----products-and-ndcs--------------------------------------------------------
if (run_live) {
  bb_oral_both <- search_drug(
    beta_blocker_names,
    return = "both",
    route = "ORAL",
    ndc_status = "ACTIVE",
    include_combos = TRUE,
    concept_status = "active"
  )
} else {
  bb_oral_both <- read_rxref_example("bb_oral_both_search.rds")
}

names(bb_oral_both)

## ----products-and-ndcs-products-----------------------------------------------
bb_oral_both$products

## ----products-and-ndcs-ndcs---------------------------------------------------
bb_oral_both$ndcs

## ----qc-nonoral-names---------------------------------------------------------
bb_oral_products |>
  filter(grepl(
    "Injection|Injectable|Ophthalmic|Topical|Transdermal|Nasal|Inhalation",
    name,
    ignore.case = TRUE
  ))

## ----qc-routes----------------------------------------------------------------
bb_oral_products |>
  count(route, dose_form_group, sort = TRUE)

## ----qc-combinations----------------------------------------------------------
bb_oral_products |>
  filter(n_ingredients > 1) |>
  count(ingredient_name, sort = TRUE)

## ----qc-without-ndcs----------------------------------------------------------
bb_oral_products |>
  anti_join(
    bb_oral_ndcs |>
      distinct(product_rxcui),
    by = "product_rxcui"
  ) |>
  select(
    ingredient_name,
    product_rxcui,
    name,
    tty,
    route,
    dose_form
  )

#' Get RxNav status for NDCs
#'
#' `ndc_status()` requests the status of an NDC entry for one or more user-supplied
#' NDC numbers. The returned status will be one "ACTIVE", "OBSOLETE", or "UNSPECIFIED".
#' On occasion, a NULL value may be pulled (resulting in an `NA` in the returned tibble),
#' when RxNorm has no status for a given NDC, or if an invalid NDC is supplied.
#'
#' @param ndc character vector; hyphenated or digits
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble with three columns: ndc_input, ndc11, ndc_hyph, status
#'
#' @export
ndc_status <- function(ndc, show_progress = interactive()) {
  stopifnot(is.character(ndc))

  .rxref_progress_map_dfr(
    ndc,
    function(x) {
      ndc11 <- ndc_to_11(x)
      ndc_h <- hyphenate_ndc_5_4_2(ndc11)

      st <- tryCatch(
        rx_get_json("/ndcstatus", query = list(ndc = ndc_h)),
        error = function(e) NULL
      )

      tibble::tibble(
        ndc_input = x,
        ndc11 = ndc11,
        ndc_hyph = ndc_h,
        status = if (
          is.null(st) ||
          is.null(st$ndcStatus$status)
        ) {
          NA_character_
        } else {
          as.character(st$ndcStatus$status)
        }
      )
    },
    name = "Checking NDC status",
    show_progress = show_progress
  )
}

#' Clinical attributes from the concept (SCD/SBD) or related SCD/SBD
#'
#' If `rxcui` is already a clinical drug (TTY = SCD or SBD), parse attributes directly
#' from its own name. Otherwise, query \code{/rxcui/{id}/related?tty=SCD,SBD} and parse.
#'
#' In addition to strength and dose form, this returns route, dose-form group (DFG),
#' brand/generic flags, ingredient summaries, and a simple active/inactive status.
#'
#' Note there is a fair amount of parsing of the RxNorm STR value to try to extract
#' relevant information (e.g., strength, dose_form), so check closely before trusting.
#' There may be edge cases that are not correctly parsed.
#'
#' For combination products, ingredient-related columns may contain multiple
#' semicolon-delimited values, such as `"amlodipine; valsartan"`.
#'
#' @param rxcui Character vector of RxCUIs
#' @param include_historical Logical. If `TRUE`, use RxCUI history status
#'   metadata as a fallback for RxCUIs that do not return active clinical
#'   attributes. This is useful for obsolete, remapped, quantified, or otherwise
#'   non-current RxCUIs found in historical prescribing data.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble with columns:
#'   \describe{
#'     \item{rxcui}{Input RxCUI}
#'     \item{related_rxcui}{Clinical drug RxCUI (SCD/SBD) used for attributes}
#'     \item{name}{Clinical drug name}
#'     \item{tty}{Term type (SCD/SBD, etc.)}
#'     \item{strength}{Parsed strength string (e.g. "500 MG")}
#'     \item{dose_form}{Parsed dose form (e.g. "Extended Release Oral Tablet")}
#'     \item{route}{Route parsed from dose form / DFG (e.g. "ORAL", "INJECTION")}
#'     \item{dose_form_group}{Dose form group (DFG), if available}
#'     \item{is_brand}{Logical; TRUE for branded clinical concepts (SBD/BPCK)}
#'     \item{is_generic}{Logical; TRUE for generic clinical concepts (SCD/GPCK)}
#'     \item{ingredient_count}{Number of distinct ingredients}
#'     \item{ingredient_rxcui}{Ingredient RxCUI. For combination products,
#'       multiple values are returned as semicolon-delimited strings.}
#'     \item{ingredient_name}{Ingredient name. For combination products,
#'       multiple values are returned as semicolon-delimited strings.}
#'     \item{ingredient_tty}{Ingredient term type (IN/PIN/MIN). For combination
#'       products, multiple values are returned as semicolon-delimited strings.}
#'     \item{is_multi_ingredient}{Logical; TRUE if >1 ingredient}
#'     \item{suppress}{Raw RxNorm suppress flag from properties}
#'     \item{status}{Simple status derived from suppress: "ACTIVE" vs "INACTIVE"}
#'   }
#' @examples
#' \dontrun{
#' get_clinical_attributes(c("861007","860975")) |>
#'   dplyr::select(rxcui, related_rxcui, name, strength, dose_form, route, tty) |>
#'   head()
#' }
#' @export
get_clinical_attributes <- function(rxcui,
                                    include_historical = FALSE,
                                    show_progress = interactive()) {
  stopifnot(is.character(rxcui))

  rxcui_ids <- unique(stats::na.omit(rxcui))

  # 1. Core clinical mapping
  core <- .rxref_progress_map_dfr(
    rxcui_ids,
    function(id) {
      # 1) Pull properties to know TTY and get name
      props <- tryCatch(
        rx_get_json(paste0("/rxcui/", id, "/properties")),
        error = function(e) NULL
      )

      p <- rx_pluck(props, "properties", .default = list())
      tty_self <- rx_scalar_chr(rx_pluck_chr(p, "tty"))
      name_self <- rx_scalar_chr(rx_pluck_chr(p, "name"))

      # If the concept is already SCD/SBD, parse it
      if (!is.na(tty_self) && tty_self %in% c("SCD", "SBD")) {
        pr <- .rxref_parse_strength_dose(name_self)

        return(tibble::tibble(
          rxcui         = id,
          related_rxcui = id,
          name          = name_self,
          tty           = tty_self,
          strength      = pr$strength,
          dose_form     = pr$dose_form
        ))
      }

      # Otherwise, fetch related SCD/SBD
      rel <- tryCatch(
        rx_get_json(
          paste0("/rxcui/", id, "/related"),
          query = list(tty = "SCD SBD")
        ),
        error = function(e) NULL
      )

      groups <- rx_pluck_list(rel, "relatedGroup", "conceptGroup")

      if (!length(groups)) {
        return(tibble::tibble(
          rxcui         = id,
          related_rxcui = NA_character_,
          name          = NA_character_,
          tty           = NA_character_,
          strength      = NA_character_,
          dose_form     = NA_character_
        ))
      }

      # Flatten conceptProperties across groups robustly
      concepts <- purrr::map(groups, \(g) rx_pluck_list(g, "conceptProperties")) |>
        purrr::compact() |>
        unlist(recursive = FALSE)

      if (!length(concepts)) {
        return(tibble::tibble(
          rxcui         = id,
          related_rxcui = NA_character_,
          name          = NA_character_,
          tty           = NA_character_,
          strength      = NA_character_,
          dose_form     = NA_character_
        ))
      }

      purrr::map_dfr(concepts, function(cp) {
        nm <- rx_scalar_chr(rx_pluck_chr(cp, "name"))
        tty <- rx_scalar_chr(rx_pluck_chr(cp, "tty"))
        pr <- .rxref_parse_strength_dose(nm)

        tibble::tibble(
          rxcui         = id,
          related_rxcui = rx_scalar_chr(rx_pluck_chr(cp, "rxcui")),
          name          = nm,
          tty           = tty,
          strength      = pr$strength,
          dose_form     = pr$dose_form
        )
      })
    },
    name = "Getting clinical attributes",
    show_progress = show_progress
  )

  # nothing else to do if everything failed
  if (!nrow(core)) return(core)

  # 2. Brand / generic flags
  core <- core |>
    dplyr::mutate(
      is_brand   = .data$tty %in% c("SBD", "BPCK"),
      is_generic = .data$tty %in% c("SCD", "GPCK")
    )

  clinical_ids <- unique(core$related_rxcui[!is.na(core$related_rxcui)])

  # if no clinical concepts, just add empty columns and return
  if (!length(clinical_ids)) {
    out <- core |>
      dplyr::mutate(
        dose_form_group = NA_character_,
        ingredient_count = NA_integer_,
        ingredient_rxcui = NA_character_,
        ingredient_name = NA_character_,
        ingredient_tty = NA_character_,
        is_multi_ingredient = NA,
        suppress = NA_character_,
        status = NA_character_,
        route = NA_character_
      )

    if (!isTRUE(include_historical)) {
      return(out)
    }

    hist <- .rxref_clinical_attributes_historical(
      unique(out$rxcui),
      show_progress = show_progress
    ) |>
      dplyr::filter(
        !is.na(.data$name) |
          !is.na(.data$tty) |
          !is.na(.data$dose_form) |
          !is.na(.data$dose_form_group) |
          !is.na(.data$ingredient_count)
      )

    return(dplyr::bind_rows(
      dplyr::filter(out, !(.data$rxcui %in% hist$rxcui)),
      hist
    ) |>
      dplyr::distinct()
    )
  }

  # 3. Dose form group (DFG) from RxNorm (TTY = DFG)
  dfg_tbl <- purrr::map_dfr(clinical_ids, function(id) {
    rel <- tryCatch(
      rx_get_json(paste0("/rxcui/", id, "/related"), query = list(tty = "DFG")),
      error = function(e) NULL
    )
    groups <- rx_pluck_list(rel, "relatedGroup", "conceptGroup")
    if (!length(groups)) {
      return(tibble::tibble(
        related_rxcui   = id,
        dose_form_group = NA_character_
      ))
    }
    concepts <- purrr::map(groups, \(g) rx_pluck_list(g, "conceptProperties")) |>
      purrr::compact() |>
      unlist(recursive = FALSE)

    if (!length(concepts)) {
      return(tibble::tibble(
        related_rxcui   = id,
        dose_form_group = NA_character_
      ))
    }
    # take the first DFG name (usually there is only one)
    nm <- null2na(concepts[[1]]$name)
    tibble::tibble(
      related_rxcui   = id,
      dose_form_group = nm
    )
  })

  # 4. Ingredient summary
  ing_raw <- tryCatch(
    .rxref_get_ingredients_for_rxcui(
      clinical_ids,
      include_pin = TRUE,
      include_min = FALSE
    ),
    error = function(e) NULL
  )

  if (!is.null(ing_raw) && nrow(ing_raw)) {
    ing_raw <- ing_raw |>
      dplyr::filter(
        !is.na(.data$ingredient_rxcui) | !is.na(.data$ingredient_name)
      )
  }

  if (is.null(ing_raw) || !nrow(ing_raw)) {
    ing_summary <- tibble::tibble(
      related_rxcui        = clinical_ids,
      ingredient_count     = NA_integer_,
      ingredient_rxcui     = NA_character_,
      ingredient_name      = NA_character_,
      ingredient_tty       = NA_character_,
      is_multi_ingredient  = NA
    )
  } else {
    # Prefer IN over PIN when both present for a given product
    ing_raw <- ing_raw |>
      dplyr::group_by(.data$related_rxcui) |>
      dplyr::filter(!(.data$ingredient_tty == "PIN" & any(.data$ingredient_tty == "IN"))) |>
      dplyr::ungroup()

    ing_raw2 <- ing_raw |>
      dplyr::mutate(
        ingredient_key = dplyr::coalesce(
          .data$ingredient_rxcui,
          stringr::str_to_lower(.data$ingredient_name)
        )
      ) |>
      dplyr::filter(!is.na(.data$ingredient_key), nzchar(.data$ingredient_key)) |>
      dplyr::distinct(
        .data$related_rxcui,
        .data$ingredient_key,
        .keep_all = TRUE
      )

    ing_summary <- ing_raw2 |>
      dplyr::group_by(.data$related_rxcui) |>
      dplyr::summarise(
        ingredient_count = dplyr::n(),
        ingredient_rxcui = .rxref_collapse_values(.data$ingredient_rxcui),
        ingredient_name  = .rxref_collapse_values(.data$ingredient_name),
        ingredient_tty   = .rxref_collapse_values(.data$ingredient_tty),
        .groups = "drop"
      ) |>
      dplyr::mutate(
        is_multi_ingredient = .data$ingredient_count > 1L
      )
  }

  # 5. Status from suppress (via get_properties)
  props_clin <- tryCatch(
    get_properties(clinical_ids, show_progress = FALSE),
    error = function(e) tibble::tibble(rxcui = clinical_ids, suppress = NA_character_)
  )

  status_tbl <- props_clin |>
    dplyr::transmute(
      related_rxcui = .data$rxcui,
      suppress      = .data$suppress,
      status        = dplyr::case_when(
        is.na(.data$suppress) ~ NA_character_,
        .data$suppress == "N" ~ "ACTIVE",
        TRUE                  ~ "INACTIVE"
      )
    )

  # 6. Merge everything + expanded route mapping
  out <- core |>
    dplyr::left_join(dfg_tbl,     by = "related_rxcui") |>
    dplyr::left_join(ing_summary, by = "related_rxcui") |>
    dplyr::left_join(status_tbl,  by = "related_rxcui") |>
    dplyr::mutate(
      is_multi_ingredient = dplyr::if_else(
        is.na(.data$ingredient_count),
        NA,
        .data$ingredient_count > 1L
      ),
      route = .rxref_route_from_dose_form(
        .data$dose_form,
        .data$dose_form_group
      )
    )

  if (!isTRUE(include_historical)) {
    return(out)
  }

  needs_hist <- out |>
    dplyr::filter(
      is.na(.data$related_rxcui) |
        is.na(.data$dose_form) |
        is.na(.data$dose_form_group) |
        is.na(.data$route) |
        is.na(.data$ingredient_count)
    ) |>
    dplyr::pull(.data$rxcui) |>
    unique()

  if (!length(needs_hist)) {
    return(out)
  }

  hist <- .rxref_clinical_attributes_historical(
    needs_hist,
    show_progress = show_progress
  ) |>
    dplyr::filter(
      !is.na(.data$name) |
        !is.na(.data$tty) |
        !is.na(.data$dose_form) |
        !is.na(.data$dose_form_group) |
        !is.na(.data$ingredient_count)
    )

  out_active_keep <- out |>
    dplyr::filter(!(.data$rxcui %in% hist$rxcui))

  dplyr::bind_rows(
    out_active_keep,
    hist
  ) |>
    dplyr::distinct()
}



# Internal: collapse multiple scalar values into a semicolon-delimited string
#' @keywords internal
#' @noRd
.rxref_collapse_values <- function(x, delim = "; ") {
  if (is.null(x) || length(x) == 0L) {
    return(NA_character_)
  }

  x <- as.character(x)
  x <- trimws(x)
  x <- x[!is.na(x) & nzchar(x)]
  x <- unique(x)

  if (!length(x)) {
    return(NA_character_)
  }

  paste(x, collapse = delim)
}

# Internal: get ingredient concepts (IN/PIN/MIN) for one or more RxCUIs
#' @keywords internal
#' @noRd
.rxref_get_ingredients_for_rxcui <- function(rxcui,
                                             include_pin = TRUE,
                                             include_min = TRUE) {
  stopifnot(is.character(rxcui))

  ttys <- c("IN")
  if (include_pin) ttys <- c(ttys, "PIN")
  if (include_min) ttys <- c(ttys, "MIN")

  # tty_query <- paste(ttys, collapse = "+")
  tty_query <- paste(ttys, collapse = " ")  # "IN PIN MIN" -> encoded as IN+PIN+MIN

  purrr::map_dfr(rxcui, function(id) {
    if (is.na(id) || !nzchar(id)) {
      return(tibble::tibble(
        related_rxcui    = character(0),
        ingredient_rxcui = character(0),
        ingredient_name  = character(0),
        ingredient_tty   = character(0)
      ))
    }

    rel <- tryCatch(
      rx_get_json(
        paste0("/rxcui/", id, "/related"),
        query = list(tty = tty_query)
      ),
      error = function(e) NULL
    )

    groups <- rx_pluck_list(rel, "relatedGroup", "conceptGroup")
    if (!length(groups)) {
      return(tibble::tibble(
        related_rxcui    = character(0),
        ingredient_rxcui = character(0),
        ingredient_name  = character(0),
        ingredient_tty   = character(0)
      ))
    }

    concepts <- purrr::map(groups, \(g) rx_pluck_list(g, "conceptProperties")) |>
      purrr::compact() |>
      unlist(recursive = FALSE)

    if (!length(concepts)) {
      return(tibble::tibble(
        related_rxcui    = character(0),
        ingredient_rxcui = character(0),
        ingredient_name  = character(0),
        ingredient_tty   = character(0)
      ))
    }

    purrr::map_dfr(concepts, function(cp) {
      tibble::tibble(
        related_rxcui    = id,
        ingredient_rxcui = rx_scalar_chr(rx_pluck_chr(cp, "rxcui")),
        ingredient_name  = rx_scalar_chr(rx_pluck_chr(cp, "name")),
        ingredient_tty   = rx_scalar_chr(rx_pluck_chr(cp, "tty"))
      )
    })
  }) |>
    dplyr::distinct()
}

#' Parse strength and dose form from an RxNorm name
#'
#' @keywords internal
#' @noRd
.rxref_parse_strength_dose <- function(nm) {
  if (is.na(nm) || !nzchar(nm)) {
    return(list(strength = NA_character_, dose_form = NA_character_))
  }

  rx <- paste0(
    "(",
    "\\d{1,4}(?:[\\d,]*)(?:\\.\\d+)?",
    "\\s*(?:MCG|MG|G|KG|ML|L|MEQ|MMOL|IU|UNITS|%)\\b",
    "\\s*(?:/\\s*",
    "(?:\\d{1,4}(?:[\\d,]*)(?:\\.\\d+)?)?",
    "\\s*(?:MCG|MG|G|KG|ML|L|MEQ|MMOL|IU|UNITS|%|HR)\\b",
    ")?",
    ")"
  )

  m <- gregexpr(rx, nm, ignore.case = TRUE, perl = TRUE)
  starts <- m[[1]]

  if (length(starts) == 1L && starts[1] == -1L) {
    return(list(strength = NA_character_, dose_form = NA_character_))
  }

  lens <- attr(m[[1]], "match.length")
  matches <- regmatches(nm, m)[[1]]
  matches <- trimws(matches)

  has_ratio <- grepl("(/|per|%)", matches, ignore.case = TRUE)

  if (any(has_ratio)) {
    idx <- tail(which(has_ratio), 1L)
    strength <- matches[idx]
  } else if (length(matches) > 1L) {
    strength <- paste(matches, collapse = " / ")
    idx <- length(matches)
  } else {
    idx <- 1L
    strength <- matches[idx]
  }

  pos_end <- starts[idx] + lens[idx] - 1L
  tail <- trimws(substr(nm, pos_end + 1L, nchar(nm)))
  tail <- sub("^[,;:.-]\\s*", "", tail)

  list(
    strength = strength,
    dose_form = if (nzchar(tail)) tail else NA_character_
  )
}

#' Infer route from dose form and dose form group
#'
#' @keywords internal
#' @noRd
.rxref_route_from_dose_form <- function(dose_form, dose_form_group) {
  dplyr::case_when(
    dose_form_group %in% c(
      "Oral Product",
      "Disintegrating Oral Product",
      "Oral Liquid Product",
      "Oral Cream Product",
      "Oral Foam Product",
      "Oral Gel Product",
      "Oral Ointment Product",
      "Oral Paste Product",
      "Oral Powder Product",
      "Oral Spray Product",
      "Oral Strip Product",
      "Flake Product",
      "Granule Product",
      "Pellet Product",
      "Pill",
      "Lozenge Product",
      "Buccal Product",
      "Sublingual Product",
      "Wafer Product",
      "Mouthwash Product",
      "Toothpaste Product",
      "Dental Product"
    ) ~ "ORAL",

    dose_form_group %in% c(
      "Injectable Product",
      "Intraperitoneal Product",
      "Intratracheal Product",
      "Irrigation Product"
    ) ~ "INJECTION",

    dose_form_group == "Ophthalmic Product" ~ "OPHTHALMIC",
    dose_form_group == "Otic Product" ~ "OTIC",
    dose_form_group %in% c("Nasal Product", "Inhalant Product") ~ "INHALATION",

    dose_form_group == "Rectal Product" ~ "RECTAL",
    dose_form_group == "Vaginal Product" ~ "VAGINAL",
    dose_form_group == "Urethral Product" ~ "URETHRAL",

    dose_form_group %in% c(
      "Topical Product",
      "Transdermal Product",
      "Mucosal Product",
      "Medicated Pad or Tape",
      "Shampoo Product",
      "Soap Product"
    ) ~ "TOPICAL",

    dose_form_group == "Drug Implant Product" ~ "IMPLANT",

    stringr::str_detect(dose_form, "(?i)ophthalmic") ~ "OPHTHALMIC",
    stringr::str_detect(dose_form, "(?i)otic") ~ "OTIC",
    stringr::str_detect(dose_form, "(?i)nasal|inhal") ~ "INHALATION",
    stringr::str_detect(dose_form, "(?i)rectal|enema") ~ "RECTAL",
    stringr::str_detect(dose_form, "(?i)vaginal|douche") ~ "VAGINAL",
    stringr::str_detect(dose_form, "(?i)urethral") ~ "URETHRAL",
    stringr::str_detect(dose_form, "(?i)implant") ~ "IMPLANT",
    stringr::str_detect(dose_form, "(?i)transdermal|topical|gel|cream|ointment|lotion|patch") ~ "TOPICAL",
    stringr::str_detect(dose_form, "(?i)oral") ~ "ORAL",

    TRUE ~ NA_character_
  )
}

#' Get clinical attributes from RxCUI history status
#'
#' @keywords internal
#' @noRd
.rxref_clinical_attributes_historical <- function(rxcui,
                                                  show_progress = interactive()) {
  stopifnot(is.character(rxcui))

  rxcui <- unique(stats::na.omit(rxcui))

  empty_hist_ings <- function() {
    tibble::tibble(
      ingredient_rxcui = character(),
      ingredient_name = character(),
      ingredient_tty = character()
    )
  }

  empty_dose_forms <- function() {
    tibble::tibble(
      dose_form_rxcui = character(),
      dose_form = character()
    )
  }

  empty_dfg <- function() {
    tibble::tibble(
      dose_form_group_rxcui = character(),
      dose_form_group = character()
    )
  }

  .rxref_progress_map_dfr(
    rxcui,
    function(id) {
      hs <- tryCatch(
        rx_get_json(paste0("/rxcui/", id, "/historystatus")),
        error = function(e) NULL
      )

      hist <- rx_pluck_list(hs, "rxcuiStatusHistory")

      if (is.null(hist) || !length(hist)) {
        return(tibble::tibble(
          rxcui = id,
          related_rxcui = id,
          name = NA_character_,
          tty = NA_character_,
          strength = NA_character_,
          dose_form = NA_character_,
          route = NA_character_,
          dose_form_group = NA_character_,
          is_brand = NA,
          is_generic = NA,
          ingredient_count = NA_integer_,
          ingredient_rxcui = NA_character_,
          ingredient_name = NA_character_,
          ingredient_tty = NA_character_,
          is_multi_ingredient = NA,
          suppress = NA_character_,
          status = NA_character_,
          history_status = NA_character_,
          history_is_current = NA_character_
        ))
      }

      meta <- hist$metaData
      attrs <- hist$attributes
      defs <- hist$definitionalFeatures
      derived <- hist$derivedConcepts

      nm <- rx_scalar_chr(attrs$name)
      tty <- rx_scalar_chr(attrs$tty)
      pr <- .rxref_parse_strength_dose(nm)

      dose_forms <- as_rx_records(
        defs$doseFormConcept,
        id_field = "doseFormRxcui"
      )

      dose_forms_tbl <- if (!length(dose_forms)) {
        empty_dose_forms()
      } else {
        purrr::map_dfr(dose_forms, function(z) {
          tibble::tibble(
            dose_form_rxcui = rx_scalar_chr(z$doseFormRxcui),
            dose_form = rx_scalar_chr(z$doseFormName)
          )
        })
      }

      dfgs <- as_rx_records(
        defs$doseFormGroupConcept,
        id_field = "doseFormGroupRxcui"
      )

      dfg_tbl <- if (!length(dfgs)) {
        empty_dfg()
      } else {
        purrr::map_dfr(dfgs, function(z) {
          tibble::tibble(
            dose_form_group_rxcui = rx_scalar_chr(z$doseFormGroupRxcui),
            dose_form_group = rx_scalar_chr(z$doseFormGroupName)
          )
        })
      }

      ing_concepts <- as_rx_records(
        derived$ingredientConcept,
        id_field = "ingredientRxcui"
      )

      ingredients_tbl <- if (!length(ing_concepts)) {
        empty_hist_ings()
      } else {
        purrr::map_dfr(ing_concepts, function(z) {
          tibble::tibble(
            ingredient_rxcui = rx_scalar_chr(z$ingredientRxcui),
            ingredient_name = rx_scalar_chr(z$ingredientName),
            ingredient_tty = "IN"
          )
        })
      }

      # Fallback: ingredientAndStrength often has useful ingredient fields even
      # when derived ingredientConcept is absent.
      ing_strength <- as_rx_records(
        defs$ingredientAndStrength,
        id_field = "baseRxcui"
      )

      strength_ingredients_tbl <- if (!length(ing_strength)) {
        empty_hist_ings()
      } else {
        purrr::map_dfr(ing_strength, function(z) {
          tibble::tibble(
            ingredient_rxcui = rx_scalar_chr(z$baseRxcui),
            ingredient_name = rx_scalar_chr(z$baseName),
            ingredient_tty = "IN"
          )
        })
      }

      ingredients_tbl <- dplyr::bind_rows(
        ingredients_tbl,
        strength_ingredients_tbl
      ) |>
        dplyr::filter(
          !is.na(.data$ingredient_rxcui) | !is.na(.data$ingredient_name)
        ) |>
        dplyr::mutate(
          ingredient_key = dplyr::coalesce(
            .data$ingredient_rxcui,
            stringr::str_to_lower(.data$ingredient_name)
          )
        ) |>
        dplyr::filter(!is.na(.data$ingredient_key), nzchar(.data$ingredient_key)) |>
        dplyr::group_by(.data$ingredient_key) |>
        dplyr::summarise(
          ingredient_rxcui = dplyr::first(stats::na.omit(.data$ingredient_rxcui), default = NA_character_),
          ingredient_name  = dplyr::first(stats::na.omit(.data$ingredient_name), default = NA_character_),
          ingredient_tty   = dplyr::first(stats::na.omit(.data$ingredient_tty), default = NA_character_),
          .groups = "drop"
        )

      ingredient_count <- nrow(ingredients_tbl)

      if (ingredient_count == 0L) {
        ingredient_count <- NA_integer_
      }

      dose_form_hist <- .rxref_collapse_values(dose_forms_tbl$dose_form)
      dose_form_group_hist <- .rxref_collapse_values(dfg_tbl$dose_form_group)

      # Prefer parsed dose form from the name if available; otherwise use the
      # structured dose form concept name from historystatus.
      dose_form_final <- dplyr::coalesce(pr$dose_form, dose_form_hist)

      tibble::tibble(
        rxcui = id,
        related_rxcui = id,
        name = nm,
        tty = tty,
        strength = pr$strength,
        dose_form = dose_form_final,
        route = .rxref_route_from_dose_form(
          dose_form_final,
          dose_form_group_hist
        ),
        dose_form_group = dose_form_group_hist,
        is_brand = tty %in% c("SBD", "BPCK"),
        is_generic = tty %in% c("SCD", "GPCK"),
        ingredient_count = ingredient_count,
        ingredient_rxcui = .rxref_collapse_values(ingredients_tbl$ingredient_rxcui),
        ingredient_name = .rxref_collapse_values(ingredients_tbl$ingredient_name),
        ingredient_tty = .rxref_collapse_values(ingredients_tbl$ingredient_tty),
        is_multi_ingredient = dplyr::if_else(
          is.na(ingredient_count),
          NA,
          ingredient_count > 1L
        ),
        suppress = NA_character_,
        status = rx_scalar_chr(meta$status),
        history_status = rx_scalar_chr(meta$status),
        history_is_current = rx_scalar_chr(meta$isCurrent)
      )
    },
    name = "Getting historical clinical attributes",
    show_progress = show_progress
  ) |>
    dplyr::distinct()
}

test_that(".rxref_collapse_values() trims, drops missing values, and collapses unique values", {
  expect_equal(
    .rxref_collapse_values(c("17767", "69749")),
    "17767; 69749"
  )

  expect_equal(
    .rxref_collapse_values(c("amlodipine", "valsartan ")),
    "amlodipine; valsartan"
  )

  expect_equal(
    .rxref_collapse_values(c("IN", "IN")),
    "IN"
  )

  expect_equal(
    .rxref_collapse_values(c(" amlodipine ", "amlodipine", NA, "")),
    "amlodipine"
  )

  expect_equal(
    .rxref_collapse_values(character(0)),
    NA_character_
  )

  expect_equal(
    .rxref_collapse_values(c(NA_character_, "")),
    NA_character_
  )
})


mock_hist_status_two_ingredients <- function() {
  list(
    rxcuiStatusHistory = list(
      metaData = list(
        status = "Obsolete",
        isCurrent = "NO"
      ),
      attributes = list(
        rxcui = "999999",
        name = "amlodipine 5 MG / valsartan 160 MG Oral Tablet",
        tty = "SCD",
        isMultipleIngredient = "YES"
      ),
      definitionalFeatures = list(
        doseFormConcept = list(
          list(
            doseFormRxcui = "317541",
            doseFormName = "Oral Tablet"
          )
        ),
        doseFormGroupConcept = list(
          list(
            doseFormGroupRxcui = "1151131",
            doseFormGroupName = "Oral Product"
          )
        ),
        ingredientAndStrength = list(
          list(
            baseRxcui = "17767",
            baseName = "amlodipine"
          ),
          list(
            baseRxcui = "69749",
            baseName = "valsartan"
          )
        )
      ),
      derivedConcepts = list(
        ingredientConcept = list(
          # Deliberately duplicated with ingredientAndStrength above.
          # This protects the de-duplication logic.
          list(
            ingredientRxcui = "17767",
            ingredientName = "amlodipine"
          )
        )
      )
    )
  )
}


mock_hist_status_no_ingredients <- function() {
  list(
    rxcuiStatusHistory = list(
      metaData = list(
        status = "Obsolete",
        isCurrent = "NO"
      ),
      attributes = list(
        rxcui = "888888",
        name = "obsolete product with unavailable ingredients",
        tty = "SCD",
        isMultipleIngredient = NA_character_
      ),
      definitionalFeatures = list(
        doseFormConcept = list(),
        doseFormGroupConcept = list(),
        ingredientAndStrength = list()
      ),
      derivedConcepts = list(
        ingredientConcept = list()
      )
    )
  )
}


test_that("historical clinical attributes combine and deduplicate ingredient sources", {
  testthat::local_mocked_bindings(
    rx_get_json = function(path, query = NULL) {
      expect_equal(path, "/rxcui/999999/historystatus")
      mock_hist_status_two_ingredients()
    }
  )

  out <- .rxref_clinical_attributes_historical(
    "999999",
    show_progress = FALSE
  )

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1L)

  expect_equal(out$rxcui, "999999")
  expect_equal(out$related_rxcui, "999999")
  expect_equal(out$tty, "SCD")

  expect_equal(out$ingredient_count, 2L)
  expect_equal(out$ingredient_rxcui, "17767; 69749")
  expect_equal(out$ingredient_name, "amlodipine; valsartan")
  expect_equal(out$ingredient_tty, "IN")
  expect_true(out$is_multi_ingredient)

  expect_equal(out$dose_form, "Oral Tablet")
  expect_equal(out$dose_form_group, "Oral Product")
  expect_equal(out$route, "ORAL")
})


test_that("historical clinical attributes return NA ingredient count when no ingredients are found", {
  testthat::local_mocked_bindings(
    rx_get_json = function(path, query = NULL) {
      expect_equal(path, "/rxcui/888888/historystatus")
      mock_hist_status_no_ingredients()
    }
  )

  out <- .rxref_clinical_attributes_historical(
    "888888",
    show_progress = FALSE
  )

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1L)

  expect_equal(out$rxcui, "888888")
  expect_equal(out$related_rxcui, "888888")

  expect_true(is.na(out$ingredient_count))
  expect_true(is.na(out$ingredient_rxcui))
  expect_true(is.na(out$ingredient_name))
  expect_true(is.na(out$ingredient_tty))
  expect_true(is.na(out$is_multi_ingredient))
})


test_that("missing active ingredient count triggers historical fallback", {
  testthat::local_mocked_bindings(
    rx_get_json = function(path, query = NULL) {
      if (identical(path, "/rxcui/999999/properties")) {
        return(list(
          properties = list(
            rxcui = "999999",
            name = "amlodipine 5 MG / valsartan 160 MG Oral Tablet",
            tty = "SCD",
            suppress = "N"
          )
        ))
      }

      if (
        identical(path, "/rxcui/999999/related") &&
        identical(query, list(tty = "DFG"))
      ) {
        return(list(
          relatedGroup = list(
            conceptGroup = list()
          )
        ))
      }

      if (identical(path, "/rxcui/999999/historystatus")) {
        return(mock_hist_status_two_ingredients())
      }

      stop("Unexpected mocked rx_get_json() call: ", path)
    },
    .rxref_get_ingredients_for_rxcui = function(rxcui,
                                                include_pin = TRUE,
                                                include_min = FALSE) {
      tibble::tibble(
        related_rxcui = character(),
        ingredient_rxcui = character(),
        ingredient_name = character(),
        ingredient_tty = character()
      )
    }
  )

  out <- get_clinical_attributes(
    "999999",
    include_historical = TRUE,
    show_progress = FALSE
  )

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1L)

  expect_equal(out$rxcui, "999999")
  expect_equal(out$related_rxcui, "999999")

  expect_equal(out$ingredient_count, 2L)
  expect_equal(out$ingredient_rxcui, "17767; 69749")
  expect_equal(out$ingredient_name, "amlodipine; valsartan")
  expect_equal(out$ingredient_tty, "IN")
  expect_true(out$is_multi_ingredient)

  expect_equal(out$dose_form_group, "Oral Product")
  expect_equal(out$route, "ORAL")
  expect_equal(out$history_status, "Obsolete")
  expect_equal(out$history_is_current, "NO")
})

.onLoad <- function(libname, pkgname) {
  op <- options()

  op.rxref <- list(
    rxref.base_url = "https://rxnav.nlm.nih.gov/REST",
    rxref.rxclass_base_url = "https://rxnav.nlm.nih.gov/REST/rxclass",
    rxref.user_agent = sprintf(
      "rxref/%s (+https://github.com/ssmithm/rxref)",
      utils::packageVersion("rxref")
    ),
    rxref.cache = cachem::cache_mem(),
    rxref.rate_delay = 0.1
  )

  toset <- !(names(op.rxref) %in% names(op))
  if (any(toset)) options(op.rxref[toset])

  invisible()
}

#' Resolve free text, RxCUI, or NDC to RxCUI and preferred name
#'
#' Vectorized over `x`. For free text, uses RxNorm approximateTerm.
#' For NDC, uses findRxcuiById. For RxCUI, validates and returns properties.
#'
#' @param x Character vector: drug string, RxCUI, or NDC (10/11-digit or hyphenated)
#' @param type One of c("auto","name","rxcui","ndc"). Default "auto" infers.
#' @param max_entries Integer, passed to approximateTerm for name queries.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble with columns: input, type, rxcui, name, tty, score (if name),
#' ndc11 (if ndc input), matched_term (if name input)
#' @export
resolve <- function(
    x,
    type = c("auto", "name", "rxcui", "ndc"),
    max_entries = 1,
    show_progress = interactive()
) {
  stopifnot(is.character(x))
  type <- match.arg(type)

  .rxref_progress_map_dfr(
    x,
    function(val) {
      t <- switch(
        type,
        auto = {
          if (is_ndcish(val)) {
            "ndc"
          } else if (is_rxcui(val)) {
            "rxcui"
          } else {
            "name"
          }
        },
        name = "name",
        rxcui = "rxcui",
        ndc = "ndc"
      )

      switch(
        t,
        name = resolve_name(val, max_entries = max_entries),
        ndc = resolve_ndc(val),
        rxcui = resolve_rxcui(val)
      )
    },
    name = "Resolving inputs",
    show_progress = show_progress
  )
}

#' @keywords internal
#' @noRd
resolve_name <- function(term, max_entries = 1) {
  # spelling suggestions (optional)
  # NOTE: this is mostly a hint; we don't use it downstream here
  #       but may be useful later for UX / alternative candidates?
  try({
    rx_get_json("/spellingsuggestions", query = list(name = term))
  }, silent = TRUE)

  # approximate matching (primary)
  approx <- rx_get_json("/approximateTerm", query = list(
    term = term,
    maxEntries = max_entries
  ))

  cand <- rx_pluck_list(approx, "approximateGroup", "candidate")

  if (!length(cand)) {
    return(tibble::tibble(
      input = term,
      type = "name",
      rxcui = NA_character_,
      name = NA_character_,
      tty = NA_character_,
      score = NA_real_,
      matched_term = NA_character_
    ))
  }

  cand <- cand[seq_len(min(length(cand), max_entries))]
  rows <- purrr::map(cand, function(c) {
    rxcui <- rx_scalar_chr(rx_pluck_chr(c, "rxcui"))

    props <- if (!is.na(rxcui)) {
      rx_get_json(paste0("/rxcui/", rxcui, "/properties"))
    } else {
      NULL
    }

    tibble::tibble(
      input = term,
      type = "name",
      rxcui = rxcui,
      name = rx_scalar_chr(rx_pluck_chr(props, "properties", "name")),
      tty = rx_scalar_chr(rx_pluck_chr(props, "properties", "tty")),
      score = suppressWarnings(as.numeric(rx_scalar_chr(rx_pluck_chr(c, "score")))),
      matched_term = rx_scalar_chr(rx_pluck_chr(c, "name"))
    )
  })

  dplyr::bind_rows(rows)
}

#' @keywords internal
#' @noRd
resolve_ndc <- function(ndc) {
  ndc_norm <- ndc_to_11(ndc)

  if (is.na(ndc_norm) || !nzchar(ndc_norm)) {
    return(
      tibble::tibble(
        input = ndc,
        type = "ndc",
        ndc11 = NA_character_,
        rxcui = NA_character_,
        name = NA_character_,
        tty = NA_character_,
        score = NA_real_,
        matched_term = NA_character_
      )
    )
  }

  res <- rx_get_json("/rxcui", query = list(idtype = "NDC", id = ndc_norm))

  rx <- rx_pluck_chr(res, "idGroup", "rxnormId")
  rxcui <- rx_scalar_chr(rx)

  props <- if (!is.na(rxcui)) {
    rx_get_json(paste0("/rxcui/", rxcui, "/properties"))
  } else {
    NULL
  }

  tibble::tibble(
    input = ndc,
    type = "ndc",
    ndc11 = ndc_norm,
    rxcui = rxcui,
    name = rx_scalar_chr(rx_pluck_chr(props, "properties", "name")),
    tty  = rx_scalar_chr(rx_pluck_chr(props, "properties", "tty")),
    score = NA_real_,
    matched_term = NA_character_
  )
}

#' @keywords internal
#' @noRd
resolve_rxcui <- function(rxcui) {
  props <- rx_get_json(paste0("/rxcui/", rxcui, "/properties"))

  tibble::tibble(
    input = rxcui,
    type = "rxcui",
    rxcui = rxcui,
    name = rx_scalar_chr(rx_pluck_chr(props, "properties", "name")),
    tty = rx_scalar_chr(rx_pluck_chr(props, "properties", "tty")),
    score = NA_real_,
    matched_term = NA_character_
  )
}

#' Map between NDCs and RxCUIs
#'
#' `map_to()` maps identifiers between National Drug Codes (NDCs) and RxNorm
#' Concept Unique Identifiers (RxCUIs).
#'
#' When mapping from NDC to RxCUI, NDCs may be supplied as 10-digit,
#' 11-digit, or hyphenated values. Input NDCs are normalized to 11-digit
#' format before querying RxNorm.
#'
#' When mapping from RxCUI to NDC, the `history` argument controls whether
#' only currently active NDCs are retrieved or whether historical NDC
#' associations are also included.
#'
#' @param x Character vector of NDCs or RxCUIs.
#' @param to Direction of mapping. One of `"rxcui"` or `"ndc"`.
#'   Use `"rxcui"` to map NDCs to RxCUIs and `"ndc"` to map RxCUIs to NDCs.
#' @param status For `to = "ndc"` only, optional character vector of NDC
#'   statuses to retain, such as `"ACTIVE"`, `"OBSOLETE"`, `"ALIEN"`, or
#'   `"UNKNOWN"`. If `NULL`, no status-based filtering is applied to the
#'   retrieved NDCs. Note that when `history = "active"`, RxNorm's active NDC
#'   endpoint is used, so returned NDCs are expected to be active even when
#'   `status = NULL`.
#' @param history For `to = "ndc"` only, the NDC association history to
#'   retrieve. One of `"active"`, `"direct"`, or `"all"`.
#'
#'   - `"active"` retrieves currently active NDCs associated with the supplied
#'     RxCUI.
#'   - `"direct"` retrieves NDCs ever directly associated with the supplied
#'     RxCUI.
#'   - `"all"` retrieves NDCs ever directly or indirectly associated with the
#'     supplied RxCUI, including associations through remapped or archived
#'     concepts.
#'
#'   If `NULL` and `to = "ndc"`, defaults to `"active"`. Ignored when
#'   `to = "rxcui"`.
#' @param show_progress Logical; if `TRUE`, show a progress bar for vectorized
#'   mapping operations. Defaults to `interactive()`.
#'
#' @return A tibble.
#'
#' For `to = "rxcui"`, returns one row per input NDC/RxCUI mapping with columns:
#' \describe{
#'   \item{input}{Original input NDC.}
#'   \item{ndc11}{Normalized 11-digit NDC.}
#'   \item{rxcui}{Mapped RxCUI, if found.}
#' }
#'
#' For `to = "ndc"` and `history = "active"`, returns one row per RxCUI/NDC
#' mapping with columns:
#' \describe{
#'   \item{rxcui}{Input RxCUI.}
#'   \item{ndc11}{Mapped NDC.}
#'   \item{ndc_status}{NDC status returned by RxNorm, when available.}
#' }
#'
#' For `to = "ndc"` and `history` equal to `"direct"` or `"all"`, the output
#' may also include:
#' \describe{
#'   \item{related_rxcui}{RxCUI associated with the historical NDC record. This
#'     may differ from the input RxCUI when historical or indirect associations
#'     are retrieved.}
#'   \item{ndc_start_date}{Start date of the NDC association, when available.}
#'   \item{ndc_end_date}{End date of the NDC association, when available.}
#' }
#'
#' @details
#' RxNorm distinguishes between currently active NDCs and historical NDC
#' associations. The default behavior, `history = "active"`, uses RxNorm's
#' active NDC endpoint and preserves the earlier behavior of `map_to()`.
#'
#' To retrieve obsolete, discontinued, or otherwise historical NDCs, use
#' `history = "direct"` or `history = "all"`. Setting `status = NULL` does not
#' by itself request historical NDCs; it only means that no status filter is
#' applied after NDCs are retrieved.
#'
#' @examples
#' \dontrun{
#' # Map an NDC to RxCUI
#' map_to("00093-7424-56", to = "rxcui")
#'
#' # Map an RxCUI to currently active NDCs
#' map_to("1049630", to = "ndc")
#'
#' # Map an RxCUI to all directly associated historical NDCs
#' map_to("1049630", to = "ndc", history = "direct")
#'
#' # Map an RxCUI to all historical NDCs and retain obsolete NDCs only
#' map_to("1049630", to = "ndc", history = "all", status = "OBSOLETE")
#' }
#'
#' @export
map_to <- function(
    x,
    to = c("rxcui", "ndc"),
    status = NULL,
    history = NULL,
    show_progress = interactive()
) {
  to <- match.arg(to)

  stopifnot(is.character(x))

  if (to == "ndc") {
    if (is.null(history)) {
      history <- "active"
    }

    history <- match.arg(history, c("active", "direct", "all"))
  } else {
    if (!is.null(history)) {
      cli::cli_warn("{.arg history} is only used when {.code to = \"ndc\"}; it will be ignored.")
    }

    if (!is.null(status)) {
      cli::cli_warn("{.arg status} is only used when {.code to = \"ndc\"}; it will be ignored.")
    }
  }

  progress_name <- if (to == "rxcui") {
    "Mapping NDCs to RxCUIs"
  } else {
    "Mapping RxCUIs to NDCs"
  }

  if (to == "rxcui") {
    .rxref_progress_map_dfr(
      x,
      function(ndc) {
        ndc11 <- ndc_to_11(ndc)

        if (is.na(ndc11) || !nzchar(ndc11)) {
          return(
            tibble::tibble(
              input = ndc,
              ndc11 = NA_character_,
              rxcui = NA_character_
            )
          )
        }

        res <- rx_get_json(
          "/rxcui",
          query = list(idtype = "NDC", id = ndc11)
        )

        rxs <- rx_pluck_chr(res, "idGroup", "rxnormId")

        rxs <- if (length(rxs)) {
          unlist(rxs, use.names = FALSE)
        } else {
          NA_character_
        }

        tibble::tibble(
          input = ndc,
          ndc11 = ndc11,
          rxcui = as.character(rxs)
        )
      },
      name = progress_name,
      show_progress = show_progress
    ) |>
      dplyr::distinct()
  } else {
    .rxref_progress_map_dfr(
      x,
      function(id) {
        if (is.na(id) || !nzchar(id)) {
          return(tibble::tibble(
            rxcui = id,
            ndc11 = NA_character_,
            ndc_status = NA_character_
          ))
        }

        if (history == "active") {
          res <- rx_get_json(paste0("/rxcui/", id, "/ndcs"))
          ndcs <- rx_pluck_chr(res, "ndcGroup", "ndcList", "ndc")

          ndcs <- if (length(ndcs)) {
            unlist(ndcs, use.names = FALSE)
          } else {
            NA_character_
          }

          out <- tibble::tibble(
            rxcui = id,
            ndc11 = ndcs
          )
        } else {
          history_value <- switch(
            history,
            direct = 1,
            all = 2
          )

          res <- rx_get_json(
            paste0("/rxcui/", id, "/allhistoricalndcs"),
            query = list(history = history_value)
          )

          hist <- rx_pluck_list(res, "historicalNdcConcept", "historicalNdcTime")

          if (is.null(hist) || !length(hist)) {
            out <- tibble::tibble(
              rxcui = id,
              related_rxcui = NA_character_,
              ndc11 = NA_character_,
              ndc_start_date = NA_character_,
              ndc_end_date = NA_character_
            )
          } else {
            out <- purrr::map_dfr(hist, function(h) {
              ndc_time <- rx_pluck_list(h, "ndcTime")

              if (is.null(ndc_time) || !length(ndc_time)) {
                return(tibble::tibble(
                  rxcui = id,
                  related_rxcui = rx_scalar_chr(rx_pluck_chr(h, "rxcui")),
                  ndc11 = NA_character_,
                  ndc_start_date = NA_character_,
                  ndc_end_date = NA_character_
                ))
              }

              purrr::map_dfr(ndc_time, function(nt) {
                ndcs <- rx_pluck_chr(nt, "ndc")

                ndcs <- if (length(ndcs)) {
                  unlist(ndcs, use.names = FALSE)
                } else {
                  NA_character_
                }

                tibble::tibble(
                  rxcui = id,
                  related_rxcui = rx_scalar_chr(rx_pluck_chr(h, "rxcui")),
                  ndc11 = as.character(ndcs),
                  ndc_start_date = rx_scalar_chr(rx_pluck_chr(nt, "startDate")),
                  ndc_end_date = rx_scalar_chr(rx_pluck_chr(nt, "endDate"))
                )
              })
            })
          }

          ndcs <- out$ndc11
        }

        stats <- purrr::map_chr(ndcs, function(n) {
          if (is.na(n) || !nzchar(n)) {
            return(NA_character_)
          }

          # hyphenate to 5-4-2 for the status endpoint
          ndc_h <- hyphenate_ndc_5_4_2(n)

          st <- tryCatch(
            rx_get_json("/ndcstatus", query = list(ndc = ndc_h)),
            error = function(e) NULL
          )

          # common JSON shape: ndcStatus$status
          status_value <- rx_scalar_chr(rx_pluck_chr(st, "ndcStatus", "status"))

          if (is.na(status_value) || !nzchar(status_value)) {
            return(NA_character_)
          }

          status_value
        })

        out$ndc_status <- stats

        if (!is.null(status)) {
          keep <- tolower(out$ndc_status) %in% tolower(status)
          out <- out[keep, , drop = FALSE]
        }

        out
      },
      name = progress_name,
      show_progress = show_progress
    ) |>
      dplyr::distinct()
  }
}


#' Map NDCs to RxCUIs
#'
#' Convenience wrapper around [map_to()] for mapping National Drug Codes
#' (NDCs) to RxNorm Concept Unique Identifiers (RxCUIs).
#'
#' NDCs may be supplied as 10-digit, 11-digit, or hyphenated values. Input
#' NDCs are normalized to 11-digit format before querying RxNorm.
#'
#' @param x Character vector of NDCs.
#' @param show_progress Logical; if `TRUE`, show a progress bar for vectorized
#'   mapping operations. Defaults to `interactive()`.
#'
#' @return A tibble with columns:
#' \describe{
#'   \item{input}{Original input NDC.}
#'   \item{ndc11}{Normalized 11-digit NDC.}
#'   \item{rxcui}{Mapped RxCUI, if found.}
#' }
#'
#' @examples
#' \dontrun{
#' map_ndc_to_rxcui("00093-7424-56")
#' }
#'
#' @export
map_ndc_to_rxcui <- function(x, show_progress = interactive()) {
  map_to(x, to = "rxcui", show_progress = show_progress)
}


#' Map RxCUIs to NDCs
#'
#' Convenience wrapper around [map_to()] for mapping RxNorm Concept Unique
#' Identifiers (RxCUIs) to National Drug Codes (NDCs).
#'
#' By default, this function retrieves currently active NDCs associated with
#' the supplied RxCUIs. Use `history = "direct"` or `history = "all"` to
#' retrieve historical NDC associations.
#'
#' @param x Character vector of RxCUIs.
#' @param status Optional character vector of NDC statuses to retain, such as
#'   `"ACTIVE"`, `"OBSOLETE"`, `"ALIEN"`, or `"UNKNOWN"`. If `NULL`, no
#'   status-based filtering is applied to the retrieved NDCs. Note that
#'   `history = "active"` uses RxNorm's active NDC endpoint, so returned NDCs
#'   are expected to be active even when `status = NULL`.
#' @param history For `to = "ndc"` only, the NDC association history to
#'   retrieve. One of `"active"`, `"direct"`, or `"all"`.
#'
#'   - `"active"` retrieves currently active NDCs associated with the supplied
#'     RxCUI.
#'   - `"direct"` retrieves NDCs ever directly associated with the supplied
#'     RxCUI.
#'   - `"all"` retrieves NDCs ever directly or indirectly associated with the
#'     supplied RxCUI, including associations through remapped or archived
#'     concepts.
#'
#'   If `NULL` and `to = "ndc"`, defaults to `"active"`. Ignored when
#'   `to = "rxcui"`.
#' @param show_progress Logical; if `TRUE`, show a progress bar for vectorized
#'   mapping operations. Defaults to `interactive()`.
#'
#' @return A tibble with one row per RxCUI/NDC mapping.
#'
#' For `history = "active"`, returns columns:
#' \describe{
#'   \item{rxcui}{Input RxCUI.}
#'   \item{ndc11}{Mapped NDC.}
#'   \item{ndc_status}{NDC status returned by RxNorm, when available.}
#' }
#'
#' For `history = "direct"` or `"all"`, the output may also include:
#' \describe{
#'   \item{related_rxcui}{RxCUI associated with the historical NDC record. This
#'     may differ from the input RxCUI when historical or indirect associations
#'     are retrieved.}
#'   \item{ndc_start_date}{Start date of the NDC association, when available.}
#'   \item{ndc_end_date}{End date of the NDC association, when available.}
#' }
#'
#' @details
#' `status = NULL` means that no status filter is applied after NDCs are
#' retrieved. It does not, by itself, request historical NDCs. To retrieve
#' historical NDCs, set `history = "direct"` or `history = "all"`.
#'
#' @examples
#' \dontrun{
#' # Current active NDCs
#' map_rxcui_to_ndc("1049630")
#'
#' # NDCs ever directly associated with this RxCUI
#' map_rxcui_to_ndc("1049630", history = "direct")
#'
#' # All historical NDCs, retaining obsolete NDCs only
#' map_rxcui_to_ndc(
#'   "1049630",
#'   history = "all",
#'   status = "OBSOLETE"
#' )
#' }
#'
#' @export
map_rxcui_to_ndc <- function(
    x,
    status = NULL,
    history = c("active", "direct", "all"),
    show_progress = interactive()
) {
  history <- match.arg(history)

  map_to(
    x,
    to = "ndc",
    status = status,
    history = history,
    show_progress = show_progress
  )
}

#' Safely pluck an object
#'
#' @keywords internal
#' @noRd
rx_pluck <- function(x, ..., .default = NULL) {
  purrr::pluck(x, ..., .default = .default)
}


#' Safely pluck a character vector
#'
#' @keywords internal
#' @noRd
rx_pluck_chr <- function(x, ...) {
  out <- purrr::pluck(x, ..., .default = character())

  if (is.null(out)) {
    character()
  } else {
    as.character(out)
  }
}


#' Safely pluck a list
#'
#' @keywords internal
#' @noRd
rx_pluck_list <- function(x, ...) {
  out <- purrr::pluck(x, ..., .default = list())

  if (is.null(out)) {
    list()
  } else if (is.list(out)) {
    out
  } else {
    list(out)
  }
}


#' Return first character value, or NA
#'
#' @keywords internal
#' @noRd
rx_scalar_chr <- function(x) {
  x <- as.character(x)

  if (length(x) == 0L) {
    NA_character_
  } else {
    x[[1]]
  }
}

#' Catalogue of RxNorm TTY (Term Types)
#'
#' Returns a tibble describing common RxNorm TTYs you may want to use when
#' expanding ingredients to products. Includes whether each TTY typically maps
#' cleanly to NDCs and whether it's included in the package's default or
#' extended TTY sets.
#'
#' Columns:
#' - `tty`: RxNorm term type code.
#' - `label`: Short, human-friendly name.
#' - `description`: What the TTY represents in RxNorm.
#' - `maps_to_ndc`: Logical; whether CUIs of this TTY usually map to NDCs via
#'    `/rxcui/{rxcui}/ndcs`.
#' - `typical_role`: How it’s commonly used (product, component, group, name).
#' - `include_default`: Logical; included in `.rxref_default_ttys`.
#' - `include_extended`: Logical; included in `.rxref_extended_ttys`.
#'
#' @return A tibble with metadata for key TTYs.
#' @examples
#' tty_catalogue()
#'
#' # TTYs that map to NDCs
#' subset(tty_catalogue(), maps_to_ndc)$tty
#'
#' # See what your defaults and extended sets contain
#' subset(tty_catalogue(), include_default)$tty
#' subset(tty_catalogue(), include_extended)$tty
#'
#' # Pick a custom set: products + components
#' with(tty_catalogue(),
#'      tty[tty %in% c("SCD","SBD","GPCK","BPCK","SCDC","SBDC")])
#' @export
tty_catalogue <- function() {
  # guard in case user hasn't defined these constants yet
  default_set  <- if (exists(".rxref_default_ttys", inherits = TRUE))
    get(".rxref_default_ttys", inherits = TRUE) else c("SCD","SBD","GPCK","BPCK")
  extended_set <- if (exists(".rxref_extended_ttys", inherits = TRUE))
    get(".rxref_extended_ttys", inherits = TRUE)
  else c("SCDC","SBDC","SCDF","SBDF","SCDFP","SBDFP","SCDG","SCDGP","BN","MIN")

  df <- tibble::tribble(
    ~tty,    ~label,                  ~description,                                                                                 ~maps_to_ndc, ~typical_role,
    "SCD",   "Semantic Clinical Drug","Normalized clinical product (ingredient+strength+dose form), no brand.",                     TRUE,         "product",
    "SBD",   "Semantic Branded Drug", "Normalized branded product (brand + SCD composition).",                                      TRUE,         "product",
    "GPCK",  "Generic Pack",          "Package that contains one or more SCD items.",                                               TRUE,         "pack",
    "BPCK",  "Branded Pack",          "Package that contains one or more SBD items.",                                               TRUE,         "pack",

    "SCDC",  "SCD Component",         "Component of an SCD (e.g., within a combo).",                                               FALSE,        "component",
    "SBDC",  "SBD Component",         "Component of an SBD.",                                                                       FALSE,        "component",

    "SCDF",  "SCD + Dose Form",       "Group of SCDs sharing a dose form; structural grouping.",                                   FALSE,        "group",
    "SBDF",  "SBD + Dose Form",       "Group of SBDs sharing a dose form; structural grouping.",                                   FALSE,        "group",
    "SCDFP", "SCD + DF + Pack",       "Group of SCDs sharing dose form and pack; structural grouping.",                            FALSE,        "group",
    "SBDFP", "SBD + DF + Pack",       "Group of SBDs sharing dose form and pack; structural grouping.",                            FALSE,        "group",
    "SCDG",  "SCD Group",             "Higher-level group of related SCDs.",                                                        FALSE,        "group",
    "SCDGP", "SCD Group + Pack",      "Group of SCDs by pack status.",                                                             FALSE,        "group",

    "BN",    "Brand Name",            "Brand name term (name concept, not a product instance).",                                   FALSE,        "name",
    "MIN",   "Multi-Ingredient Name", "Name concept representing a specific multi-ingredient combo (not a product instance).",      FALSE,        "name"
  )

  df |>
    dplyr::mutate(
      include_default  = .data$tty %in% default_set,
      include_extended = .data$tty %in% extended_set
    )
}


# Some helpers:
# Default product-ish TTYs (good for mapping to NDC)
.rxref_default_ttys <- c(
  "SCD",  # Semantic Clinical Drug
  "SBD",  # Semantic Branded Drug
  "GPCK", # Generic Pack
  "BPCK"  # Brand Pack
)

.rxref_extended_product_ttys <- c(
  .rxref_default_ttys,
  "SCDG",  # Semantic Clinical Dose Form Group
  "SBDG",  # Semantic Branded Dose Form Group
  "SCDF",  # Semantic Clinical Dose Form
  "SBDF",  # Semantic Branded Dose Form
  "SBDFP", # Semantic Branded Drug Form Precise
  "SCDFP", # Semantic Clinical Drug Form Precise
  "SCDGP" # Semantic Clinical Dose Form Group Precise
)

# Extended “structure/group” TTYs for richer CUIs; note many of these will rarely map to NDC,
# but may be useful in RXCUI searches, particularly where RXCUIs are not cleanly mapped to a
# very limited set of prescribable products.
# - Default product TTYs: SCD, SBD, GPCK, BPCK
# - Components: SCDC, SBDC
# - Drug forms / groups: SCDF, SBDF, SCDFP, SBDFP, SCDG, SCDGP
# - Names / ingredients: BN (brand name), MIN (multi-ingredient), IN (ingredients)
.rxref_extended_ttys <- c(
  .rxref_extended_product_ttys,
  "SCDC",  # Semantic Clinical Drug Component
  "SBDC",  # Semantic Branded Drug Component
  "BN",    # Brand Name
  "MIN",   # Multiple Ingredients
  "PIN",   # Precise Ingredient
  "IN"     # Ingredient
  )

#' Default RxNorm product term types
#'
#' Returns the default RxNorm term types used by rxref when identifying
#' drug products.
#'
#' The default set is intentionally focused on product-level concepts:
#' semantic clinical drugs, semantic branded drugs, generic packs, and
#' branded packs.
#'
#' @return A character vector of RxNorm term type abbreviations.
#'
#' @examples
#' default_product_ttys()
#'
#' @export
default_product_ttys <- function() {
  .rxref_default_ttys
}


#' Extended RxNorm product term types
#'
#' Returns an extended set of RxNorm term types that includes the default
#' product term types plus dose-form and dose-form-group concepts.
#'
#' This can be useful when a broader set of product-related RxNorm concepts
#' is desired.
#'
#' @return A character vector of RxNorm term type abbreviations.
#'
#' @examples
#' extended_product_ttys()
#'
#' @export
extended_product_ttys <- function() {
  .rxref_extended_product_ttys
}

#' Extended RxNorm term types
#'
#' Returns an extended set of RxNorm term types that includes the default and extended
#' product term types plus drug component, brand name, and ingredient concepts.
#' This list includes essentially all term types that capture a specific ingredient,
#' thus it excludes dose form, dose form group, prescribable name, synonyms, and others
#' that are not associated with a specific ingredient.
#'
#' This can be useful when the broadest set of RxNorm concepts that still capture an ingredient
#' is desired.
#'
#' @return A character vector of RxNorm term type abbreviations.
#'
#' @examples
#' extended_ttys()
#'
#' @export
extended_ttys <- function() {
  .rxref_extended_ttys
}

#' RxNorm product term type sets
#'
#' Returns common RxNorm term type sets used by rxref.
#'
#' @param set One of `"default"`, `"extended_product"`, or `"extended"`.
#'
#' @return A character vector of RxNorm term type abbreviations.
#'
#' @examples
#' product_ttys()
#' product_ttys("extended")
#'
#' @export
product_ttys <- function(set = c("default", "extended_product", "extended")) {
  set <- match.arg(set)

  switch(
    set,
    default = default_product_ttys(),
    extended_product = extended_product_ttys(),
    extended = extended_ttys()
  )
}

test_that("products_for_ingredients active only excludes historical products", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))

  out <- products_for_ingredients(
    "1369",
    ttys = "SCD",
    route = NULL,
    include_combos = FALSE,
    concept_status = "active"
  )

  expect_false("concept_status" %in% names(out))
  expect_false("active_start_date" %in% names(out))
  expect_false("active_end_date" %in% names(out))
  expect_false("release_start_date" %in% names(out))
  expect_false("release_end_date" %in% names(out))

  expect_equal(nrow(out), 2)
  expect_setequal(out$product_rxcui, c("197394", "308614"))
  expect_true(all(out$tty == "SCD"))
  expect_true(all(out$n_ingredients == 1L))
})

test_that("products_for_ingredients active_and_historical includes obsolete SCD products", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))
  skip_if_not(identical(Sys.getenv("RXREF_SLOW"), "1"))

  out <- products_for_ingredients(
    "1369",
    ttys = "SCD",
    route = NULL,
    include_combos = FALSE,
    concept_status = "active_and_historical"
  )

  expect_true("concept_status" %in% names(out))
  expect_true("active_start_date" %in% names(out))
  expect_true("active_end_date" %in% names(out))
  expect_true("release_start_date" %in% names(out))
  expect_true("release_end_date" %in% names(out))

  expect_setequal(
    out$product_rxcui,
    c("197394", "308614", "197393", "308617", "429502")
  )

  expect_setequal(out$concept_status, c("Active", "Obsolete"))
  expect_true(all(out$tty == "SCD"))
  expect_true(all(out$n_ingredients == 1L))
})

test_that("historical combination products have correct ingredient counts", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))
  skip_if_not(identical(Sys.getenv("RXREF_SLOW"), "1"))

  out <- products_for_ingredients(
    "1369",
    ttys = product_ttys("extended_product"),
    route = NULL,
    include_combos = TRUE,
    concept_status = "active_and_historical"
  )

  x <- dplyr::filter(out, product_rxcui == "104267")

  expect_equal(nrow(x), 1)
  expect_equal(x$n_ingredients, 2L)
  expect_equal(x$concept_status, "Obsolete")
})

test_that("products_for_ingredients validates concept_status", {
  expect_error(
    products_for_ingredients(
      "1369",
      concept_status = "current"
    ),
    "should be one of"
  )
})

test_that("historical_status can limit returned historical statuses", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))
  skip_if_not(identical(Sys.getenv("RXREF_SLOW"), "1"))

  out <- products_for_ingredients(
    "1369",
    ttys = "SCD",
    route = NULL,
    include_combos = FALSE,
    concept_status = "active_and_historical",
    historical_status = "Obsolete"
  )

  expect_true(all(out$concept_status %in% c("Active", "Obsolete")))
})

test_that("products_for_ingredients rejects deprecated concept_status values", {
  expect_error(
    products_for_ingredients(
      "11289",
      concept_status = "current_and_historical"
    ),
    "active"
  )
})

test_that("default_product_ttys returns expected product TTYs", {
  expect_type(default_product_ttys(), "character")
  expect_equal(
    default_product_ttys(),
    c("SCD", "SBD", "GPCK", "BPCK")
  )
})

test_that("extended_product_ttys includes default product TTYs", {
  expect_type(extended_product_ttys(), "character")
  expect_true(all(default_product_ttys() %in% extended_product_ttys()))
  expect_true(length(extended_product_ttys()) > length(default_product_ttys()))
})

test_that("product_ttys returns requested TTY set", {
  expect_equal(product_ttys("default"), default_product_ttys())
  expect_equal(product_ttys("extended_product"), extended_product_ttys())
  expect_equal(product_ttys("extended"), extended_ttys())
})

#' Abort with an rxref API error
#'
#' @keywords internal
#' @noRd
rx_abort_api <- function(
    message,
    class = NULL,
    parent = NULL,
    call = rlang::caller_env()
) {
  rlang::abort(
    message = message,
    class = c(class, "rxref_api_error"),
    parent = parent,
    call = call
  )
}

test_that("rx_abort_api creates rxref API errors", {
  expect_error(
    rx_abort_api("Test error.", class = "rxref_test_error"),
    class = "rxref_test_error"
  )

  expect_error(
    rx_abort_api("Test error.", class = "rxref_test_error"),
    class = "rxref_api_error"
  )
})

test_that("rx_perform_json() gives a friendly error for 429 responses", {
  req <- httr2::request("https://example.com")

  local_mocked_bindings(
    req_perform = function(req, ...) {
      httr2::response(status_code = 429)
    },
    .package = "httr2"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    class = "rxref_rate_limit_error"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    regexp = "rate-limited"
  )
})

test_that("rx_perform_json() gives a friendly error for 404 responses", {
  req <- httr2::request("https://example.com")

  local_mocked_bindings(
    req_perform = function(req, ...) {
      httr2::response(status_code = 404)
    },
    .package = "httr2"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    class = "rxref_not_found_error"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    regexp = "status 404"
  )
})

test_that("rx_perform_json() gives a friendly error for 500 responses", {
  req <- httr2::request("https://example.com")

  local_mocked_bindings(
    req_perform = function(req, ...) {
      httr2::response(status_code = 500)
    },
    .package = "httr2"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    class = "rxref_server_error"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    regexp = "temporary server-side issue"
  )
})

test_that("rx_perform_json() gives a friendly error for other 400 responses", {
  req <- httr2::request("https://example.com")

  local_mocked_bindings(
    req_perform = function(req, ...) {
      httr2::response(status_code = 400)
    },
    .package = "httr2"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    class = "rxref_http_error"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    regexp = "unsuccessful response"
  )
})

test_that("rx_perform_json() gives a friendly error for connection failures", {
  req <- httr2::request("https://example.com")

  local_mocked_bindings(
    req_perform = function(req, ...) {
      stop("Could not resolve host")
    },
    .package = "httr2"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    class = "rxref_connection_error"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    regexp = "could not connect"
  )
})

test_that("rx_perform_json() gives a friendly error for malformed JSON", {
  req <- httr2::request("https://example.com")

  local_mocked_bindings(
    req_perform = function(req, ...) {
      httr2::response(
        status_code = 200,
        body = charToRaw("{not valid json")
      )
    },
    .package = "httr2"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    class = "rxref_json_error"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    regexp = "could not be parsed"
  )
})

test_that("rx_perform_json() gives a friendly error for empty response bodies", {
  req <- httr2::request("https://example.com")

  local_mocked_bindings(
    req_perform = function(req, ...) {
      httr2::response(
        status_code = 200,
        body = charToRaw("")
      )
    },
    .package = "httr2"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    class = "rxref_empty_response_error"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    regexp = "response was empty"
  )
})

test_that("rx_perform_json() parses successful JSON responses", {
  req <- httr2::request("https://example.com")

  local_mocked_bindings(
    req_perform = function(req, ...) {
      httr2::response(
        status_code = 200,
        body = charToRaw('{"idGroup":{"rxnormId":["123"]}}')
      )
    },
    .package = "httr2"
  )

  out <- rx_perform_json(req, service = "RxNorm")

  expect_type(out, "list")
  expect_equal(out$idGroup$rxnormId[[1]], "123")
})

test_that("specific API errors inherit from rxref_api_error", {
  req <- httr2::request("https://example.com")

  local_mocked_bindings(
    req_perform = function(req, ...) {
      httr2::response(status_code = 429)
    },
    .package = "httr2"
  )

  expect_error(
    rx_perform_json(req, service = "RxNorm"),
    class = "rxref_api_error"
  )
})

#' Should this function show a progress bar?
#'
#' @keywords internal
#' @noRd
.rxref_show_progress <- function(show_progress, x, min_n = 5L) {
  isTRUE(show_progress) && length(x) >= min_n
}


#' Map with optional cli progress bar
#'
#' @keywords internal
#' @noRd
.rxref_progress_map_dfr <- function(
    .x,
    .f,
    ...,
    name = "Working",
    show_progress = interactive(),
    min_n = 5L
) {
  use_progress <- .rxref_show_progress(show_progress, .x, min_n = min_n)

  progress_id <- NULL

  if (use_progress) {
    progress_id <- cli::cli_progress_bar(
      name = name,
      total = length(.x),
      type = "iterator"
    )

    on.exit(cli::cli_progress_done(id = progress_id), add = TRUE)
  }

  purrr::map_dfr(.x, function(.val) {
    out <- .f(.val, ...)

    if (use_progress) {
      cli::cli_progress_update(id = progress_id)
    }

    out
  })
}

#' Configure rxref
#'
#' @param base_url Override the RxNav base URL (e.g., a local mirror)
#' @param rate_delay Seconds to wait between HTTP calls
#' @param cache A cachem cache object used by memoised calls
#' @return A named list of current settings
#' @export
rxref_conf <- function(base_url = getOption("rxref.base_url"),
                       rate_delay = getOption("rxref.rate_delay"),
                       cache = getOption("rxref.cache")) {
  if (!missing(base_url)) {
    options(rxref.base_url = base_url)
  }

  if (!missing(rate_delay)) {
    options(rxref.rate_delay = rate_delay)
  }

  if (!missing(cache)) {
    if (!inherits(cache, "cachem")) {
      stop(
        "`cache` must be a cachem cache object, such as `cachem::cache_mem()`.",
        call. = FALSE
      )
    }

    options(rxref.cache = cache)
  }

  list(
    base_url = getOption("rxref.base_url"),
    rate_delay = getOption("rxref.rate_delay"),
    cache = getOption("rxref.cache")
  )
}

test_that("rxref_conf sets and returns options", {
  old <- options(
    rxref.base_url = getOption("rxref.base_url"),
    rxref.rate_delay = getOption("rxref.rate_delay"),
    rxref.cache = getOption("rxref.cache")
  )
  on.exit(options(old), add = TRUE)

  test_cache <- cachem::cache_mem()

  out <- rxref_conf(
    base_url = "https://example.com",
    rate_delay = 0.1,
    cache = test_cache
  )

  expect_equal(out$base_url, "https://example.com")
  expect_equal(out$rate_delay, 0.1)
  expect_identical(out$cache, test_cache)
})

test_that("rxref_conf validates cache objects", {
  expect_error(
    rxref_conf(cache = NULL),
    "`cache` must be a cachem cache object",
    fixed = TRUE
  )

  expect_error(
    rxref_conf(cache = "not-a-cache"),
    "`cache` must be a cachem cache object",
    fixed = TRUE
  )
})

#' rxref: Tidy RxNorm utilities
#'
#' `rxref` provides tidy, API-first tools for working with RxNorm and RxClass
#' data in R. The package helps users resolve drug names and identifiers,
#' inspect RxNorm concepts, map between RxCUIs and NDCs, expand ingredients to
#' product-level concepts, retrieve drug class information, and derive clinically
#' useful product attributes.
#'
#' Common workflows include:
#'
#' * Resolving drug names or identifiers with [resolve()] and [search_drug()].
#' * Retrieving concept metadata with [get_properties()].
#' * Building product-level medication lists with [find_ingredients()],
#'   [ingredients_for_rxcui()], and [products_for_ingredients()].
#' * Mapping between RxCUIs and NDCs with [map_to()],
#'   [map_ndc_to_rxcui()], and [map_rxcui_to_ndc()].
#' * Retrieving therapeutic class information with [get_classes()],
#'   [find_classes()], [get_class_members()], [get_atc()], [get_epc()],
#'   and [get_va()].
#' * Deriving clinical product attributes with [get_clinical_attributes()] and
#'   [filter_products_by_route()].
#'
#' Configure package behavior with [rxref_conf()].
#'
#' @keywords internal
"_PACKAGE"


#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr mutate select rename relocate arrange distinct filter left_join bind_rows bind_cols
#' @importFrom tidyr unnest unnest_wider unnest_longer replace_na
#' @importFrom purrr map map2 map_chr map_int map_lgl imap compact pluck possibly quietly
#' @importFrom stringr str_detect str_trim str_remove_all str_replace_all str_squish
#' @importFrom vctrs vec_recycle vec_cast vec_c vec_rbind
#' @importFrom jsonlite fromJSON
#' @importFrom memoise memoise
#' @importFrom rlang .data .env
NULL

#' Get core RxNorm properties for one or more RxCUIs
#'
#' @param rxcui Character vector of RxCUIs
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'
#' @return tibble with rxcui, name, synonym, tty, language, suppress, umlscui.
#'
#' @examples
#' if (identical(Sys.getenv("RXREF_ONLINE"), "1")) {
#' get_properties(c("860975","1049630"))
#' }
#' @export
get_properties <- function(rxcui, show_progress = interactive()) {
  stopifnot(is.character(rxcui))

  .rxref_progress_map_dfr(
    rxcui,
    function(id) {
      if (is.na(id) || !nzchar(id)) {
        return(empty_properties_result(id))
      }

      res <- rx_get_json(paste0("/rxcui/", id, "/properties"))
      p <- rx_pluck(res, "properties", .default = list())

      api_rxcui <- rx_scalar_chr(rx_pluck_chr(p, "rxcui"))

      tibble::tibble(
        rxcui = if (is.na(api_rxcui) || !nzchar(api_rxcui)) id else api_rxcui,
        name = rx_scalar_chr(rx_pluck_chr(p, "name")),
        synonym = rx_scalar_chr(rx_pluck_chr(p, "synonym")),
        tty = rx_scalar_chr(rx_pluck_chr(p, "tty")),
        language = rx_scalar_chr(rx_pluck_chr(p, "language")),
        suppress = rx_scalar_chr(rx_pluck_chr(p, "suppress")),
        umlscui = rx_scalar_chr(rx_pluck_chr(p, "umlscui"))
      )
    },
    name = "Getting properties",
    show_progress = show_progress
  )
}

#' Empty properties result
#'
#' @keywords internal
#' @noRd
empty_properties_result <- function(rxcui = NA_character_) {
  tibble::tibble(
    rxcui = rxcui,
    name = NA_character_,
    synonym = NA_character_,
    tty = NA_character_,
    language = NA_character_,
    suppress = NA_character_,
    umlscui = NA_character_
  )
}

test_that("get_properties parses RxNorm properties response", {
  fake_rx_get_json <- function(path, query = list()) {
    expect_equal(path, "/rxcui/29046/properties")

    list(
      properties = list(
        rxcui = "29046",
        name = "lisinopril",
        synonym = "",
        tty = "IN",
        language = "ENG",
        suppress = "N",
        umlscui = "C0065374"
      )
    )
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- get_properties("29046")

  expect_s3_class(out, "tbl_df")
  expect_equal(out$rxcui, "29046")
  expect_equal(out$name, "lisinopril")
  expect_equal(out$tty, "IN")
  expect_equal(out$suppress, "N")
})

test_that("get_properties() returns expected columns for missing input", {
  out <- get_properties(NA_character_, show_progress = FALSE)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1L)
  expect_named(out, c(
    "rxcui",
    "name",
    "synonym",
    "tty",
    "language",
    "suppress",
    "umlscui"
  ))
  expect_true(is.na(out$rxcui))
  expect_true(is.na(out$name))
})

#' Filter RxNorm product concepts by route of administration.
#'
#' Filters a product-level rxref table using route information from
#' [get_clinical_attributes()]. Useful when a drug or class includes
#' products across multiple routes, such as oral tablets, ophthalmic solutions,
#' injectables, patches, or topical formulations.
#'
#' @param products A tibble containing a `product_rxcui` column, such as the
#'   output of [products_for_ingredients()].
#' @param route Character vector of routes to keep. Common values include
#'   `"ORAL"`, `"INJECTION"`, `"OPHTHALMIC"`, `"OTIC"`, `"INHALATION"`,
#'   `"TOPICAL"`, `"RECTAL"`, `"VAGINAL"`, `"URETHRAL"`, and `"IMPLANT"`.
#' @param keep_route_info Logical. If `TRUE` (default), append route, dose form, and
#'   dose-form group summaries to the returned table.
#' @param include_historical Logical. If `TRUE`, use RxCUI history status
#'   metadata as a fallback for RxCUIs that do not return active clinical
#'   attributes. This is useful for obsolete, remapped, quantified, or otherwise
#'   non-current RxCUIs found in historical prescribing data. If `products`
#'   contains a `concept_status` column with non-`"active"` value, historical
#'   lookup is enabled automatically.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble containing only rows whose `product_rxcui` has at least one
#'   matching route.
#'
#' @examples
#' \dontrun{
#' ing <- find_ingredients("metoprolol")
#'
#' prods <- products_for_ingredients(
#'   ing$rxcui,
#'   include_combos = TRUE
#' )
#'
#' filter_products_by_route(prods, route = "ORAL")
#'
#' # Historical/current products can also be route-filtered
#' prods_hist <- products_for_ingredients(
#'   ing$rxcui,
#'   include_combos = TRUE,
#'   concept_status = "active_and_historical"
#' )
#'
#' filter_products_by_route(prods_hist, route = "ORAL")
#' }
#'
#' @export
filter_products_by_route <- function(
    products,
    route = "ORAL",
    keep_route_info = TRUE,
    include_historical = FALSE,
    show_progress = interactive()
) {
  stopifnot(is.data.frame(products))
  stopifnot(is.character(route), length(route) >= 1)

  if (!"product_rxcui" %in% names(products)) {
    cli::cli_abort("{.arg products} must contain a {.field product_rxcui} column.")
  }

  route_keep <- toupper(route)

  # infer whether historical lookup needed if concept_status in `products` tibble.
  include_historical <- isTRUE(include_historical) ||
    (
      "concept_status" %in% names(products) &&
        any(!is.na(products$concept_status) & products$concept_status != "Active")
    )

  product_ids <- unique(stats::na.omit(products$product_rxcui))

  if (!length(product_ids)) {
    return(products[0, , drop = FALSE])
  }

  attrs <- get_clinical_attributes(
    product_ids,
    include_historical = include_historical,
    show_progress = show_progress
  )

  collapse_nonmissing <- function(x) {
    x <- sort(unique(stats::na.omit(x)))
    if (!length(x)) {
      return(NA_character_)
    }
    paste(x, collapse = "; ")
  }

  route_tbl <- attrs |>
    dplyr::group_by(product_rxcui = .data$rxcui) |>
    dplyr::summarise(
      route_match = any(.data$route %in% route_keep, na.rm = TRUE),
      route = collapse_nonmissing(.data$route),
      dose_form = collapse_nonmissing(.data$dose_form),
      dose_form_group = collapse_nonmissing(.data$dose_form_group),
      .groups = "drop"
    )

  products_for_join <- products |>
    dplyr::select(
      -dplyr::any_of(c(
        "route_match",
        "route",
        "routes",
        "dose_form",
        "dose_forms",
        "dose_form_group",
        "dose_form_groups"
      ))
    )

  out <- products_for_join |>
    dplyr::left_join(route_tbl, by = "product_rxcui") |>
    dplyr::filter(.data$route_match %in% TRUE)

  if (!isTRUE(keep_route_info)) {
    out <- out |>
      dplyr::select(
        -dplyr::any_of(c(
          "route_match",
          "route",
          "dose_form",
          "dose_form_group"
        ))
      )
  } else {
    out <- out |>
      dplyr::select(-dplyr::any_of("route_match"))
  }

  out
}

#' @importFrom digest digest
NULL

#' @keywords internal
#' @noRd
rx_http_client <- function() {
  httr2::request(getOption("rxref.base_url")) |>
    httr2::req_user_agent(getOption("rxref.user_agent"))
}


#' @keywords internal
#' @noRd
rx_sleep <- function() {
  delay <- getOption("rxref.rate_delay", 0)
  if (isTRUE(delay > 0)) Sys.sleep(delay)
}

#' Perform an RxNav/RxClass request and parse JSON
#'
#' @param req An httr2 request object.
#' @param service Character label for the service, e.g. "RxNorm" or "RxClass".
#'
#' @return Parsed JSON as a list.
#'
#' @keywords internal
#' @noRd
rx_perform_json <- function(req, service = "RxNav") {
  resp <- tryCatch(
    httr2::req_perform(req),
    error = function(e) {
      rx_abort_api(
        paste(
          service, "API request failed.",
          "rxref could not connect to the API.",
          "Check your internet connection and try again.",
          sep = "\n"
        ),
        class = "rxref_connection_error",
        parent = e
      )
    }
  )

  status <- httr2::resp_status(resp)

  if (identical(status, 429L)) {
    rx_abort_api(
      paste(
        service, "API request failed with status 429.",
        "The request may have been rate-limited.",
        "Try increasing `rate_delay` or retrying later.",
        sep = "\n"
      ),
      class = "rxref_rate_limit_error"
    )
  }

  if (identical(status, 404L)) {
    rx_abort_api(
      paste(
        service, "API request failed with status 404.",
        "The requested endpoint or resource was not found.",
        sep = "\n"
      ),
      class = "rxref_not_found_error"
    )
  }

  if (status >= 500L) {
    rx_abort_api(
      paste(
        service, "API request failed with status", status, ".",
        "This appears to be a temporary server-side issue.",
        "Please retry later.",
        sep = "\n"
      ),
      class = "rxref_server_error"
    )
  }

  if (status >= 400L) {
    rx_abort_api(
      paste(
        service, "API request failed with status", status, ".",
        "The API returned an unsuccessful response.",
        sep = "\n"
      ),
      class = "rxref_http_error"
    )
  }

  txt <- tryCatch(
    httr2::resp_body_string(resp),
    error = function(e) {
      msg <- conditionMessage(e)

      if (grepl("empty body", msg, ignore.case = TRUE)) {
        rx_abort_api(
          paste(
            service, "API response was empty.",
            "The API returned an empty response body.",
            sep = "\n"
          ),
          class = "rxref_empty_response_error",
          parent = e
        )
      }

      rx_abort_api(
        paste(
          service, "API response could not be read.",
          "The API returned a response, but rxref could not read the response body.",
          sep = "\n"
        ),
        class = "rxref_response_error",
        parent = e
      )
    }
  )

  if (!nzchar(txt)) {
    rx_abort_api(
      paste(
        service, "API response was empty.",
        "The API returned an empty response body.",
        sep = "\n"
      ),
      class = "rxref_empty_response_error"
    )
  }

  tryCatch(
    jsonlite::fromJSON(txt, simplifyVector = FALSE),
    error = function(e) {
      rx_abort_api(
        paste(
          service, "API response could not be parsed.",
          "The API returned malformed or unexpected JSON.",
          "Please retry later.",
          sep = "\n"
        ),
        class = "rxref_json_error",
        parent = e
      )
    }
  )
}

# helpers for rx_get_json() caching
.rxref_memo_env <- new.env(parent = emptyenv())

.rxref_get_cache <- function() {
  opt <- getOption("rxref.cache")

  if (inherits(opt, "cachem")) {
    return(opt)
  }

  cachem::cache_mem()
}

.make_rx_get_json <- function(cache) {
  memoise::memoise(
    function(path, query = list()) {
      rx_sleep()

      path_json <- if (grepl("\\.json$", path)) {
        path
      } else {
        paste0(path, ".json")
      }

      req <- rx_http_client() |>
        httr2::req_url_path_append(path_json) |>
        httr2::req_url_query(!!!query, .multi = "explode")

      rx_perform_json(req, service = "RxNorm")
    },
    cache = cache
  )
}

#' @keywords internal
#' @noRd
rx_get_json <- function(path, query = list()) {
  cache <- .rxref_get_cache()

  if (
    is.null(.rxref_memo_env$rx_get_json) ||
    !identical(cache, .rxref_memo_env$rx_get_json_cache)
  ) {
    .rxref_memo_env$rx_get_json <- .make_rx_get_json(cache)
    .rxref_memo_env$rx_get_json_cache <- cache
  }

  .rxref_memo_env$rx_get_json(path = path, query = query)
}

#' @keywords internal
#' @noRd
rxclass_http_client <- function() {
  httr2::request(
    getOption(
      "rxref.rxclass_base_url",
      "https://rxnav.nlm.nih.gov/REST/rxclass"
    )
  ) |>
    httr2::req_user_agent(getOption("rxref.user_agent"))
}

#' @keywords internal
#' @noRd
.rxclass_collapse_query <- function(x) {
  if (is.null(x)) return(NULL)
  if (length(x) == 0) return(NULL)

  paste(x, collapse = " ")
}

.make_rxclass_get_json <- function(cache) {
  memoise::memoise(
    function(path, query = list()) {
      rx_sleep()

      path_json <- if (grepl("\\.json$", path)) {
        path
      } else {
        paste0(path, ".json")
      }

      req <- rxclass_http_client() |>
        httr2::req_url_path_append(path_json) |>
        httr2::req_url_query(!!!query, .multi = "explode")

      rx_perform_json(req, service = "RxClass")
    },
    cache = cache
  )
}

#' @keywords internal
#' @noRd
rxclass_get_json <- function(path, query = list()) {
  cache <- .rxref_get_cache()

  if (
    is.null(.rxref_memo_env$rxclass_get_json) ||
    !identical(cache, .rxref_memo_env$rxclass_get_json_cache)
  ) {
    .rxref_memo_env$rxclass_get_json <- .make_rxclass_get_json(cache)
    .rxref_memo_env$rxclass_get_json_cache <- cache
  }

  .rxref_memo_env$rxclass_get_json(path = path, query = query)
}

#' @keywords internal
#' @noRd
null2chr <- function(x) {
  if (is.null(x) || length(x) == 0) return(NA_character_)
  as.character(x[[1]])
}


# tidy helpers --------------------------------------------------------------

#' @keywords internal
#' @noRd
.as_tibble <- function(x) tibble::as_tibble(x, .name_repair = "unique")

#' @keywords internal
#' @noRd
null2na <- function(x) if (is.null(x)) NA else x

#' @keywords internal
#' @noRd
vec_recycle_len <- function(x, n) vctrs::vec_recycle(x, n)

#' @keywords internal
#' @noRd
is_rxcui <- function(x) {
  if (!is.character(x) || length(x) != 1L || is.na(x)) return(FALSE)

  # Only digits
  if (!grepl("^[0-9]+$", x)) return(FALSE)

  # If it's 10 or 11 digits, treat as NDC-ish, not RxCUI
  !nchar(x) %in% c(10L, 11L)
}

#' @keywords internal
#' @noRd
is_ndcish <- function(x) {
  if (!is.character(x) || length(x) != 1L || is.na(x)) return(FALSE)

  digits <- gsub("[^0-9]", "", x)

  grepl("^[0-9]{10,11}$", digits)
}


#' @keywords internal
#' @noRd
.ndc_parts_from_string <- function(x) {
  if (!is.character(x) || length(x) != 1L || is.na(x)) {
    return(list(
      labeler = NA_character_,
      product = NA_character_,
      package = NA_character_,
      raw = x
    ))
  }

  raw <- x
  x <- trimws(x)

  # Preserve hyphenated structure if present
  cleaned <- gsub("[^0-9-]", "", x)

  if (grepl("^\\d+-\\d+-\\d+$", cleaned)) {
    p <- strsplit(cleaned, "-", fixed = TRUE)[[1]]

    return(list(
      labeler = p[[1]],
      product = p[[2]],
      package = p[[3]],
      raw = raw
    ))
  }

  digits <- gsub("[^0-9]", "", x)

  if (!grepl("^[0-9]{10,11}$", digits)) {
    return(list(
      labeler = NA_character_,
      product = NA_character_,
      package = NA_character_,
      raw = raw
    ))
  }

  # If 11 digits, assume already normalized 5-4-2
  if (nchar(digits) == 11L) {
    return(list(
      labeler = substr(digits, 1L, 5L),
      product = substr(digits, 6L, 9L),
      package = substr(digits, 10L, 11L),
      raw = raw
    ))
  }

  # If 10 digits and not hyphenated, grouping cannot be known reliably
  list(
    labeler = NA_character_,
    product = NA_character_,
    package = NA_character_,
    raw = raw
  )
}

#' @keywords internal
#' @noRd
ndc_to_11 <- function(ndc) {
  stopifnot(is.character(ndc))

  purrr::map_chr(ndc, function(x) {
    if (is.na(x) || !nzchar(x)) {
      return(NA_character_)
    }

    x <- trimws(x)

    # If already an 11-digit NDC, return digits only
    digits_only <- gsub("[^0-9]", "", x)

    if (nchar(digits_only) == 11 && !grepl("-", x)) {
      return(digits_only)
    }

    # Handle hyphenated FDA 10-digit formats
    parts <- strsplit(x, "-", fixed = TRUE)[[1]]

    if (length(parts) == 3) {
      parts <- trimws(parts)

      labeler <- parts[[1]]
      product <- parts[[2]]
      package <- parts[[3]]

      widths <- nchar(parts)

      # 4-4-2 format: pad labeler to 5 digits
      if (identical(widths, c(4L, 4L, 2L))) {
        return(paste0(
          stringr::str_pad(labeler, width = 5, side = "left", pad = "0"),
          product,
          package
        ))
      }

      # 5-3-2 format: pad product to 4 digits
      if (identical(widths, c(5L, 3L, 2L))) {
        return(paste0(
          labeler,
          stringr::str_pad(product, width = 4, side = "left", pad = "0"),
          package
        ))
      }

      # 5-4-1 format: pad package to 2 digits
      if (identical(widths, c(5L, 4L, 1L))) {
        return(paste0(
          labeler,
          product,
          stringr::str_pad(package, width = 2, side = "left", pad = "0")
        ))
      }
    }

    # Do not infer 11-digit format from non-hyphenated 10-digit NDCs.
    # Without hyphens, the original FDA format is ambiguous, e.g.,
    # 4-4-2, 5-3-2, or 5-4-1?
    if (nchar(digits_only) == 10L && !grepl("-", x)) {
      return(NA_character_)
    }


    # Fallback: if removing punctuation gives 11 digits, use that
    if (nchar(digits_only) == 11) {
      return(digits_only)
    }

    NA_character_
  })
}

#' @keywords internal
#' @noRd
hyphenate_ndc_5_4_2 <- function(ndc) {
  if (!is.character(ndc)) {
    return(ndc)
  }

  purrr::map_chr(ndc, function(x) {
    if (is.na(x)) return(NA_character_)

    digits <- gsub("[^0-9]", "", x)

    if (!grepl("^[0-9]{11}$", digits)) {
      return(x)
    }

    paste0(
      substr(digits, 1L, 5L), "-",
      substr(digits, 6L, 9L), "-",
      substr(digits, 10L, 11L)
    )
  })
}

test_that("ndc_to_11 handles common FDA 10-digit formats", {
  expect_equal(rxref:::ndc_to_11("0002-0800-01"), "00002080001") # 4-4-2
  expect_equal(rxref:::ndc_to_11("50242-040-62"), "50242004062") # 5-3-2
  expect_equal(rxref:::ndc_to_11("60505-0829-1"), "60505082901") # 5-4-1
  expect_equal(rxref:::ndc_to_11("60505-0829-01"), "60505082901") # 5-4-2

  # Unhyphenated 10-digit NDCs are ambiguous because the original grouping
  # could be 4-4-2, 5-3-2, or 5-4-1.
  expect_equal(rxref:::ndc_to_11("1234567890"), NA_character_)
})

test_that("ndc_to_11 handles non-hyphenated values", {
  expect_equal(rxref:::ndc_to_11("60505082901"), "60505082901")
  expect_equal(rxref:::ndc_to_11("00002080001"), "00002080001")
  expect_equal(rxref:::ndc_to_11("50242004062"), "50242004062")
})

test_that("is_ndcish identifies 10/11 digit NDC-like strings", {
  expect_true(rxref:::is_ndcish("1234567890"))
  expect_true(rxref:::is_ndcish("12345-6789-01"))
  expect_false(rxref:::is_ndcish("lisinopril"))
})

test_that("is_ndcish recognizes 10- and 11-digit NDC-like values", {
  expect_true(rxref:::is_ndcish("0002-0800-01"))
  expect_true(rxref:::is_ndcish("00002080001"))
  expect_true(rxref:::is_ndcish("50242-040-62"))
  expect_true(rxref:::is_ndcish("60505-0829-1"))

  expect_false(rxref:::is_ndcish(NA_character_))
  expect_false(rxref:::is_ndcish("lisinopril"))
  expect_false(rxref:::is_ndcish("12345"))
})

test_that(".ndc_parts_from_string parses hyphenated and normalized NDCs", {
  x <- rxref:::.ndc_parts_from_string("0002-0800-01")

  expect_equal(x$labeler, "0002")
  expect_equal(x$product, "0800")
  expect_equal(x$package, "01")

  y <- rxref:::.ndc_parts_from_string("00002080001")

  expect_equal(y$labeler, "00002")
  expect_equal(y$product, "0800")
  expect_equal(y$package, "01")
})

test_that(".ndc_parts_from_string does not infer grouping for unhyphenated 10-digit NDCs", {
  x <- rxref:::.ndc_parts_from_string("0002080001")

  expect_true(is.na(x$labeler))
  expect_true(is.na(x$product))
  expect_true(is.na(x$package))
})

test_that("hyphenate_ndc_5_4_2 hyphenates normalized 11-digit NDCs", {
  expect_equal(
    rxref:::hyphenate_ndc_5_4_2("00002080001"),
    "00002-0800-01"
  )

  expect_equal(
    rxref:::hyphenate_ndc_5_4_2("50242004062"),
    "50242-0040-62"
  )
})

test_that("hyphenate_ndc_5_4_2 is vectorized and preserves invalid inputs", {
  expect_equal(
    rxref:::hyphenate_ndc_5_4_2(c("00002080001", "not-an-ndc", NA_character_)),
    c("00002-0800-01", "not-an-ndc", NA_character_)
  )
})

test_that("resolve() handles invalid NDCs without API calls", {
  out <- resolve("not-an-ndc", type = "ndc")

  expect_s3_class(out, "tbl_df")
  expect_equal(out$input, "not-an-ndc")
  expect_equal(out$type, "ndc")
  expect_true(is.na(out$ndc11))
  expect_true(is.na(out$rxcui))
  expect_true(is.na(out$name))
  expect_true(is.na(out$tty))
})

test_that("map_to() handles invalid NDCs without API calls", {
  out <- map_to("not-an-ndc", to = "rxcui")

  expect_s3_class(out, "tbl_df")
  expect_equal(out$input, "not-an-ndc")
  expect_true(is.na(out$ndc11))
  expect_true(is.na(out$rxcui))
})

test_that("find_ingredients finds lisinopril online", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))

  out <- find_ingredients("lisinopril")

  expect_s3_class(out, "tbl_df")
  expect_true(any(out$rxcui == "29046"))
  expect_true(any(out$tty == "IN"))
})

test_that("get_properties retrieves known RxCUI online", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))

  out <- get_properties("29046")

  expect_equal(out$rxcui, "29046")
  expect_equal(tolower(out$name), "lisinopril")
  expect_equal(out$tty, "IN")
})

test_that("rx_get_json gives informative error when API is unavailable", {
  old <- getOption("rxref.base_url")
  on.exit(options(rxref.base_url = old), add = TRUE)

  options(rxref.base_url = "http://127.0.0.1:9")

  expect_error(
    rx_get_json("/rxcui", query = list(name = "lisinopril")),
    class = "rxref_api_error"
  )
})

test_that("rxclass_get_json gives informative error when API is unavailable", {
  old <- getOption("rxref.rxclass_base_url")
  on.exit(options(rxref.rxclass_base_url = old), add = TRUE)

  options(rxref.rxclass_base_url = "http://127.0.0.1:9")

  expect_error(
    rxclass_get_json("/class/byDrugName", query = list(drugName = "lisinopril")),
    class = "rxref_api_error"
  )
})

test_that(".rxclass_collapse_query collapses vectors with spaces", {
  expect_equal(rxref:::.rxclass_collapse_query(c("ATC", "EPC")), "ATC EPC")
  expect_null(rxref:::.rxclass_collapse_query(NULL))
  expect_null(rxref:::.rxclass_collapse_query(character()))
})

#' Exclude products containing specified ingredients
#'
#' Remove all rows for products that contain one or more user-specified
#' ingredients. This is useful when broad ingredient-based product searches
#' return fixed-dose combination products that contain an ingredient of interest
#' but are not clinically relevant to a specific study definition.
#'
#' Exclusions are applied at the product level. If any ingredient in a product
#' matches the exclusion list, all rows for that product are removed.
#'
#' @param data A data frame containing product RxCUIs and ingredient information,
#'   such as that produced by [products_for_ingredients()].
#' @param ingredients Optional character vector of ingredient names to exclude.
#'   These are resolved to ingredient RxCUIs using [find_ingredients()].
#' @param ingredient_rxcuis Optional character vector of ingredient RxCUIs to
#'   exclude. These are used directly and do not require name resolution.
#' @param product_id_col Name of the column containing product RxCUIs. Default
#'   is `product_rxcui`.
#' @param ingredient_name_col Name of the column containing ingredient names.
#'   Used for reporting and, when needed, fallback matching. Default is
#'   `ingredient_name`.
#' @param ingredient_rxcui_col Name of the column containing ingredient RxCUIs.
#'   Default is `ingredient_rxcui`.
#' @param return_excluded Logical. If `FALSE`, return the filtered data frame.
#'   If `TRUE`, return a list with `data`, `excluded`, and
#'   `resolved_ingredients`.
#'
#' @return
#' If `return_excluded = FALSE`, a data frame of the same general shape as
#' `data`, with excluded product rows removed.
#'
#' If `return_excluded = TRUE`, a list with:
#' \describe{
#'   \item{data}{The filtered data frame.}
#'   \item{excluded}{Rows from the input data belonging to excluded products.}
#'   \item{resolved_ingredients}{The ingredient names/RxCUIs used for exclusion.}
#' }
#'
#' @examples
#' \dontrun{
#' antihtn_clean <- antihtn_products |>
#'   exclude_products_with_ingredients(ingredients = "sacubitril")
#'
#' antihtn_audit <- antihtn_products |>
#'   exclude_products_with_ingredients(
#'     ingredients = "sacubitril",
#'     return_excluded = TRUE
#'   )
#'
#' antihtn_audit$data
#' antihtn_audit$excluded
#' antihtn_audit$resolved_ingredients
#' }
#'
#' @export
exclude_products_with_ingredients <- function(
    data,
    ingredients = NULL,
    ingredient_rxcuis = NULL,
    product_id_col = "product_rxcui",
    ingredient_name_col = "ingredient_name",
    ingredient_rxcui_col = "ingredient_rxcui",
    return_excluded = FALSE
) {
  if (!is.data.frame(data)) {
    stop("`data` must be a data frame.", call. = FALSE)
  }

  if (!product_id_col %in% names(data)) {
    stop(
      "`data` must contain the product ID column `",
      product_id_col,
      "`.",
      call. = FALSE
    )
  }

  has_ingredient_rxcui_col <- ingredient_rxcui_col %in% names(data)
  has_ingredient_name_col <- ingredient_name_col %in% names(data)

  if (!has_ingredient_rxcui_col && !has_ingredient_name_col) {
    stop(
      "`data` must contain either `",
      ingredient_rxcui_col,
      "` or `",
      ingredient_name_col,
      "`.",
      call. = FALSE
    )
  }

  if (is.null(ingredients) && is.null(ingredient_rxcuis)) {
    stop(
      "Supply at least one of `ingredients` or `ingredient_rxcuis`.",
      call. = FALSE
    )
  }

  ingredient_rxcuis <- unique(as.character(ingredient_rxcuis))
  ingredient_rxcuis <- ingredient_rxcuis[!is.na(ingredient_rxcuis) & ingredient_rxcuis != ""]

  resolved_ingredients <- tibble::tibble(
    input = character(),
    ingredient_rxcui = character(),
    ingredient_name = character(),
    tty = character(),
    score = numeric()
  )

  if (!is.null(ingredients)) {
    ingredients <- unique(as.character(ingredients))
    ingredients <- ingredients[!is.na(ingredients) & ingredients != ""]

    if (length(ingredients) > 0) {
      resolved_ingredients <- find_ingredients(ingredients) |>
        dplyr::transmute(
          input = .data$input,
          ingredient_rxcui = .data$rxcui,
          ingredient_name = .data$name,
          tty = .data$tty,
          score = .data$score
        )

      resolved_rxcuis <- unique(as.character(resolved_ingredients$ingredient_rxcui))
      resolved_rxcuis <- resolved_rxcuis[!is.na(resolved_rxcuis) & resolved_rxcuis != ""]

      ingredient_rxcuis <- unique(c(ingredient_rxcuis, resolved_rxcuis))
    }
  }

  if (length(ingredient_rxcuis) == 0 && !has_ingredient_name_col) {
    stop(
      "No valid ingredient RxCUIs were supplied or resolved.",
      call. = FALSE
    )
  }

  ingredient_map <- normalize_product_ingredients(
    data = data,
    product_id_col = product_id_col,
    ingredient_name_col = ingredient_name_col,
    ingredient_rxcui_col = ingredient_rxcui_col
  )

  if (length(ingredient_rxcuis) > 0 && "ingredient_rxcui" %in% names(ingredient_map)) {
    products_to_exclude <- ingredient_map |>
      dplyr::filter(.data$ingredient_rxcui %in% ingredient_rxcuis) |>
      dplyr::distinct(.data$product_id)
  } else {
    products_to_exclude <- tibble::tibble(product_id = character())
  }

  # Fallback: if ingredient names were supplied but no RxCUI match was possible,
  # use normalized exact name matching.
  if (
    nrow(products_to_exclude) == 0 &&
    !is.null(ingredients) &&
    has_ingredient_name_col &&
    "ingredient_name" %in% names(ingredient_map)
  ) {
    ingredients_norm <- normalize_ingredient_name(ingredients)

    products_to_exclude <- ingredient_map |>
      dplyr::filter(normalize_ingredient_name(.data$ingredient_name) %in% ingredients_norm) |>
      dplyr::distinct(.data$product_id)
  }

  product_ids_to_exclude <- products_to_exclude$product_id

  filtered_data <- data |>
    dplyr::filter(!as.character(.data[[product_id_col]]) %in% product_ids_to_exclude)

  excluded_data <- data |>
    dplyr::filter(as.character(.data[[product_id_col]]) %in% product_ids_to_exclude)

  if (return_excluded) {
    return(
      list(
        data = filtered_data,
        excluded = excluded_data,
        resolved_ingredients = resolved_ingredients
      )
    )
  }

  filtered_data
}

# Internal helpers

#' Normalize product ingredient fields
#'
#' Converts product-level ingredient columns into a long product-ingredient
#' mapping table. Handles single values, semicolon-delimited fixed-dose
#' combination values, and older pseudo-vector strings.
#'
#' @param data A data frame containing product and ingredient columns.
#' @param product_id_col Name of the product identifier column.
#' @param ingredient_name_col Name of the ingredient name column.
#' @param ingredient_rxcui_col Name of the ingredient RxCUI column.
#'
#' @return A tibble with `product_id`, `ingredient_rxcui`, and
#'   `ingredient_name`.
#'
#' @noRd
normalize_product_ingredients <- function(
    data,
    product_id_col,
    ingredient_name_col,
    ingredient_rxcui_col
) {
  product_id <- as.character(data[[product_id_col]])

  has_rxcui <- ingredient_rxcui_col %in% names(data)
  has_name <- ingredient_name_col %in% names(data)

  rxcui_list <- if (has_rxcui) {
    split_ingredient_field(data[[ingredient_rxcui_col]])
  } else {
    vector("list", length(product_id))
  }

  name_list <- if (has_name) {
    split_ingredient_field(data[[ingredient_name_col]])
  } else {
    vector("list", length(product_id))
  }

  pieces <- lapply(seq_along(product_id), function(i) {
    rxcuis <- if (has_rxcui) rxcui_list[[i]] else NA_character_
    names_ <- if (has_name) name_list[[i]] else NA_character_

    n <- max(length(rxcuis), length(names_))

    tibble::tibble(
      product_id = rep(product_id[[i]], n),
      ingredient_rxcui = rep_len(as.character(rxcuis), n),
      ingredient_name = rep_len(as.character(names_), n)
    )
  })

  dplyr::bind_rows(pieces) |>
    dplyr::mutate(
      ingredient_rxcui = trimws(.data$ingredient_rxcui),
      ingredient_name = trimws(.data$ingredient_name)
    ) |>
    dplyr::filter(!is.na(.data$product_id), .data$product_id != "")
}

#' Split ingredient fields
#'
#' Splits ingredient fields that may contain single values, semicolon-delimited
#' values, or older pseudo-vector strings such as `c("a", "b")`.
#'
#' @param x A character vector.
#'
#' @return A list of character vectors.
#'
#' @noRd
split_ingredient_field <- function(x) {
  lapply(x, function(value) {
    if (is.null(value) || length(value) == 0 || is.na(value)) {
      return(NA_character_)
    }

    value <- as.character(value)

    # Handle semicolon-delimited values, e.g. "sacubitril; valsartan"
    if (grepl(";", value, fixed = TRUE)) {
      return(trimws(strsplit(value, ";", fixed = TRUE)[[1]]))
    }

    # Handle older pseudo-vector strings, e.g. c("sacubitril", "valsartan")
    if (grepl("^c\\(", value)) {
      value <- gsub("^c\\(|\\)$", "", value)
      value <- gsub("\"|'", "", value)
      return(trimws(strsplit(value, ",", fixed = TRUE)[[1]]))
    }

    trimws(value)
  })
}

#' Normalize ingredient names for matching
#'
#' Converts ingredient names to trimmed lower-case character values.
#'
#' @param x A character vector.
#'
#' @return A normalized character vector.
#'
#' @noRd
normalize_ingredient_name <- function(x) {
  x |>
    as.character() |>
    trimws() |>
    tolower()
}

test_that("exclude_products_with_ingredients removes entire matching product", {
  dat <- tibble::tibble(
    product_rxcui = c("1", "1", "2"),
    name = c(
      "sacubitril / valsartan tablet",
      "sacubitril / valsartan tablet",
      "valsartan tablet"
    ),
    ingredient_rxcui = c("161", "69749", "69749"),
    ingredient_name = c("sacubitril", "valsartan", "valsartan")
  )

  out <- exclude_products_with_ingredients(
    dat,
    ingredient_rxcuis = "161"
  )

  expect_equal(unique(out$product_rxcui), "2")
  expect_false("1" %in% out$product_rxcui)
})


test_that("exclude_products_with_ingredients handles semicolon-delimited ingredients", {
  dat <- tibble::tibble(
    product_rxcui = c("1", "2"),
    name = c(
      "sacubitril / valsartan tablet",
      "valsartan tablet"
    ),
    ingredient_rxcui = c("161; 69749", "69749"),
    ingredient_name = c("sacubitril; valsartan", "valsartan")
  )

  out <- exclude_products_with_ingredients(
    dat,
    ingredient_rxcuis = "161"
  )

  expect_equal(nrow(out), 1)
  expect_equal(out$product_rxcui, "2")
})


test_that("exclude_products_with_ingredients returns audit output", {
  dat <- tibble::tibble(
    product_rxcui = c("1", "2"),
    name = c(
      "sacubitril / valsartan tablet",
      "valsartan tablet"
    ),
    ingredient_rxcui = c("161; 69749", "69749"),
    ingredient_name = c("sacubitril; valsartan", "valsartan")
  )

  out <- exclude_products_with_ingredients(
    dat,
    ingredient_rxcuis = "161",
    return_excluded = TRUE
  )

  expect_named(out, c("data", "excluded", "resolved_ingredients"))
  expect_equal(out$data$product_rxcui, "2")
  expect_equal(out$excluded$product_rxcui, "1")
})


test_that("exclude_products_with_ingredients errors when no exclusion input supplied", {
  dat <- tibble::tibble(
    product_rxcui = "1",
    ingredient_rxcui = "161"
  )

  expect_error(
    exclude_products_with_ingredients(dat),
    "Supply at least one"
  )
})


test_that("exclude_products_with_ingredients errors without product column", {
  dat <- tibble::tibble(
    rxcui = "1",
    ingredient_rxcui = "161"
  )

  expect_error(
    exclude_products_with_ingredients(dat, ingredient_rxcuis = "161"),
    "product ID column"
  )
})

#' Resolve a free-text drug name to ingredient CUIs (IN/PIN)
#'
#' Uses RxNav approximateTerm, then fetches properties for each candidate and
#' filters to ingredient-type concepts (TTY `IN` and, optionally, `PIN`).
#'
#' @param term Character vector (free text).
#' @param max_entries Integer. Max approximate-term candidates per input (default 10).
#' @param include_pin Logical. Include precise-ingredient (`PIN`) CUIs (default TRUE).
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble with columns: `input`, `rxcui`, `name`, `tty`, `score`.
#'
#' @examples
#' if (identical(Sys.getenv("RXREF_ONLINE"), "1")) {
#'   find_ingredients("metformin")
#' }
#' @export
find_ingredients <- function(
    term,
    max_entries = 10,
    include_pin = TRUE,
    show_progress = interactive()
) {
  stopifnot(is.character(term))

  .rxref_progress_map_dfr(
    term,
    function(t) {
      approx <- rx_get_json(
        "/approximateTerm",
        query = list(term = t, maxEntries = max_entries)
      )

      cand <- rx_pluck_list(approx, "approximateGroup", "candidate")

      if (!length(cand)) {
        return(tibble::tibble(
          input = character(),
          rxcui = character(),
          name = character(),
          tty = character(),
          score = numeric()
        ))
      }

      rows <- purrr::map_dfr(cand, function(c) {
        id <- rx_scalar_chr(rx_pluck_chr(c, "rxcui"))

        pr <- if (!is.na(id)) {
          rx_get_json(paste0("/rxcui/", id, "/properties"))
        } else {
          NULL
        }

        tibble::tibble(
          input = t,
          rxcui = id,
          name = rx_scalar_chr(rx_pluck_chr(pr, "properties", "name")),
          tty = rx_scalar_chr(rx_pluck_chr(pr, "properties", "tty")),
          score = suppressWarnings(as.numeric(rx_scalar_chr(rx_pluck_chr(c, "score"))))
        )
      })

      keep_ttys <- if (include_pin) c("IN", "PIN") else "IN"

      rows |>
        dplyr::filter(.data$tty %in% keep_ttys) |>
        dplyr::arrange(dplyr::desc(.data$score))
    },
    name = "Finding ingredients",
    show_progress = show_progress
  ) |>
    dplyr::distinct()
}


#' Expand ingredient CUIs to product CUIs that truly contain the ingredient
#'
#' Tries multiple RxNav endpoints and verifies candidates truly contain the
#' queried ingredient or one of its acceptable related ingredient concepts.
#' The function unions candidates from multiple sources, verifies ingredient
#' containment, and reports the number of ingredients represented by each
#' product concept.
#'
#' @param ingredient_rxcui Character vector of ingredient CUIs (TTY `IN` or `PIN`).
#' @param ttys Character vector of TTYs to include. Defaults to product-facing
#'   TTYs returned by [product_ttys()]. Pass a larger set if you want groups,
#'   components, branded concepts, or other product-related concepts, for example
#'   `product_ttys("extended_product")`.
#' @param route Optional character vector of routes to retain. If `NULL`, no
#'   route filtering is performed. Route filtering uses [get_clinical_attributes()].
#'   Route filtering is intended for product-level TTYs and may not filter well
#'   on broader group or package TTYs.
#' @param include_combos Logical. If `FALSE`, keep only single-ingredient
#'   products, where ingredient count is based on distinct `IN` concepts when
#'   available and otherwise falls back to distinct `PIN` concepts.
#' @param concept_status Character. Which RxNorm concept universe to search.
#'   `"active"` uses active-scope RxNav relationship endpoints and is the
#'   default. `"active_and_historical"` also searches historical RxNorm concepts
#'   using all-status concept retrieval and RxCUI history status metadata.
#'   Historical searching is slower and is intended for mapping older
#'   prescribing or dispensing data.
#' @param historical_status Character vector of historical RxNorm statuses to
#'   include when `concept_status = "active_and_historical"`. Defaults to
#'   `c("Obsolete", "Remapped", "Quantified", "NotCurrent")`. These values use
#'   RxNorm status definitions:
#'   \describe{
#'     \item{`"Obsolete"`}{The concept is obsolete in the current RxNorm data set,
#'       and RxNorm has not designated an active concept as equivalent.}
#'     \item{`"Remapped"`}{The concept was active or obsolete at one time, is no
#'       longer in the current data set, and has been remapped to one or more
#'       active or obsolete concepts.}
#'     \item{`"Quantified"`}{The concept has been designated as non-dispensable
#'       because it lacks a quantity factor; related concepts with quantity
#'       factors may be available.}
#'     \item{`"NotCurrent"`}{The concept either exists in the current data set
#'       without RxNorm vocabulary terms, or existed in a previous monthly
#'       release but has since been removed and not remapped.}
#'   }
#'   See the RxNorm API documentation for concept status values:
#'   \url{https://lhncbc.nlm.nih.gov/RxNav/APIs/api-RxNorm.getAllConceptsByStatus.html}.
#' @param show_progress Logical. Show progress bars for long-running API
#'   retrieval, product matching, and optional route filtering steps. Defaults
#'   to `interactive()`.
#'
#' @return A tibble with one row per matched ingredient/product concept pair.
#'   For `concept_status = "active"`, columns include `ingredient_rxcui`,
#'   `product_rxcui`, `name`, `tty`, and `n_ingredients`. When
#'   `concept_status = "active_and_historical"`, additional columns include
#'   `concept_status`, `active_start_date`, `active_end_date`,
#'   `release_start_date`, and `release_end_date`.
#'
#' @export
products_for_ingredients <- function(ingredient_rxcui,
                                     ttys = .rxref_default_ttys,
                                     route = NULL,
                                     include_combos = TRUE,
                                     concept_status = c("active", "active_and_historical"),
                                     historical_status = c("Obsolete", "Remapped", "Quantified", "NotCurrent"),
                                     show_progress = interactive()) {

  concept_status <- match.arg(concept_status)

  include_historical <- identical(concept_status, "active_and_historical")

  stopifnot(
    is.character(ingredient_rxcui),
    is.character(ttys),
    length(ttys) >= 1,
    is.character(historical_status)
  )

  tty_vec <- unique(ttys)
  historical_status <- unique(historical_status)

  ingredient_ids <- unique(stats::na.omit(ingredient_rxcui))

  # helper: safe scalarization
  nz1 <- function(x) rx_scalar_chr(x)

  # helper: collect concepts, using group tty when concept tty is missing
  collect_concepts <- function(groups, allowed_ttys) {
    if (is.null(groups) || !length(groups)) {
      return(tibble::tibble(
        product_rxcui = character(),
        name = character(),
        tty  = character()
      ))
    }

    rows <- purrr::map_dfr(groups, function(grp) {
      grp_tty <- nz1(grp$tty)
      cps <- grp$conceptProperties
      if (is.null(cps) || !length(cps)) {
        return(tibble::tibble(
          product_rxcui = character(),
          name = character(),
          tty  = character()
        ))
      }
      purrr::map_dfr(cps, function(cp) {
        tty_here <- nz1(cp$tty)
        tibble::tibble(
          product_rxcui = nz1(cp$rxcui),
          name          = nz1(cp$name),
          tty           = if (!is.na(tty_here) && nzchar(tty_here)) tty_here else grp_tty
        )
      })
    })
    if (!nrow(rows)) {
      return(tibble::tibble(
        product_rxcui = character(),
        name = character(),
        tty  = character()
      ))
    }
    rows |>
      dplyr::filter(
        !is.na(.data$product_rxcui),
        !is.na(.data$tty),
        .data$tty %in% allowed_ttys
      ) |>
      dplyr::distinct()
  }

  # acceptance set: input IN + its PIN children; also a loose name pattern
  acceptance_for_ing <- function(ing) {
    pr <- tryCatch(rx_get_json(paste0("/rxcui/", ing, "/properties")), error = function(e) NULL)
    ing_name <- tolower(rx_scalar_chr(rx_pluck_chr(pr, "properties", "name")))

    if (is.na(ing_name) || !nzchar(ing_name)) {
      base_pat <- "a^"
    } else {
      base_token <- sub("\\s+.*$", "", ing_name)
      base_pat <- paste0("\\b", gsub("([\\W_])", "\\\\\\1", base_token), "\\b")
    }

    pins <- character(0)
    tty_self <- rx_scalar_chr(rx_pluck_chr(pr, "properties", "tty"))
    if (!identical(tty_self, "PIN")) {
      rel <- tryCatch(
        rx_get_json(paste0("/rxcui/", ing, "/related"),
                    query = list(tty = "PIN", rela = "has_precise_ingredient")),
        error = function(e) NULL
      )
      pins_tbl <- collect_concepts(
        rx_pluck_list(rel, "relatedGroup", "conceptGroup"),
        "PIN"
      )
      pins <- pins_tbl$product_rxcui
    }
    list(cui_ok = unique(c(ing, pins)), name_pat = base_pat)
  }

  # fetchers that each return a tibble: product_rxcui, name, tty
  fetch_via_rela <- function(ing) {
    rel <- tryCatch(
      rx_get_json(paste0("/rxcui/", ing, "/related"),
                  query = list(tty = tty_vec, rela = "ingredient_of")),
      error = function(e) NULL
    )
    collect_concepts(rx_pluck_list(rel, "relatedGroup", "conceptGroup"), tty_vec)
  }
  fetch_via_related <- function(ing) {
    rel <- tryCatch(
      rx_get_json(paste0("/rxcui/", ing, "/related"),
                  query = list(tty = tty_vec)),
      error = function(e) NULL
    )
    collect_concepts(rx_pluck_list(rel, "relatedGroup", "conceptGroup"), tty_vec)
  }
  fetch_via_allrelated <- function(ing) {
    rel <- tryCatch(
      rx_get_json(paste0("/rxcui/", ing, "/allrelated")),
      error = function(e) NULL
    )

    collect_concepts(
      rx_pluck_list(rel, "allRelatedGroup", "conceptGroup"),
      tty_vec
    )
  }
  fetch_via_drugs_name <- function(ing) {
    props <- tryCatch(
      rx_get_json(paste0("/rxcui/", ing, "/properties")),
      error = function(e) NULL
    )

    nm <- rx_scalar_chr(rx_pluck_chr(props, "properties", "name"))

    if (is.na(nm) || !nzchar(nm)) {
      return(tibble::tibble(
        product_rxcui = character(),
        name = character(),
        tty = character()
      ))
    }

    dg <- tryCatch(
      rx_get_json("/drugs", query = list(name = nm)),
      error = function(e) NULL
    )

    collect_concepts(
      rx_pluck_list(dg, "drugGroup", "conceptGroup"),
      tty_vec
    )
  }

  fetch_allstatus_products <- function() {
    .rxref_progress_map_dfr(
      historical_status,
      function(status) {
        res <- tryCatch(
          rx_get_json("/allstatus", query = list(status = status)),
          error = function(e) NULL
        )

        concepts <- rx_pluck_list(res, "minConceptGroup", "minConcept")
        concepts <- as_rx_records(concepts)

        if (!length(concepts)) {
          return(tibble::tibble(
            product_rxcui = character(),
            name = character(),
            tty = character(),
            concept_status = character()
          ))
        }

        purrr::map_dfr(concepts, function(cp) {
          tibble::tibble(
            product_rxcui = nz1(cp$rxcui),
            name = nz1(cp$name),
            tty = nz1(cp$tty),
            concept_status = status
          )
        })
      },
      name = "Fetching historical products from RxNorm",
      show_progress = show_progress,
      min_n = 1L
    ) |>
      dplyr::filter(
        !is.na(.data$product_rxcui),
        !is.na(.data$name),
        !is.na(.data$tty),
        .data$tty %in% tty_vec
      ) |>
      dplyr::distinct()
  }

  filter_allstatus_name <- function(allstatus_products, name_pat) {
    if (identical(name_pat, "a^") || !nrow(allstatus_products)) {
      return(tibble::tibble(
        product_rxcui = character(),
        name = character(),
        tty = character(),
        concept_status = character()
      ))
    }

    allstatus_products |>
      dplyr::filter(grepl(name_pat, tolower(.data$name), perl = TRUE)) |>
      dplyr::distinct()
  }

  # verify that product lists the ingredient (IN/PIN) by CUI or name
  verify_contains <- function(prod_rxcui, cui_ok, name_pat) {
    rel_ing <- tryCatch(
      rx_get_json(
        paste0("/rxcui/", prod_rxcui, "/related"),
        query = list(tty = c("IN", "PIN"))
      ),
      error = function(e) NULL
    )

    gs <- rx_pluck_list(rel_ing, "relatedGroup", "conceptGroup")

    if (is.null(gs) || !length(gs)) {
      return(tibble::tibble(
        product_rxcui = prod_rxcui,
        n_ingredients = NA_integer_,
        contains = FALSE
      ))
    }

    ings <- purrr::map_dfr(gs, function(grp) {
      cps <- rx_pluck_list(grp, "conceptProperties")

      if (is.null(cps)) {
        return(tibble::tibble())
      }

      purrr::map_dfr(cps, function(z) {
        tibble::tibble(
          rxcui = nz1(z$rxcui),
          tty = nz1(z$tty),
          name = tolower(nz1(z$name))
        )
      })
    }) |>
      dplyr::filter(.data$tty %in% c("IN", "PIN")) |>
      dplyr::filter(!is.na(.data$rxcui) | !is.na(.data$name)) |>
      dplyr::mutate(
        ingredient_key = dplyr::coalesce(
          .data$rxcui,
          .data$name
        )
      ) |>
      dplyr::filter(!is.na(.data$ingredient_key), nzchar(.data$ingredient_key)) |>
      dplyr::distinct(.data$tty, .data$ingredient_key, .keep_all = TRUE)

    n_in <- ings |>
      dplyr::filter(.data$tty == "IN") |>
      dplyr::distinct(.data$ingredient_key) |>
      nrow()

    n_pin <- ings |>
      dplyr::filter(.data$tty == "PIN") |>
      dplyr::distinct(.data$ingredient_key) |>
      nrow()

    n_total <- if (n_in > 0L) n_in else n_pin

    if (n_total == 0L) {
      n_total <- NA_integer_
    }

    contains <- any(ings$rxcui %in% cui_ok, na.rm = TRUE) ||
      any(grepl(name_pat, ings$name, perl = TRUE), na.rm = TRUE)

    tibble::tibble(
      product_rxcui = prod_rxcui,
      n_ingredients = as.integer(n_total),
      contains = contains
    )
  }

  verify_contains_historical <- function(prod_rxcui, cui_ok, name_pat) {
    hs <- tryCatch(
      rx_get_json(paste0("/rxcui/", prod_rxcui, "/historystatus")),
      error = function(e) NULL
    )

    x <- rx_pluck_list(hs, "rxcuiStatusHistory")

    if (is.null(x) || !length(x)) {
      return(tibble::tibble(
        product_rxcui = prod_rxcui,
        n_ingredients = NA_integer_,
        contains = FALSE,
        concept_status = NA_character_,
        active_start_date = NA_character_,
        active_end_date = NA_character_,
        release_start_date = NA_character_,
        release_end_date = NA_character_
      ))
    }

    meta <- x$metaData
    attrs <- x$attributes
    defs <- x$definitionalFeatures
    derived <- x$derivedConcepts

    empty_hist_ings <- function() {
      tibble::tibble(
        rxcui = character(),
        name = character()
      )
    }

    ing_strength <- as_rx_records(
      defs$ingredientAndStrength,
      id_field = "baseRxcui"
    )

    # Canonical fallback ingredients from ingredientAndStrength.
    # These may contribute to n_ingredients.
    strength_base_ings <- if (!length(ing_strength)) {
      empty_hist_ings()
    } else {
      purrr::map_dfr(ing_strength, function(z) {
        tibble::tibble(
          rxcui = nz1(z$baseRxcui),
          name = tolower(nz1(z$baseName))
        )
      })
    }

    # Broader ingredient-related concepts from ingredientAndStrength.
    # These are useful for confirming containment, but are not counted
    # as separate ingredients.
    strength_match_ings <- if (!length(ing_strength)) {
      empty_hist_ings()
    } else {
      purrr::map_dfr(ing_strength, function(z) {
        tibble::tibble(
          rxcui = c(
            nz1(z$baseRxcui),
            nz1(z$bossRxcui),
            nz1(z$activeIngredientRxcui),
            nz1(z$moietyRxcui)
          ),
          name = tolower(c(
            nz1(z$baseName),
            nz1(z$bossName),
            nz1(z$activeIngredientName),
            nz1(z$moietyName)
          ))
        )
      })
    }

    ing_concepts <- as_rx_records(
      derived$ingredientConcept,
      id_field = "ingredientRxcui"
    )

    # Preferred canonical ingredient source.
    derived_ings <- if (!length(ing_concepts)) {
      empty_hist_ings()
    } else {
      purrr::map_dfr(ing_concepts, function(z) {
        tibble::tibble(
          rxcui = nz1(z$ingredientRxcui),
          name = tolower(nz1(z$ingredientName))
        )
      })
    }

    # Count canonical ingredients only.
    # Prefer derived ingredientConcept when available; fall back to base
    # ingredientAndStrength concepts otherwise.
    count_ings <- if (nrow(derived_ings)) {
      derived_ings
    } else {
      strength_base_ings
    }

    count_ings <- count_ings |>
      dplyr::filter(!is.na(.data$rxcui) | !is.na(.data$name)) |>
      dplyr::mutate(
        ingredient_key = dplyr::coalesce(
          .data$rxcui,
          .data$name
        )
      ) |>
      dplyr::filter(!is.na(.data$ingredient_key), nzchar(.data$ingredient_key)) |>
      dplyr::distinct(.data$ingredient_key, .keep_all = TRUE)

    # Use the broader table only for containment verification.
    match_ings <- dplyr::bind_rows(
      derived_ings,
      strength_base_ings,
      strength_match_ings
    ) |>
      dplyr::filter(!is.na(.data$rxcui) | !is.na(.data$name)) |>
      dplyr::distinct()

    n_total <- nrow(count_ings)

    if (n_total == 0L) {
      n_total <- NA_integer_
    }

    contains <- any(match_ings$rxcui %in% cui_ok, na.rm = TRUE) ||
      any(grepl(name_pat, match_ings$name, perl = TRUE), na.rm = TRUE)

    tibble::tibble(
      product_rxcui = prod_rxcui,
      n_ingredients = as.integer(n_total),
      contains = contains,
      concept_status = nz1(meta$status),
      active_start_date = nz1(meta$activeStartDate),
      active_end_date = nz1(meta$activeEndDate),
      release_start_date = nz1(meta$releaseStartDate),
      release_end_date = nz1(meta$releaseEndDate)
    )
  }

  allstatus_products <- if (isTRUE(include_historical)) {
    fetch_allstatus_products()
  } else {
    tibble::tibble(
      product_rxcui = character(),
      name = character(),
      tty = character(),
      concept_status = character()
    )
  }

  out <- .rxref_progress_map_dfr(
    ingredient_ids,
    function(ing) {
      ## orig.
      acc <- acceptance_for_ing(ing)
      cand_rela <- fetch_via_rela(ing)
      cand_related <- fetch_via_related(ing)
      cand_allrel <- fetch_via_allrelated(ing)
      cand_drugs <- fetch_via_drugs_name(ing)

      hist_prods <- if (isTRUE(include_historical)) {
        filter_allstatus_name(allstatus_products, acc$name_pat)
      } else {
        tibble::tibble(
          product_rxcui = character(),
          name = character(),
          tty = character(),
          concept_status = character()
        )
      }

      active_prods <- dplyr::bind_rows(
        cand_rela,
        cand_related,
        cand_allrel,
        cand_drugs
      ) |>
        dplyr::distinct()

      prods <- dplyr::bind_rows(
        dplyr::mutate(active_prods, concept_status = "Active"),
        hist_prods
      ) |>
        dplyr::distinct()

      if (!nrow(prods)) {
        return(empty_products(
          include_history_cols = isTRUE(include_historical)
        ))
      }

      active_ids <- prods |>
        dplyr::filter(.data$concept_status == "Active") |>
        dplyr::pull(.data$product_rxcui) |>
        unique()

      historical_ids <- prods |>
        dplyr::filter(.data$concept_status != "Active") |>
        dplyr::pull(.data$product_rxcui) |>
        unique()

      chk_active <- purrr::map_dfr(
        active_ids,
        verify_contains,
        cui_ok = acc$cui_ok,
        name_pat = acc$name_pat
      ) |>
        dplyr::mutate(
          concept_status = "Active",
          active_start_date = NA_character_,
          active_end_date = NA_character_,
          release_start_date = NA_character_,
          release_end_date = NA_character_
        )

      chk_historical <- purrr::map_dfr(
        historical_ids,
        verify_contains_historical,
        cui_ok = acc$cui_ok,
        name_pat = acc$name_pat
      )

      chk <- dplyr::bind_rows(chk_active, chk_historical)

      out_ <- prods |>
        dplyr::select(-dplyr::any_of("concept_status")) |>
        dplyr::left_join(chk, by = "product_rxcui") |>
        dplyr::filter(.data$contains %in% TRUE) |>
        dplyr::mutate(ingredient_rxcui = ing)

      if (isTRUE(include_historical)) {
        out_ <- out_ |>
          dplyr::select(dplyr::all_of(c(
            "ingredient_rxcui",
            "product_rxcui",
            "name",
            "tty",
            "n_ingredients",
            "concept_status",
            "active_start_date",
            "active_end_date",
            "release_start_date",
            "release_end_date"
          )))
      } else {
        out_ <- out_ |>
          dplyr::select(dplyr::all_of(c(
            "ingredient_rxcui",
            "product_rxcui",
            "name",
            "tty",
            "n_ingredients"
          )))
      }

      if (!isTRUE(include_combos)) {
        out_ <- dplyr::filter(out_, .data$n_ingredients <= 1L)
      }

      out_
    },
    name = "Matching products to ingredients",
    show_progress = show_progress
  ) |>
    dplyr::distinct()

  if (!is.null(route)) {
    out <- filter_products_by_route(
      out,
      route = route,
      include_historical = isTRUE(include_historical),
      show_progress = show_progress
    )
  }

  out
}


#' Search free-text drug name and return product CUIs and/or NDCs
#'
#' High-level convenience: free text -> ingredient(s) (IN/PIN) -> verified product
#' CUIs, and optionally expand to NDCs with status filtering.
#'
#' @param term Character vector; free-text drug names.
#' @param return One of `c("rxcui","ndc","both")`. Note that `"both"` will return a list
#'  with both an rxcui tibble and an ndc tibble.
#' @param ndc_status Optional character vector to filter NDCs. Options are "ACTIVE",
#'  "OBSOLETE", "UNSPECIFIED" (the API may also return no value, which will appear as NA).
#' @param ttys Character vector of TTYs to include in product search.
#'   Defaults to `.rxref_default_ttys`. Other prespecified option is `.rxref_extended_ttys`
#'   or a character vector of explicit TTYs. Run `tty_catalogue()` to review options.
#' @param route Optional character vector of routes to retain before returning
#'   products or mapping to NDCs. If `NULL`, no route filtering is performed.
#'   Common values include `"ORAL"`, `"INJECTION"`, `"OPHTHALMIC"`,
#'   `"INHALATION"`, and `"TOPICAL"`.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#' @param ... Passed to `products_for_ingredients()` (e.g., include_combos = FALSE)
#' @return If `return="rxcui"`: tibble of products.
#'   If `"ndc"`: tibble of NDCs with `ingredient_rxcui`, `product_rxcui`, `ndc11`, `ndc_status`.
#'   If `"both"`: list(products=…, ndcs=…).
#' @export
search_drug <- function(term,
                        return = c("rxcui","ndc","both"),
                        ndc_status = NULL,
                        ttys = .rxref_default_ttys,
                        route = NULL,
                        show_progress = interactive(),
                        ...) {
  return <- match.arg(return)

  ings <- find_ingredients(
    term,
    show_progress = show_progress
  )
  ing_ids <- unique(stats::na.omit(ings$rxcui))

  empty_products <- tibble::tibble(
    ingredient_rxcui = character(),
    product_rxcui    = character(),
    name             = character(),
    tty              = character(),
    n_ingredients    = integer()
  )
  empty_ndc <- tibble::tibble(
    ingredient_rxcui = character(),
    product_rxcui    = character(),
    ndc11            = character(),
    ndc_status       = character()
  )

  if (!length(ing_ids)) {
    if (return == "rxcui") return(empty_products)
    if (return == "ndc")   return(empty_ndc)
    return(list(products = empty_products, ndcs = empty_ndc))
  }

  prods <- products_for_ingredients(
    ing_ids,
    ttys = ttys,
    route = route,
    show_progress = show_progress,
    ...
  )
  if (return == "rxcui") return(prods)

  # Only product-ish TTYs cleanly map to NDCs
  ndc_map_ttys <- c("SCD","SBD","GPCK","BPCK")
  prods_for_ndc <- dplyr::filter(prods, .data$tty %in% ndc_map_ttys)

  prod_ids <- unique(stats::na.omit(prods_for_ndc$product_rxcui))
  if (!length(prod_ids)) {
    if (return == "ndc")   return(empty_ndc)
    return(list(products = prods, ndcs = empty_ndc))
  }

  # Map each product rxcui to NDCs; rename rxcui -> product_rxcui for join
  ndcs <- map_rxcui_to_ndc(
    prod_ids,
    status = ndc_status,
    show_progress = show_progress
  ) |>
    dplyr::filter(!is.na(.data$ndc11)) |>
    dplyr::mutate(product_rxcui = .data$rxcui) |>
    dplyr::select(.data$product_rxcui, .data$ndc11, dplyr::any_of("ndc_status")) |>
    dplyr::left_join(
      dplyr::select(prods_for_ndc, .data$product_rxcui, .data$ingredient_rxcui),
      by = "product_rxcui"
    ) |>
    dplyr::relocate(.data$ingredient_rxcui, .before = 1L) |>
    dplyr::distinct()

  if (return == "ndc") return(ndcs)
  list(products = prods, ndcs = ndcs)
}


#' Get ingredient concepts for RxCUIs
#'
#' Maps one or more RxCUIs to related ingredient concepts, returning ingredient
#' RxCUIs, names, and term types. This is useful when the input is already a
#' specific RxNorm product concept rather than a free-text drug name.
#'
#' @param rxcui Character vector of RxCUIs.
#' @param include_pin Logical. Include precise ingredient concepts (`PIN`).
#' @param include_min Logical. Include multiple ingredient concepts (`MIN`).
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'
#' @return A tibble with columns `rxcui`, `ingredient_rxcui`,
#'   `ingredient_name`, and `ingredient_tty`.
#'
#' @export
ingredients_for_rxcui <- function(
    rxcui,
    include_pin = TRUE,
    include_min = FALSE,
    show_progress = interactive()
) {
  stopifnot(is.character(rxcui))

  rxcui_ids <- unique(stats::na.omit(rxcui))

  .rxref_progress_map_dfr(
    rxcui_ids,
    function(x) {
      .rxref_get_ingredients_for_rxcui(
        x,
        include_pin = include_pin,
        include_min = include_min
      )
    },
    name = "Finding ingredients",
    show_progress = show_progress
  ) |>
    dplyr::rename(rxcui = "related_rxcui")
}

#' Normalize RxNav records to a list of records
#'
#' RxNav sometimes returns repeated elements as a list of lists, but when there
#' is only one result it may return a single named list. This helper normalizes
#' both cases to a list of records.
#'
#' @keywords internal
#' @noRd
as_rx_records <- function(x, id_field = "rxcui") {
  if (is.null(x) || !length(x)) {
    return(list())
  }

  if (is.data.frame(x)) {
    return(split(x, seq_len(nrow(x))))
  }

  if (is.list(x) && !is.null(x[[id_field]])) {
    return(list(x))
  }

  x
}


#' Empty products tibble
#'
#' @keywords internal
#' @noRd
empty_products <- function(include_history_cols = FALSE) {
  out <- tibble::tibble(
    ingredient_rxcui = character(),
    product_rxcui = character(),
    name = character(),
    tty = character(),
    n_ingredients = integer()
  )

  if (isTRUE(include_history_cols)) {
    out <- dplyr::mutate(
      out,
      concept_status = character(),
      active_start_date = character(),
      active_end_date = character(),
      release_start_date = character(),
      release_end_date = character()
    )
  }

  out
}

test_that("products_for_ingredients can exclude combination products", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))

  ing <- find_ingredients("lisinopril") |>
    dplyr::filter(tty == "IN")

  single_only <- products_for_ingredients(
    ing$rxcui,
    ttys = c("SCD", "SBD"),
    include_combos = FALSE
  )

  expect_s3_class(single_only, "tbl_df")
  expect_true(nrow(single_only) > 0)
  expect_true(all(single_only$n_ingredients <= 1))
})

test_that("products_for_ingredients includes combination products when requested", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))

  ing <- find_ingredients("lisinopril") |>
    dplyr::filter(tty == "IN")

  with_combos <- products_for_ingredients(
    ing$rxcui,
    ttys = c("SCD", "SBD"),
    include_combos = TRUE
  )

  expect_true(any(with_combos$n_ingredients > 1))
  expect_true(any(grepl("hydrochlorothiazide", with_combos$name, ignore.case = TRUE)))
})

test_that("filter_products_by_route requires product_rxcui", {
  bad <- tibble::tibble(rxcui = "123")

  expect_error(
    filter_products_by_route(bad, route = "ORAL"),
    "product_rxcui"
  )
})


test_that("filter_products_by_route keeps oral metoprolol products", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))

  ing <- find_ingredients("metoprolol") |>
    dplyr::filter(tty == "IN")

  prods <- products_for_ingredients(
    ing$rxcui,
    include_combos = TRUE
  )

  oral <- filter_products_by_route(prods, route = "ORAL")

  expect_true(nrow(oral) > 0)
  expect_true(all(grepl("Oral|Tablet|Capsule|Solution", oral$name, ignore.case = TRUE)))
})

test_that("search_drug can filter products by route", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))

  res <- search_drug(
    "timolol",
    return = "rxcui",
    route = "ORAL",
    include_combos = TRUE
  )

  expect_true(is.data.frame(res))

  if (nrow(res) > 0) {
    expect_false(any(grepl("ophthalmic", res$name, ignore.case = TRUE)))
  }
})

test_that("filter_products_by_route filters using mocked clinical route attributes", {
  fake_get_clinical_attributes <- function(rxcui, ...) {
    expect_equal(sort(rxcui), c("10", "11", "12"))

    tibble::tibble(
      rxcui = c("10", "11", "12"),
      dose_form = c(
        "Oral Tablet",
        "Ophthalmic Solution",
        "Injectable Solution"
      ),
      dose_form_group = c(
        "PILL",
        "OPHTHALMIC",
        "INJECTION"
      ),
      route = c(
        "ORAL",
        "OPHTHALMIC",
        "INJECTION"
      )
    )
  }

  testthat::local_mocked_bindings(
    get_clinical_attributes = fake_get_clinical_attributes,
    .package = "rxref"
  )

  x <- tibble::tibble(
    ingredient_rxcui = c("1", "1", "1"),
    product_rxcui = c("10", "11", "12"),
    name = c(
      "timolol 10 MG Oral Tablet",
      "timolol 0.5 % Ophthalmic Solution",
      "timolol 5 MG/ML Injectable Solution"
    ),
    tty = c("SCD", "SCD", "SCD"),
    n_ingredients = c(1L, 1L, 1L)
  )

  out <- filter_products_by_route(x, route = "ORAL")

  expect_equal(nrow(out), 1)
  expect_equal(out$product_rxcui, "10")
  expect_equal(out$route, "ORAL")
  expect_equal(out$dose_form, "Oral Tablet")
  expect_equal(out$dose_form_group, "PILL")
})

test_that("filter_products_by_route can drop appended route columns", {
  fake_get_clinical_attributes <- function(rxcui, ...) {
    tibble::tibble(
      rxcui = c("10", "11"),
      dose_form = c("Oral Tablet", "Ophthalmic Solution"),
      dose_form_group = c("PILL", "OPHTHALMIC"),
      route = c("ORAL", "OPHTHALMIC")
    )
  }

  testthat::local_mocked_bindings(
    get_clinical_attributes = fake_get_clinical_attributes,
    .package = "rxref"
  )

  x <- tibble::tibble(
    ingredient_rxcui = c("1", "1"),
    product_rxcui = c("10", "11"),
    name = c(
      "timolol 10 MG Oral Tablet",
      "timolol 0.5 % Ophthalmic Solution"
    ),
    tty = c("SCD", "SCD"),
    n_ingredients = c(1L, 1L)
  )

  out <- filter_products_by_route(
    x,
    route = "ORAL",
    keep_route_info = FALSE
  )

  expect_equal(nrow(out), 1)
  expect_equal(out$product_rxcui, "10")
  expect_false("route" %in% names(out))
  expect_false("dose_form" %in% names(out))
  expect_false("dose_form_group" %in% names(out))
})

test_that("estradiol clinical attributes can be unnested", {
  skip_if_not(identical(Sys.getenv("RXREF_ONLINE"), "1"))

  a1 <- find_ingredients("estradiol")
  a2 <- products_for_ingredients(a1$rxcui)
  a3 <- get_clinical_attributes(a2$product_rxcui)

  expect_s3_class(a3, "tbl_df")
  expect_true(all(c(
    "ingredient_rxcui",
    "ingredient_name",
    "ingredient_tty"
  ) %in% names(a3)))

  expect_no_error(
    tidyr::unnest(a3, c(ingredient_rxcui, ingredient_name, ingredient_tty))
  )
})

test_that("resolve() resolves a drug name using mocked approximateTerm and properties", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/spellingsuggestions") {
      return(list())
    }

    if (path == "/approximateTerm") {
      expect_equal(query$term, "lisinopril")
      expect_equal(query$maxEntries, 1)

      return(list(
        approximateGroup = list(
          candidate = list(
            list(
              rxcui = "29046",
              name = "lisinopril",
              score = "100"
            )
          )
        )
      ))
    }

    if (path == "/rxcui/29046/properties") {
      return(list(
        properties = list(
          rxcui = "29046",
          name = "lisinopril",
          tty = "IN"
        )
      ))
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- resolve("lisinopril", type = "name", max_entries = 1)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$input, "lisinopril")
  expect_equal(out$type, "name")
  expect_equal(out$rxcui, "29046")
  expect_equal(out$name, "lisinopril")
  expect_equal(out$tty, "IN")
  expect_equal(out$score, 100)
  expect_equal(out$matched_term, "lisinopril")
})


test_that("resolve() resolves an NDC using mocked /rxcui and properties", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/rxcui") {
      expect_equal(query$idtype, "NDC")
      expect_equal(query$id, "00002080001")

      return(list(
        idGroup = list(
          rxnormId = list("860975")
        )
      ))
    }

    if (path == "/rxcui/860975/properties") {
      return(list(
        properties = list(
          rxcui = "860975",
          name = "metformin hydrochloride 500 MG Oral Tablet",
          tty = "SCD"
        )
      ))
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- resolve("0002-0800-01", type = "ndc")

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$input, "0002-0800-01")
  expect_equal(out$type, "ndc")
  expect_equal(out$ndc11, "00002080001")
  expect_equal(out$rxcui, "860975")
  expect_equal(out$tty, "SCD")
})


test_that("resolve() resolves an RxCUI using mocked properties", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/rxcui/29046/properties") {
      return(list(
        properties = list(
          rxcui = "29046",
          name = "lisinopril",
          tty = "IN"
        )
      ))
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- resolve("29046", type = "rxcui")

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$input, "29046")
  expect_equal(out$type, "rxcui")
  expect_equal(out$rxcui, "29046")
  expect_equal(out$name, "lisinopril")
  expect_equal(out$tty, "IN")
})


test_that("find_ingredients() keeps IN and PIN concepts by default", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/approximateTerm") {
      expect_equal(query$term, "lisinopril")
      expect_equal(query$maxEntries, 10)

      return(list(
        approximateGroup = list(
          candidate = list(
            list(rxcui = "29046", name = "lisinopril", score = "100"),
            list(rxcui = "12345", name = "lisinopril sodium", score = "90"),
            list(rxcui = "99999", name = "not an ingredient", score = "80")
          )
        )
      ))
    }

    if (path == "/rxcui/29046/properties") {
      return(list(
        properties = list(
          rxcui = "29046",
          name = "lisinopril",
          tty = "IN"
        )
      ))
    }

    if (path == "/rxcui/12345/properties") {
      return(list(
        properties = list(
          rxcui = "12345",
          name = "lisinopril sodium",
          tty = "PIN"
        )
      ))
    }

    if (path == "/rxcui/99999/properties") {
      return(list(
        properties = list(
          rxcui = "99999",
          name = "lisinopril 10 MG Oral Tablet",
          tty = "SCD"
        )
      ))
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- find_ingredients("lisinopril")

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 2)
  expect_equal(out$input, c("lisinopril", "lisinopril"))
  expect_equal(out$rxcui, c("29046", "12345"))
  expect_equal(out$tty, c("IN", "PIN"))
  expect_equal(out$score, c(100, 90))
})


test_that("find_ingredients() excludes PIN concepts when include_pin = FALSE", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/approximateTerm") {
      return(list(
        approximateGroup = list(
          candidate = list(
            list(rxcui = "29046", name = "lisinopril", score = "100"),
            list(rxcui = "12345", name = "lisinopril sodium", score = "90")
          )
        )
      ))
    }

    if (path == "/rxcui/29046/properties") {
      return(list(
        properties = list(
          rxcui = "29046",
          name = "lisinopril",
          tty = "IN"
        )
      ))
    }

    if (path == "/rxcui/12345/properties") {
      return(list(
        properties = list(
          rxcui = "12345",
          name = "lisinopril sodium",
          tty = "PIN"
        )
      ))
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- find_ingredients("lisinopril", include_pin = FALSE)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$rxcui, "29046")
  expect_equal(out$tty, "IN")
})


test_that("find_ingredients() returns empty tibble when no candidates are found", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/approximateTerm") {
      return(list(
        approximateGroup = list(
          candidate = NULL
        )
      ))
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- find_ingredients("definitely not a drug")

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0L)
  expect_named(out, c("input", "rxcui", "name", "tty", "score"))
})


test_that("map_ndc_to_rxcui() maps NDC to RxCUI", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/rxcui") {
      expect_equal(query$idtype, "NDC")
      expect_equal(query$id, "00002080001")

      return(list(
        idGroup = list(
          rxnormId = list("860975")
        )
      ))
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- map_ndc_to_rxcui("0002-0800-01", show_progress = FALSE)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$input, "0002-0800-01")
  expect_equal(out$ndc11, "00002080001")
  expect_equal(out$rxcui, "860975")
})


test_that("map_ndc_to_rxcui() returns NA RxCUI when no mapping exists", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/rxcui") {
      return(list(
        idGroup = list(
          rxnormId = NULL
        )
      ))
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- map_ndc_to_rxcui("0002-0800-01", show_progress = FALSE)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$input, "0002-0800-01")
  expect_equal(out$ndc11, "00002080001")
  expect_true(is.na(out$rxcui))
})


test_that("map_rxcui_to_ndc() maps RxCUI to NDCs and retrieves NDC status", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/rxcui/860975/ndcs") {
      return(list(
        ndcGroup = list(
          ndcList = list(
            ndc = list("00002080001", "00002080002")
          )
        )
      ))
    }

    if (path == "/ndcstatus") {
      if (query$ndc == "00002-0800-01") {
        return(list(
          ndcStatus = list(
            status = "ACTIVE"
          )
        ))
      }

      if (query$ndc == "00002-0800-02") {
        return(list(
          ndcStatus = list(
            status = "OBSOLETE"
          )
        ))
      }
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- map_rxcui_to_ndc("860975", show_progress = FALSE)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 2)
  expect_equal(out$rxcui, c("860975", "860975"))
  expect_equal(out$ndc11, c("00002080001", "00002080002"))
  expect_equal(out$ndc_status, c("ACTIVE", "OBSOLETE"))
})


test_that("map_rxcui_to_ndc() filters by requested NDC status", {
  fake_rx_get_json <- function(path, query = list()) {
    if (path == "/rxcui/860975/ndcs") {
      return(list(
        ndcGroup = list(
          ndcList = list(
            ndc = list("00002080001", "00002080002")
          )
        )
      ))
    }

    if (path == "/ndcstatus") {
      if (query$ndc == "00002-0800-01") {
        return(list(
          ndcStatus = list(
            status = "ACTIVE"
          )
        ))
      }

      if (query$ndc == "00002-0800-02") {
        return(list(
          ndcStatus = list(
            status = "OBSOLETE"
          )
        ))
      }
    }

    stop("Unexpected path: ", path)
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- map_rxcui_to_ndc("860975", status = "ACTIVE", show_progress = FALSE)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$rxcui, "860975")
  expect_equal(out$ndc11, "00002080001")
  expect_equal(out$ndc_status, "ACTIVE")
})


test_that("get_classes() queries RxClass by RxCUI and parses class assertions", {
  fake_rxclass_get_json <- function(path, query = list()) {
    expect_equal(path, "/class/byRxcui")
    expect_equal(query$rxcui, "29046")
    expect_equal(query$relaSource, "ATC")

    return(list(
      rxclassDrugInfoList = list(
        rxclassDrugInfo = list(
          list(
            minConcept = list(
              rxcui = "29046",
              name = "lisinopril",
              tty = "IN"
            ),
            rxclassMinConceptItem = list(
              classId = "C09AA",
              className = "ACE inhibitors, plain",
              classType = "ATC4",
              classUrl = "https://example.org/C09AA"
            ),
            rela = "has_ATC",
            relaSource = "ATC"
          )
        )
      )
    ))
  }

  testthat::local_mocked_bindings(
    rxclass_get_json = fake_rxclass_get_json,
    .package = "rxref"
  )

  withr::local_options(lifecycle_verbosity = "quiet")

  out <- get_classes(
    "29046",
    by = "rxcui",
    rela_source = "ATC",
    keep_input = TRUE
  )

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$input, "29046")
  expect_equal(out$rxcui, "29046")
  expect_equal(out$drug_name, "lisinopril")
  expect_equal(out$drug_tty, "IN")
  expect_equal(out$class_id, "C09AA")
  expect_equal(out$class_name, "ACE inhibitors, plain")
  expect_equal(out$class_type, "ATC4")
  expect_equal(out$rela, "has_ATC")
  expect_equal(out$rela_source, "ATC")
})


test_that("get_classes() queries RxClass by drug name", {
  fake_rxclass_get_json <- function(path, query = list()) {
    expect_equal(path, "/class/byDrugName")
    expect_equal(query$drugName, "lisinopril")
    expect_equal(query$relaSource, "ATC")

    return(list(
      rxclassDrugInfoList = list(
        rxclassDrugInfo = list(
          list(
            minConcept = list(
              rxcui = "29046",
              name = "lisinopril",
              tty = "IN"
            ),
            rxclassMinConceptItem = list(
              classId = "C09AA",
              className = "ACE inhibitors, plain",
              classType = "ATC4",
              classUrl = "https://example.org/C09AA"
            ),
            rela = "has_ATC",
            relaSource = "ATC"
          )
        )
      )
    ))
  }

  testthat::local_mocked_bindings(
    rxclass_get_json = fake_rxclass_get_json,
    .package = "rxref"
  )

  withr::local_options(lifecycle_verbosity = "quiet")

  out <- get_classes(
    "lisinopril",
    by = "name",
    rela_source = "ATC",
    keep_input = TRUE
  )

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$input, "lisinopril")
  expect_equal(out$rxcui, "29046")
  expect_equal(out$class_id, "C09AA")
})


test_that("get_classes() can filter class types", {
  fake_rxclass_get_json <- function(path, query = list()) {
    return(list(
      rxclassDrugInfoList = list(
        rxclassDrugInfo = list(
          list(
            minConcept = list(
              rxcui = "29046",
              name = "lisinopril",
              tty = "IN"
            ),
            rxclassMinConceptItem = list(
              classId = "C09AA",
              className = "ACE inhibitors, plain",
              classType = "ATC4",
              classUrl = "https://example.org/C09AA"
            ),
            rela = "has_ATC",
            relaSource = "ATC"
          ),
          list(
            minConcept = list(
              rxcui = "29046",
              name = "lisinopril",
              tty = "IN"
            ),
            rxclassMinConceptItem = list(
              classId = "C000000",
              className = "Some non-ATC class",
              classType = "EPC",
              classUrl = "https://example.org/EPC"
            ),
            rela = "has_EPC",
            relaSource = "FDASPL"
          )
        )
      )
    ))
  }

  testthat::local_mocked_bindings(
    rxclass_get_json = fake_rxclass_get_json,
    .package = "rxref"
  )

  withr::local_options(lifecycle_verbosity = "quiet")

  out <- get_classes(
    "29046",
    by = "rxcui",
    class_types = "ATC4",
    keep_input = FALSE
  )

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$class_id, "C09AA")
  expect_equal(out$class_type, "ATC4")
  expect_false("input" %in% names(out))
})


test_that("find_classes() searches class concepts by name", {
  fake_rxclass_get_json <- function(path, query = list()) {
    expect_equal(path, "/class/byName")
    expect_equal(query$className, "ACE inhibitors")

    return(list(
      rxclassMinConceptList = list(
        rxclassMinConcept = list(
          list(
            classId = "C09A",
            className = "ACE inhibitors",
            classType = "ATC3",
            classUrl = "https://example.org/C09A"
          ),
          list(
            classId = "C09AA",
            className = "ACE inhibitors, plain",
            classType = "ATC4",
            classUrl = "https://example.org/C09AA"
          )
        )
      )
    ))
  }

  testthat::local_mocked_bindings(
    rxclass_get_json = fake_rxclass_get_json,
    .package = "rxref"
  )

  out <- find_classes("ACE inhibitors")

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 2)
  expect_equal(out$class_id, c("C09A", "C09AA"))
  expect_equal(out$class_name, c("ACE inhibitors", "ACE inhibitors, plain"))
  expect_equal(out$class_type, c("ATC3", "ATC4"))
})


test_that("find_classes() passes class_types filter to RxClass", {
  fake_rxclass_get_json <- function(path, query = list()) {
    expect_equal(path, "/class/byName")
    expect_equal(query$className, "ACE inhibitors")
    expect_equal(query$classTypes, "ATC4 EPC")

    return(list(
      rxclassMinConceptList = list(
        rxclassMinConcept = list(
          list(
            classId = "C09AA",
            className = "ACE inhibitors, plain",
            classType = "ATC4",
            classUrl = "https://example.org/C09AA"
          )
        )
      )
    ))
  }

  testthat::local_mocked_bindings(
    rxclass_get_json = fake_rxclass_get_json,
    .package = "rxref"
  )

  out <- find_classes("ACE inhibitors", class_types = c("ATC4", "EPC"))

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$class_id, "C09AA")
  expect_equal(out$class_type, "ATC4")
})


test_that("find_classes() returns empty tibble when no class concepts are found", {
  fake_rxclass_get_json <- function(path, query = list()) {
    return(list(
      rxclassMinConceptList = list(
        rxclassMinConcept = NULL
      )
    ))
  }

  testthat::local_mocked_bindings(
    rxclass_get_json = fake_rxclass_get_json,
    .package = "rxref"
  )

  out <- find_classes("not a real class")

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0)
  expect_named(out, c("class_id", "class_name", "class_type", "class_url"))
})


test_that("get_class_members() queries RxClass members and parses node attributes", {
  fake_rxclass_get_json <- function(path, query = list()) {
    expect_equal(path, "/classMembers")
    expect_equal(query$classId, "C09AA")
    expect_equal(query$relaSource, "ATC")
    expect_equal(query$trans, "0")
    expect_equal(query$ttys, "IN PIN")

    return(list(
      drugMemberGroup = list(
        drugMember = list(
          list(
            minConcept = list(
              rxcui = "29046",
              name = "lisinopril",
              tty = "IN"
            ),
            nodeAttr = list(
              list(attrName = "SourceId", attrValue = "C09AA03"),
              list(attrName = "SourceName", attrValue = "lisinopril"),
              list(attrName = "SourceUrl", attrValue = "https://example.org/C09AA03"),
              list(attrName = "Relation", attrValue = "direct")
            )
          )
        )
      )
    ))
  }

  testthat::local_mocked_bindings(
    rxclass_get_json = fake_rxclass_get_json,
    .package = "rxref"
  )

  out <- get_class_members(
    class_id = "C09AA",
    rela_source = "ATC",
    include_indirect = TRUE,
    ttys = c("IN", "PIN")
  )

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$rxcui, "29046")
  expect_equal(out$name, "lisinopril")
  expect_equal(out$tty, "IN")
  expect_equal(out$class_id, "C09AA")
  expect_equal(out$rela_source, "ATC")
  expect_true(is.na(out$rela))
  expect_equal(out$source_id, "C09AA03")
  expect_equal(out$source_name, "lisinopril")
  expect_equal(out$relation, "direct")
})


test_that("get_class_members() passes rela and direct-only options", {
  fake_rxclass_get_json <- function(path, query = list()) {
    expect_equal(path, "/classMembers")
    expect_equal(query$classId, "C09AA")
    expect_equal(query$relaSource, "ATC")
    expect_equal(query$rela, "has_ATC")
    expect_equal(query$trans, "1")

    return(list(
      drugMemberGroup = list(
        drugMember = list(
          list(
            minConcept = list(
              rxcui = "29046",
              name = "lisinopril",
              tty = "IN"
            ),
            nodeAttr = list()
          )
        )
      )
    ))
  }

  testthat::local_mocked_bindings(
    rxclass_get_json = fake_rxclass_get_json,
    .package = "rxref"
  )

  out <- get_class_members(
    class_id = "C09AA",
    rela_source = "ATC",
    rela = "has_ATC",
    include_indirect = FALSE
  )

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1)
  expect_equal(out$rxcui, "29046")
  expect_equal(out$rela, "has_ATC")
  expect_equal(out$class_id, "C09AA")
})


test_that("get_class_members() returns empty tibble when no members are found", {
  fake_rxclass_get_json <- function(path, query = list()) {
    return(list(
      drugMemberGroup = list(
        drugMember = NULL
      )
    ))
  }

  testthat::local_mocked_bindings(
    rxclass_get_json = fake_rxclass_get_json,
    .package = "rxref"
  )

  out <- get_class_members(
    class_id = "C09AA",
    rela_source = "ATC"
  )

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0)
  expect_named(out, c(
    "rxcui",
    "name",
    "tty",
    "class_id",
    "rela_source",
    "rela",
    "source_id",
    "source_name",
    "source_url",
    "relation"
  ))
})


test_that("ingredients_for_rxcui() wraps ingredient helper and renames related_rxcui to rxcui", {
  fake_get_ingredients <- function(x, include_pin = TRUE, include_min = FALSE) {
    expect_equal(x, "860975")
    expect_true(include_pin)
    expect_false(include_min)

    tibble::tibble(
      related_rxcui = x,
      ingredient_rxcui = c("6809", "12345"),
      ingredient_name = c("metformin", "metformin hydrochloride"),
      ingredient_tty = c("IN", "PIN")
    )
  }

  testthat::local_mocked_bindings(
    .rxref_get_ingredients_for_rxcui = fake_get_ingredients,
    .package = "rxref"
  )

  out <- ingredients_for_rxcui("860975", show_progress = FALSE)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 2)
  expect_equal(out$rxcui, c("860975", "860975"))
  expect_equal(out$ingredient_rxcui, c("6809", "12345"))
  expect_equal(out$ingredient_name, c("metformin", "metformin hydrochloride"))
  expect_equal(out$ingredient_tty, c("IN", "PIN"))
  expect_false("related_rxcui" %in% names(out))
})


test_that("ingredients_for_rxcui() de-duplicates input RxCUIs", {
  calls <- 0L

  fake_get_ingredients <- function(x, include_pin = TRUE, include_min = FALSE) {
    calls <<- calls + 1L

    tibble::tibble(
      related_rxcui = x,
      ingredient_rxcui = "6809",
      ingredient_name = "metformin",
      ingredient_tty = "IN"
    )
  }

  testthat::local_mocked_bindings(
    .rxref_get_ingredients_for_rxcui = fake_get_ingredients,
    .package = "rxref"
  )

  out <- ingredients_for_rxcui(c("860975", "860975"), show_progress = FALSE)

  expect_equal(calls, 1L)
  expect_equal(nrow(out), 1)
  expect_equal(out$rxcui, "860975")
  expect_equal(out$ingredient_rxcui, "6809")
})

# test_that("products_for_ingredients() parses products from allrelated fallback", {
#   fake_rx_get_json <- function(path, query = list()) {
#     # fetch_via_rela(): return no products from /related + rela = ingredient_of
#     if (path == "/rxcui/29046/related" &&
#         identical(query$rela, "ingredient_of")) {
#       return(list(
#         relatedGroup = list(
#           conceptGroup = NULL
#         )
#       ))
#     }
#
#     # fetch_via_related(): return no products from /related without rela
#     if (path == "/rxcui/29046/related" &&
#         is.null(query$rela)) {
#       return(list(
#         allRelatedGroup = list(
#           conceptGroup = NULL
#         )
#       ))
#     }
#
#     # fetch_via_allrelated(): this is the only source that returns a product
#     if (path == "/rxcui/29046/allrelated") {
#       return(list(
#         allRelatedGroup = list(
#           conceptGroup = list(
#             list(
#               tty = "SCD",
#               conceptProperties = list(
#                 list(
#                   rxcui = "11111",
#                   name = "lisinopril 10 MG Oral Tablet",
#                   tty = "SCD"
#                 )
#               )
#             )
#           )
#         )
#       ))
#     }
#
#     # fetch_via_drugs_name() and acceptance_for_ing() both use properties
#     if (path == "/rxcui/29046/properties") {
#       return(list(
#         properties = list(
#           rxcui = "29046",
#           name = "lisinopril",
#           tty = "IN"
#         )
#       ))
#     }
#
#     # fetch_via_drugs_name(): return no products from /drugs
#     if (path == "/drugs") {
#       return(list(
#         drugGroup = list(
#           conceptGroup = NULL
#         )
#       ))
#     }
#
#     # acceptance_for_ing(): no PIN children
#     if (path == "/rxcui/29046/related" &&
#         identical(query$rela, "has_precise_ingredient")) {
#       return(list(
#         relatedGroup = list(
#           conceptGroup = NULL
#         )
#       ))
#     }
#
#     # verify_contains(): product must verify that it contains the ingredient
#     if (path == "/rxcui/11111/related") {
#       return(list(
#         relatedGroup = list(
#           conceptGroup = list(
#             list(
#               tty = "IN",
#               conceptProperties = list(
#                 list(
#                   rxcui = "29046",
#                   name = "lisinopril",
#                   tty = "IN"
#                 )
#               )
#             )
#           )
#         )
#       ))
#     }
#
#     stop("Unexpected path: ", path)
#   }
#
#   testthat::local_mocked_bindings(
#     rx_get_json = fake_rx_get_json,
#     .package = "rxref"
#   )
#
#   out <- products_for_ingredients(
#     "29046",
#     ttys = "SCD",
#     include_combos = TRUE,
#     show_progress = FALSE
#   )
#
#   expect_s3_class(out, "tbl_df")
#   expect_equal(nrow(out), 1L)
#   expect_equal(out$ingredient_rxcui, "29046")
#   expect_equal(out$product_rxcui, "11111")
#   expect_equal(out$name, "lisinopril 10 MG Oral Tablet")
#   expect_equal(out$tty, "SCD")
#   expect_equal(out$n_ingredients, 1L)
# })


test_that("get_properties() returns NA fields when properties are missing", {
  fake_rx_get_json <- function(path, query = list()) {
    list()
  }

  testthat::local_mocked_bindings(
    rx_get_json = fake_rx_get_json,
    .package = "rxref"
  )

  out <- get_properties("999999", show_progress = FALSE)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 1L)
  expect_named(out, c(
    "rxcui",
    "name",
    "synonym",
    "tty",
    "language",
    "suppress",
    "umlscui"
  ))
  expect_equal(out$rxcui, "999999")
  expect_true(is.na(out$name))
  expect_true(is.na(out$tty))
})

#' Empty resolve result
#'
#' @keywords internal
#' @noRd
empty_resolve_result <- function(query = character()) {
  tibble::tibble(
    query = query,
    rxcui = NA_character_,
    name = NA_character_,
    tty = NA_character_
  )
}

#' Empty NDC-to-RxCUI result
#'
#' @keywords internal
#' @noRd
empty_ndc_to_rxcui_result <- function(ndc = character()) {
  tibble::tibble(
    ndc = ndc,
    rxcui = NA_character_,
    name = NA_character_,
    tty = NA_character_
  )
}

#' Empty RxCUI-to-NDC result
#'
#' @keywords internal
#' @noRd
empty_rxcui_to_ndc_result <- function() {
  tibble::tibble(
    rxcui = character(),
    ndc = character()
  )
}

# R/classes.R

#' Get RxClass assertions for RxNorm drugs
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `get_classes()` returns RxClass assertions for RxNorm concepts or drug names.
#' These include drug-class relationships from sources such as ATC, FDASPL,
#' DAILYMED, MEDRT, SNOMEDCT, and VA. Depending on the source and relationship,
#' returned assertions may represent pharmacologic classes, chemical structures,
#' mechanisms of action, physiologic effects, indications, contraindications,
#' VA classes, ATC classes, or SNOMED CT dispositions/structures.
#'
#' Because different sources use different classification logic, this function
#' preserves class type, relationship, and relationship source rather than
#' collapsing results into a single class label.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param rela_source Optional RxClass relationship source filter, such as
#'   `"ATC"`, `"ATCPROD"`, `"DAILYMED"`, `"FDASPL"`, `"MEDRT"`, `"SNOMEDCT"`,
#'   or `"VA"`. May be a character vector; values are queried separately.
#' @param relas Optional relationship filter. Examples include `"has_EPC"`,
#'   `"has_MoA"`, `"has_PE"`, `"may_treat"`, `"ci_with"`,
#'   `"has_chemical_structure"`, and others supported by RxClass.
#' @param class_types Optional filter on returned class types, such as `"EPC"`,
#'   `"MOA"`, `"PE"`, `"DISEASE"`, `"CHEM"`, `"VA"`, or `"ATC1-4"`.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble with one row per RxClass assertion.
#' @export
get_classes <- function(
    x,
    by = c("rxcui", "name"),
    rela_source = NULL,
    relas = NULL,
    class_types = NULL,
    keep_input = TRUE,
    show_progress = interactive()
) {
  by <- match.arg(by)

  lifecycle::signal_stage("experimental", "get_classes()")

  stopifnot(is.character(x))

  if (is.null(rela_source)) {
    rela_source <- NA_character_
  }

  grid <- tidyr::expand_grid(
    x = unique(x),
    rela_source = rela_source
  )

  .rxref_progress_map_dfr(
    seq_len(nrow(grid)),
    function(i) {
      .x <- grid$x[[i]]
      .rela_source <- grid$rela_source[[i]]

      query <- list()

      if (by == "rxcui") {
        path <- "/class/byRxcui"
        query$rxcui <- .x
      } else {
        path <- "/class/byDrugName"
        query$drugName <- .x
      }

      if (!is.na(.rela_source)) {
        query$relaSource <- .rela_source
      }

      if (!is.null(relas)) {
        query$relas <- .rxclass_collapse_query(relas)
      }

      res <- rxclass_get_json(path, query = query)

      out <- parse_rxclass_drug_classes(res)

      if (!is.null(class_types)) {
        out <- dplyr::filter(out, .data$class_type %in% class_types)
      }

      if (keep_input) {
        out <- dplyr::mutate(out, input = .x, .before = 1)
      }

      out
    },
    name = "Getting classes",
    show_progress = show_progress
  ) |>
    dplyr::distinct()
}

#' Get class-like RxClass assertions for RxNorm drugs
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' `get_drug_classes()` is an experimental convenience function that returns
#' RxClass assertions that are likely to behave like drug-class labels.
#'
#' It combines selected assertions from ATC, ATCPROD, FDA/SPL EPC, VA, and
#' SNOMED CT disposition relationships. It intentionally excludes relationship
#' types that usually describe contraindications, indications, physiologic
#' effects, chemical structures, or other non-class assertions.
#'
#' For a more compact and ingredient-oriented output, rely on defaults for
#' include_sources. For extended list of product-level sources, consider adding
#' "ATCPROD" and "VA".
#'
#' This function is experimental because "drug class" is not a single native
#' RxClass concept. Different sources use different classification logic, and
#' this helper applies an opinionated filter to return class-like assertions.
#' For source-specific results, use [get_classes()], [get_atc()], [get_epc()],
#' [get_va()], or related functions directly.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param include_sources Character vector of class-like sources to include.
#'   Defaults to `c("ATC", "ATCPROD", "EPC", "VA", "SNOMEDCT")`.
#' @param collapse Logical; if `TRUE`, returns one row per unique class
#'   assertion per input and source, dropping drug-specific columns. If `FALSE`,
#'   returns the full source-specific rows, including matched RxCUI, drug name,
#'   and term type.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of class-like RxClass assertions.
#'
#' @export
get_drug_classes <- function(x,
                             by = c("rxcui", "name"),
                             include_sources = c("ATC", "EPC", "SNOMEDCT"),
                             collapse = TRUE,
                             keep_input = TRUE,
                             show_progress = interactive()) {
  by <- match.arg(by)

  lifecycle::signal_stage("experimental", "get_drug_classes()")

  include_sources <- match.arg(
    include_sources,
    choices = c("ATC", "ATCPROD", "EPC", "VA", "SNOMEDCT"),
    several.ok = TRUE
  )

  out <- list()

  if ("ATC" %in% include_sources) {
    out$atc <- get_atc(
      x = x,
      by = by,
      keep_input = keep_input,
      show_progress = show_progress
    )
  }

  if ("ATCPROD" %in% include_sources) {
    out$atcprod <- get_atcprod(
      x = x,
      by = by,
      keep_input = keep_input,
      show_progress = show_progress
    )
  }

  if ("EPC" %in% include_sources) {
    out$epc <- get_epc(
      x = x,
      by = by,
      keep_input = keep_input,
      show_progress = show_progress
    )
  }

  if ("VA" %in% include_sources) {
    out$va <- get_va(
      x = x,
      by = by,
      keep_input = keep_input,
      show_progress = show_progress
    )
  }

  if ("SNOMEDCT" %in% include_sources) {
    out$snomedct <- get_classes(
      x = x,
      by = by,
      rela_source = "SNOMEDCT",
      relas = "isa_disposition",
      keep_input = keep_input,
      show_progress = show_progress
    )
  }

  out <- dplyr::bind_rows(out) |>
    dplyr::distinct()

  if (isTRUE(collapse)) {
    if (isTRUE(keep_input)) {
      out <- out |>
        dplyr::select(
          .data$input,
          .data$class_id,
          .data$class_name,
          .data$class_type,
          .data$class_url,
          .data$rela,
          .data$rela_source
        )
    } else {
      out <- out |>
        dplyr::select(
          .data$class_id,
          .data$class_name,
          .data$class_type,
          .data$class_url,
          .data$rela,
          .data$rela_source
        )
    }

    out <- out |>
      dplyr::distinct()
  }

  out
}


#' Get ATC classes for RxNorm drugs
#'
#' `get_atc()` is a convenience wrapper for `get_classes()` that returns
#' Anatomical Therapeutic Chemical (ATC) class assertions for RxNorm drugs.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of ATC class assertions.
#' @export
get_atc <- function(x,
                    by = c("rxcui", "name"),
                    keep_input = TRUE,
                    show_progress = interactive()) {
  by <- match.arg(by)

  get_classes(
    x = x,
    by = by,
    rela_source = "ATC",
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Get ATC product-level classes for RxNorm drugs
#'
#' `get_atc()` is a convenience wrapper for `get_classes()` that returns
#' Anatomical Therapeutic Chemical (ATC) product-level class assertions for
#' RxNorm drugs.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of ATCPROD class assertions.
#' @export
get_atcprod <- function(x,
                        by = c("rxcui", "name"),
                        keep_input = TRUE,
                        show_progress = interactive()) {
  by <- match.arg(by)

  get_classes(
    x = x,
    by = by,
    rela_source = "ATCPROD",
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Get FDA/SPL established pharmacologic classes
#'
#' `get_epc()` is a convenience wrapper for `get_classes()` that returns
#' FDA labeling class assertions for RxNorm drugs.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of established pharmacologic class assertions.
#' @export
get_epc <- function(x,
                    by = c("rxcui", "name"),
                    keep_input = TRUE,
                    show_progress = interactive()) {
  by <- match.arg(by)

  get_classes(
    x = x,
    by = by,
    rela_source = c("DAILYMED", "FDASPL"),
    relas = "has_EPC",
    class_types = "EPC",
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Get MED-RT assertions for RxNorm drugs
#'
#' `get_medrt()` is a convenience wrapper for `get_classes()` that returns
#' MED-RT assertions for RxNorm drugs. These include, for example, asserted
#' mechanisms of action, contraindications, physiologic effects (including
#' adverse side effects), etc.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param relas Optional MED-RT relationship filter.
#' @param class_types Optional MED-RT class type filter.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of MED-RT assertions.
#' @export
get_medrt <- function(x,
                      by = c("rxcui", "name"),
                      relas = NULL,
                      class_types = NULL,
                      keep_input = TRUE,
                      show_progress = interactive()) {
  by <- match.arg(by)

  get_classes(
    x = x,
    by = by,
    rela_source = "MEDRT",
    relas = relas,
    class_types = class_types,
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Get MED-RT mechanism-of-action assertions
#'
#' `get_medrt_moa()` is a convenience wrapper for `get_medrt()` that returns
#' MED-RT mechanism-of-action assertions when available. Not all drugs have
#' MED-RT has_moa assertions; for some drugs, mechanism-like or class-like
#' information may instead be available from FDA/SPL EPC via `get_epc()` or as
#' physiologic effects via `get_medrt_pe()`.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of MED-RT mechanism-of-action assertions.
#' @export
get_medrt_moa <- function(x,
                          by = c("rxcui", "name"),
                          keep_input = TRUE,
                          show_progress = interactive()) {
  by <- match.arg(by)

  get_medrt(
    x = x,
    by = by,
    relas = "has_MoA",
    class_types = "MOA",
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Get MED-RT physiologic-effect assertions
#'
#' `get_medrt_pe()` is a convenience wrapper for `get_medrt()` that returns
#' only physiologic effect MED-RT assertions for RxNorm drugs.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of MED-RT physiologic-effect assertions.
#' @export
get_medrt_pe <- function(x,
                         by = c("rxcui", "name"),
                         keep_input = TRUE,
                         show_progress = interactive()) {
  by <- match.arg(by)

  get_medrt(
    x = x,
    by = by,
    relas = "has_PE",
    class_types = "PE",
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Get contraindication assertions from MED-RT
#'
#' `get_contraindications()` is a convenience wrapper for `get_medrt()` that
#' returns only contraindication MED-RT assertions for RxNorm drugs.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of MED-RT contraindication assertions.
#' @export
get_contraindications <- function(x,
                                  by = c("rxcui", "name"),
                                  keep_input = TRUE,
                                  show_progress = interactive()) {
  by <- match.arg(by)

  get_medrt(
    x = x,
    by = by,
    relas = "ci_with",
    class_types = "DISEASE",
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Get indication/treatment assertions from MED-RT
#'
#' `get_may_treat()` is a convenience wrapper for `get_medrt()` that returns
#' indications (disease assertions) from MED-RT for RxNorm drugs, when
#' available. Note that absence of a row (or an empty table altogether)
#' should not necessarily be interpreted as absence of an indication.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of MED-RT may-treat assertions.
#' @export
get_may_treat <- function(x,
                          by = c("rxcui", "name"),
                          keep_input = TRUE,
                          show_progress = interactive()) {
  by <- match.arg(by)

  get_medrt(
    x = x,
    by = by,
    relas = "may_treat",
    class_types = "DISEASE",
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Get chemical-structure assertions
#'
#' `get_chemical_structure()` is a convenience wrapper for `get_classes()` that
#' returns chemical-structure assertions for RxNorm drugs.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param rela_source Relationship source. Defaults to FDASPL, DAILYMED, and MEDRT.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of chemical-structure assertions.
#' @export
get_chemical_structure <- function(x,
                                   by = c("rxcui", "name"),
                                   rela_source = c("FDASPL", "DAILYMED", "MEDRT"),
                                   keep_input = TRUE,
                                   show_progress = interactive()) {
  by <- match.arg(by)

  get_classes(
    x = x,
    by = by,
    rela_source = rela_source,
    relas = "has_chemical_structure",
    class_types = "CHEM",
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Get VA drug classes for RxNorm drugs
#'
#' `get_va()` is a convenience wrapper for `get_classes()` that
#' returns VA drug class assertions for RxNorm drugs.
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param extended Logical; if `TRUE`, includes extended VA class assertions.
#' @param keep_input Logical; if `TRUE`, includes the original input value.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble of VA class assertions.
#' @export
get_va <- function(x,
                   by = c("rxcui", "name"),
                   extended = TRUE,
                   keep_input = TRUE,
                   show_progress = interactive()) {
  by <- match.arg(by)

  relas <- if (isTRUE(extended)) {
    c("has_VAClass", "has_VAClass_extended")
  } else {
    "has_VAClass"
  }

  get_classes(
    x = x,
    by = by,
    rela_source = "VA",
    relas = relas,
    class_types = "VA",
    keep_input = keep_input,
    show_progress = show_progress
  )
}


#' Find RxClass drug classes by class name
#'
#' @param query Character string to search for.
#' @param class_types Optional class type filter, such as `"EPC"`, `"ATC1"`,
#'   `"ATC2"`, `"ATC3"`, `"ATC4"`, `"CHEM"`, `"DISEASE"`, or `"VA"`.
#'
#' @return A tibble of matching class concepts.
#' @export
find_classes <- function(query, class_types = NULL) {
  stopifnot(is.character(query), length(query) == 1)

  q <- list(className = query)

  if (!is.null(class_types)) {
    q$classTypes <- .rxclass_collapse_query(class_types)
  }

  res <- rxclass_get_json(
    "/class/byName",
    query = q
  )

  parse_rxclass_find_classes(res)
}


#' Get RxNorm members of an RxClass class
#'
#' `get_class_members()` queries the RxClass API to identify members of a
#' specific drug class. Typically, this would be used to identify all
#' ingredient RxCUIs in a class, which can then be combined with other search
#' functions to extract related product RxCUIs or NDCs.
#'
#' @param class_id RxClass class identifier.
#' @param rela_source RxClass relationship source, such as `"ATC"`,
#'   `"ATCPROD"`, `"DAILYMED"`, `"FDASPL"`, `"MEDRT"`, `"SNOMEDCT"`, or `"VA"`.
#' @param rela Optional relationship filter.
#' @param include_indirect Logical; if `TRUE`, include direct and indirect
#'   members. If `FALSE`, return direct members only.
#' @param ttys Optional RxNorm term type filter, such as `"IN"`, `"PIN"`,
#'   `"SCD"`, or `"SBD"`.
#'
#' @return A tibble of RxNorm drug members.
#' @export
get_class_members <- function(class_id,
                              rela_source,
                              rela = NULL,
                              include_indirect = TRUE,
                              ttys = NULL) {
  stopifnot(is.character(class_id), length(class_id) == 1)
  stopifnot(is.character(rela_source), length(rela_source) == 1)

  q <- list(
    classId = class_id,
    relaSource = rela_source,
    trans = if (isTRUE(include_indirect)) "0" else "1"
  )

  if (!is.null(rela)) {
    q$rela <- rela
  }

  if (!is.null(ttys)) {
    q$ttys <- .rxclass_collapse_query(ttys)
  }

  res <- rxclass_get_json(
    "/classMembers",
    query = q
  )

  parse_rxclass_members(
    res = res,
    class_id = class_id,
    rela_source = rela_source,
    rela = rela
  )
}

#' Summarize RxClass relationship types for drugs
#'
#' `rxclass_relationships()` is a diagnostic helper that evaluates the type(s)
#' of relationship (rela) asserted in RxNorm MED-RT data and the number within
#' each relationship asserted. In some cases, `get_medrt()` or its various
#' convenience wrappers may return empty or shorter-than-expected tibbles
#' because some relationships expected by the user are not asserted in RxNorm
#' (for example, an MOA may not be asserted by MED-RT).
#'
#' @param x Character vector of RxCUIs or drug names.
#' @param by One of `"rxcui"` or `"name"`.
#' @param rela_source Optional RxClass relationship source filter.
#' @param show_progress Logical. Show a progress bar in interactive sessions.
#'   Progress is shown only when at least 5 inputs are supplied.
#'
#' @return A tibble summarizing available relationship/class-type combinations.
#' @export
rxclass_relationships <- function(x,
                                  by = c("rxcui", "name"),
                                  rela_source = NULL,
                                  show_progress = interactive()) {
  by <- match.arg(by)

  get_classes(
    x = x,
    by = by,
    rela_source = rela_source,
    keep_input = TRUE,
    show_progress = show_progress
  ) |>
    dplyr::count(
      .data$input,
      .data$rela_source,
      .data$rela,
      .data$class_type,
      name = "n"
    ) |>
    dplyr::arrange(
      .data$input,
      .data$rela_source,
      .data$rela,
      .data$class_type
    )
}

#' Get RxClass relationship types
#'
#' `rxclass_relas()` is a helper function that queries the RxClass API for
#' generally valid relationships within the data. Note that just because a
#' relationship is generally valid within a rela_source, not every drug or
#' RxCUI will have a corresponding relationship asserted.
#'
#' @param rela_source RxClass relationship source, such as `"ATC"`,
#'   `"DAILYMED"`, `"FDASPL"`, `"MEDRT"`, `"SNOMEDCT"`, or `"VA"`.
#'
#' @return A tibble of relationship names for the selected source.
#' @export
rxclass_relas <- function(rela_source) {
  stopifnot(is.character(rela_source), length(rela_source) == 1)

  res <- rxclass_get_json(
    "/relas",
    query = list(relaSource = rela_source)
  )

  relas <- rx_pluck_chr(res, "relaList", "relaName")

  if (is.null(relas) || length(relas) == 0) {
    return(tibble::tibble(
      rela_source = character(),
      rela = character()
    ))
  }

  tibble::tibble(
    rela_source = rela_source,
    rela = as.character(relas)
  )
}


#' Get RxClass relationship sources
#'
#' @return A tibble of RxClass relationship sources.
#' @export
rxclass_rela_sources <- function() {
  res <- rxclass_get_json("/relaSources")

  sources <- rx_pluck_chr(res, "relaSourceList", "relaSourceName")

  if (is.null(sources) || length(sources) == 0) {
    return(tibble::tibble(rela_source = character()))
  }

  tibble::tibble(rela_source = as.character(sources))
}



#' @keywords internal
#' @noRd
empty_rxclass_drug_classes <- function() {
  tibble::tibble(
    rxcui = character(),
    drug_name = character(),
    drug_tty = character(),
    class_id = character(),
    class_name = character(),
    class_type = character(),
    class_url = character(),
    rela = character(),
    rela_source = character()
  )
}


#' @keywords internal
#' @noRd
parse_rxclass_drug_classes <- function(res) {
  classes <- rx_pluck_list(res, "rxclassDrugInfoList", "rxclassDrugInfo")

  if (!length(classes)) {
    return(empty_rxclass_drug_classes())
  }

  purrr::map_dfr(classes, function(z) {
    concept <- rx_pluck(z, "minConcept", .default = list())
    cls <- rx_pluck(z, "rxclassMinConceptItem", .default = list())

    tibble::tibble(
      rxcui = rx_scalar_chr(rx_pluck_chr(concept, "rxcui")),
      drug_name = rx_scalar_chr(rx_pluck_chr(concept, "name")),
      drug_tty = rx_scalar_chr(rx_pluck_chr(concept, "tty")),
      class_id = rx_scalar_chr(rx_pluck_chr(cls, "classId")),
      class_name = rx_scalar_chr(rx_pluck_chr(cls, "className")),
      class_type = rx_scalar_chr(rx_pluck_chr(cls, "classType")),
      class_url = rx_scalar_chr(rx_pluck_chr(cls, "classUrl")),
      rela = rx_scalar_chr(rx_pluck_chr(z, "rela")),
      rela_source = rx_scalar_chr(rx_pluck_chr(z, "relaSource"))
    )
  })
}


#' @keywords internal
#' @noRd
parse_rxclass_find_classes <- function(res) {
  classes <- rx_pluck_list(res, "rxclassMinConceptList", "rxclassMinConcept")

  if (!length(classes)) {
    return(tibble::tibble(
      class_id = character(),
      class_name = character(),
      class_type = character(),
      class_url = character()
    ))
  }

  purrr::map_dfr(classes, function(z) {
    tibble::tibble(
      class_id = rx_scalar_chr(rx_pluck_chr(z, "classId")),
      class_name = rx_scalar_chr(rx_pluck_chr(z, "className")),
      class_type = rx_scalar_chr(rx_pluck_chr(z, "classType")),
      class_url = rx_scalar_chr(rx_pluck_chr(z, "classUrl"))
    )
  })
}

#' @keywords internal
#' @noRd
empty_rxclass_members <- function() {
  tibble::tibble(
    rxcui = character(),
    name = character(),
    tty = character(),
    class_id = character(),
    rela_source = character(),
    rela = character(),
    source_id = character(),
    source_name = character(),
    source_url = character(),
    relation = character()
  )
}


#' @keywords internal
#' @noRd
parse_rxclass_members <- function(res,
                                  class_id = NA_character_,
                                  rela_source = NA_character_,
                                  rela = NA_character_) {
  members <- rx_pluck_list(res, "drugMemberGroup", "drugMember")

  if (!length(members)) {
    return(empty_rxclass_members())
  }

  purrr::map_dfr(members, function(z) {
    concept <- rx_pluck(z, "minConcept", .default = list())
    attrs <- parse_rxclass_node_attrs(rx_pluck_list(z, "nodeAttr"))

    tibble::tibble(
      rxcui = rx_scalar_chr(rx_pluck_chr(concept, "rxcui")),
      name = rx_scalar_chr(rx_pluck_chr(concept, "name")),
      tty = rx_scalar_chr(rx_pluck_chr(concept, "tty")),
      class_id = class_id,
      rela_source = rela_source,
      rela = if (is.null(rela)) NA_character_ else rela,
      source_id = rx_scalar_chr(attrs$SourceId),
      source_name = rx_scalar_chr(attrs$SourceName),
      source_url = rx_scalar_chr(attrs$SourceUrl),
      relation = rx_scalar_chr(attrs$Relation)
    )
  })
}


#' @keywords internal
#' @noRd
parse_rxclass_node_attrs <- function(node_attr) {
  out <- list(
    SourceId = NA_character_,
    SourceName = NA_character_,
    SourceUrl = NA_character_,
    Relation = NA_character_
  )

  if (!length(node_attr)) {
    return(out)
  }

  for (a in node_attr) {
    nm <- rx_scalar_chr(rx_pluck_chr(a, "attrName"))
    val <- rx_scalar_chr(rx_pluck_chr(a, "attrValue"))

    if (!is.na(nm) && nm %in% names(out)) {
      out[[nm]] <- val
    }
  }

  out
}

test_that("parse_rxclass_drug_classes handles empty response", {
  res <- list(rxclassDrugInfoList = list(rxclassDrugInfo = NULL))
  out <- rxref:::parse_rxclass_drug_classes(res)

  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0)
  expect_named(out, c(
    "rxcui", "drug_name", "drug_tty",
    "class_id", "class_name", "class_type",
    "class_url", "rela", "rela_source"
  ))
})

test_that("parse_rxclass_drug_classes parses one class assertion", {
  res <- list(
    rxclassDrugInfoList = list(
      rxclassDrugInfo = list(
        list(
          minConcept = list(
            rxcui = "29046",
            name = "lisinopril",
            tty = "IN"
          ),
          rxclassMinConceptItem = list(
            classId = "C09AA",
            className = "ACE inhibitors, plain",
            classType = "ATC4",
            classUrl = "https://example.com"
          ),
          rela = "has_ATC",
          relaSource = "ATC"
        )
      )
    )
  )

  out <- rxref:::parse_rxclass_drug_classes(res)

  expect_equal(nrow(out), 1)
  expect_equal(out$rxcui, "29046")
  expect_equal(out$class_id, "C09AA")
  expect_equal(out$class_type, "ATC4")
  expect_equal(out$rela_source, "ATC")
})

---
title: "Building route-specific RxCUI and NDC lists"
author: "Steven Smith"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Building route-specific RxCUI and NDC lists}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
# Vignettes use precomputed example data by default.
# To rebuild examples with live RxNorm/RxClass API calls, set:
# Sys.setenv(RXREF_BUILD_VIGNETTES_ONLINE = "true")

online_env <- identical(
  tolower(Sys.getenv("RXREF_BUILD_VIGNETTES_ONLINE")),
  "true"
)

has_net <- tryCatch({
  requireNamespace("curl", quietly = TRUE) && curl::has_internet()
}, error = function(e) FALSE)

run_live <- online_env && has_net

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(rxref)
library(dplyr)

read_rxref_example <- function(file) {
  path <- system.file("extdata", file, package = "rxref")
  if (!nzchar(path)) {
    stop(
      "The example data file '", file, "' was not found. ",
      "Reinstall rxref or rebuild the vignette with ",
      "RXREF_BUILD_VIGNETTES_ONLINE=true."
    )
  }
  readRDS(path)
}
```

## Overview

A common use case for `rxref` is to develop a list of RxNorm product
concepts and National Drug Codes (NDCs) for a drug class.

For example, suppose we want to identify product RxCUIs and active NDCs
for oral beta-blockers. This is not as simple as searching for
beta-blocker ingredients and mapping every product to NDCs, because some
beta-blockers are available through multiple routes.

For example:

- `metoprolol`, `atenolol`, `labetalol`, `propranolol`, and `sotalol`
  may have oral and injectable products;
- `esmolol` is injectable only;
- `timolol`, `betaxolol`, and related drugs may have ophthalmic products.

For many pharmacoepidemiologic studies, we may want oral outpatient-use
products only.

This vignette walks through a tidy workflow for:

1. resolving drug names to ingredient RxCUIs;
2. expanding ingredient RxCUIs to product RxCUIs;
3. filtering product RxCUIs by route; and
4. mapping the filtered product RxCUIs to active NDCs.

The examples use precomputed data by default so the vignette can be built
without querying the live RxNorm API. To rebuild the examples with live
API calls, set:

```{r eval = FALSE}
Sys.setenv(RXREF_BUILD_VIGNETTES_ONLINE = "true")
```

## Define a beta-blocker ingredient list

In this example, we start with a curated list of beta-blocker ingredient
names.

Users can also use RxClass-related helpers, such as `find_classes()` or
`get_class_members()`, when a suitable source vocabulary is available.
However, curated ingredient lists remain useful because class membership
may not perfectly match a study-specific exposure definition.

```{r define}
beta_blocker_names <- c(
  "acebutolol",
  "atenolol",
  "betaxolol",
  "bisoprolol",
  "carvedilol",
  "labetalol",
  "metoprolol",
  "nadolol",
  "nebivolol",
  "penbutolol",
  "pindolol",
  "propranolol",
  "sotalol",
  "timolol"
)
```

## Resolve drug names to ingredient RxCUIs

First, use `find_ingredients()` to resolve each drug name to an RxNorm
ingredient concept.

```{r resolve}
if (run_live) {
  bb_ingredients <- find_ingredients(beta_blocker_names) |>
    filter(tty == "IN") |>
    distinct(
      input,
      ingredient_rxcui = rxcui,
      ingredient_name = name,
      ingredient_tty = tty
    )
} else {
  bb_ingredients <- read_rxref_example("bb_ingredients.rds")
}

bb_ingredients
```

The resulting table contains one row per resolved ingredient. These
ingredient RxCUIs are the starting point for product expansion.

## Expand ingredients to product RxCUIs

Next, use `products_for_ingredients()` to identify product-level RxNorm
concepts containing each ingredient.

In this example, we use active RxNorm concepts and include combination
products.

```{r expand}
if (run_live) {
  bb_products <- products_for_ingredients(
    bb_ingredients$ingredient_rxcui,
    ttys = product_ttys("default"),
    include_combos = TRUE,
    concept_status = "active"
  ) |>
    left_join(bb_ingredients, by = "ingredient_rxcui")
} else {
  bb_products <- read_rxref_example("bb_products.rds")
}

bb_products
```

Because `include_combos = TRUE`, combination products are retained. For
beta-blockers, this means products such as beta-blocker/thiazide
combinations may be included.

To exclude combination products, set `include_combos = FALSE`.

```{r exclude-option}
if (run_live) {
  bb_single_ingredient_products <- products_for_ingredients(
    bb_ingredients$ingredient_rxcui,
    ttys = product_ttys("default"),
    include_combos = FALSE,
    concept_status = "active"
  ) |>
    left_join(bb_ingredients, by = "ingredient_rxcui")
} else {
  bb_single_ingredient_products <- read_rxref_example(
    "bb_single_ingredient_products.rds"
  )
}

bb_single_ingredient_products
```

## Active versus historical RxNorm concepts

For many current medication list workflows, `concept_status = "active"`
is a good default. This limits product expansion to active RxNorm
concepts.

For studies covering older calendar periods, users may want to include
historical RxNorm concepts as well.

```{r historical-products, eval = FALSE}
bb_products_historical <- products_for_ingredients(
  bb_ingredients$ingredient_rxcui,
  ttys = product_ttys("default"),
  include_combos = TRUE,
  concept_status = "active_and_historical"
)
```

Historical concepts can be useful when reconstructing medication exposure
during older study periods. However, historical product concepts may have
less complete clinical attribute information than active concepts. Route
filtering should therefore be reviewed carefully when
`concept_status = "active_and_historical"` is used.

## Why route filtering matters

At this point, the product list may include products from multiple
routes. For example, some beta-blockers have oral tablets, injectable
products, or ophthalmic solutions.

We can use `get_clinical_attributes()` to inspect route and dose-form
information for the product RxCUIs.

```{r filtering-matters}
if (run_live) {
  bb_attrs <- get_clinical_attributes(
    unique(bb_products$product_rxcui)
  ) |>
    rename(product_rxcui = rxcui)
} else {
  bb_attrs <- read_rxref_example("bb_attrs.rds")
}

bb_attrs |>
  count(route, dose_form_group, sort = TRUE)
```

This helps verify whether the product list includes routes that are
outside the intended use case.

## Filter to oral products

To keep only orally available products, use `filter_products_by_route()`.

```{r filter}
if (run_live) {
  bb_oral_products <- bb_products |>
    filter_products_by_route(route = "ORAL")
} else {
  bb_oral_products <- read_rxref_example("bb_oral_products.rds")
}

bb_oral_products
```

The returned table keeps product-level metadata and appends summarized
route and dose-form information.

```{r filter-summary}
bb_oral_products |>
  count(route, dose_form_group, sort = TRUE)
```

This should retain products such as oral tablets, capsules, and oral
solutions, while excluding injectable or ophthalmic products.

Route filtering relies on clinical attribute information available
through RxNorm. If a product concept lacks route or dose-form
information, it may not be retained by route-specific filters. This is
especially important when historical concepts are included.

## Map oral product RxCUIs to active NDCs

Once the product list is restricted to oral products, use
`map_rxcui_to_ndc()` to map the product RxCUIs to NDCs.

```{r map}
if (run_live) {
  bb_oral_ndc_map <- map_rxcui_to_ndc(
    unique(bb_oral_products$product_rxcui),
    status = "ACTIVE"
  )

  bb_oral_ndcs <- bb_oral_ndc_map |>
    left_join(
      bb_oral_products,
      by = c("rxcui" = "product_rxcui")
    ) |>
    rename(
      product_rxcui = rxcui,
    ) |>
    distinct()
} else {
  bb_oral_ndcs <- read_rxref_example("bb_oral_ndcs.rds")
}

bb_oral_ndcs
```

The output is a flat tibble with one row per product RxCUI/NDC pair.
Product metadata is retained alongside the NDC.

```{r map-select}
bb_oral_ndcs |>
  select(
    ingredient_name,
    product_rxcui,
    name,
    tty,
    route,
    dose_form,
    ndc11,
    ndc_status
  )
```

The `status = "ACTIVE"` argument controls the NDC status returned by
`map_rxcui_to_ndc()`. This is distinct from `concept_status`, which
controls whether active or historical RxNorm concepts are included during
product expansion.

## Summarize the resulting list

For quality control, it is useful to summarize the number of products and
NDCs by ingredient.

```{r summarize-products}
bb_oral_products |>
  count(ingredient_name, sort = TRUE, name = "n_product_rxcuis")
```

```{r summarize-ndcs}
bb_oral_ndcs |>
  count(ingredient_name, sort = TRUE, name = "n_active_ndcs")
```

You may also want to inspect combination products separately.

```{r summarize-combinations}
bb_oral_products |>
  filter(n_ingredients > 1) |>
  select(
    ingredient_name,
    product_rxcui,
    name,
    tty,
    n_ingredients,
    route,
    dose_form
  )
```

Depending on the scientific question, combination products may be
appropriate to include or exclude. For example, beta-blocker/thiazide
combination products may be relevant for antihypertensive exposure
definitions but not for studies focused on beta-blocker monotherapy.

## A shortcut using `search_drug()`

The same core workflow can be run more compactly with `search_drug()`.
This function combines ingredient searching, product expansion, optional
route filtering, and optional NDC mapping.

To return oral product RxCUIs only:

```{r alt-search-rxcui}
if (run_live) {
  bb_oral_rxcuis <- search_drug(
    beta_blocker_names,
    return = "rxcui",
    route = "ORAL",
    include_combos = TRUE,
    concept_status = "active"
  )
} else {
  bb_oral_rxcuis <- read_rxref_example("bb_oral_rxcuis_search.rds")
}

bb_oral_rxcuis
```

To return a flat table of oral product RxCUIs and active NDCs:

```{r alt-search-ndc}
if (run_live) {
  bb_oral_ndcs_search_raw <- search_drug(
    beta_blocker_names,
    return = "ndc",
    route = "ORAL",
    ndc_status = "ACTIVE",
    include_combos = TRUE,
    concept_status = "active"
  )

  bb_oral_ndcs_search <- bb_oral_ndcs_search_raw

  if (!"ingredient_name" %in% names(bb_oral_ndcs_search)) {
    bb_oral_ndcs_search <- bb_oral_ndcs_search |>
      left_join(
        bb_ingredients |>
          select(ingredient_rxcui, ingredient_name),
        by = "ingredient_rxcui"
      )
  }

  product_cols <- c(
    "name",
    "tty",
    "route",
    "dose_form",
    "dose_form_group"
  )

  missing_product_cols <- setdiff(product_cols, names(bb_oral_ndcs_search))

  if (length(missing_product_cols) > 0) {
    bb_oral_ndcs_search <- bb_oral_ndcs_search |>
      left_join(
        bb_oral_products |>
          select(
            ingredient_rxcui,
            product_rxcui,
            all_of(missing_product_cols)
          ),
        by = c("ingredient_rxcui", "product_rxcui")
      )
  }

  bb_oral_ndcs_search <- bb_oral_ndcs_search |>
    distinct()
} else {
  bb_oral_ndcs_search <- read_rxref_example("bb_oral_ndcs_search.rds")
}

bb_oral_ndcs_search
```

The raw `search_drug(return = "ndc")` output is intentionally compact. In
this vignette, we join it back to the ingredient and product tables so
that the displayed example includes readable ingredient names, product
names, routes, and dose forms.

## Returning both products and NDCs

If you want both the unique product RxCUI table and the expanded NDC
table, use `return = "both"`.

```{r products-and-ndcs}
if (run_live) {
  bb_oral_both <- search_drug(
    beta_blocker_names,
    return = "both",
    route = "ORAL",
    ndc_status = "ACTIVE",
    include_combos = TRUE,
    concept_status = "active"
  )
} else {
  bb_oral_both <- read_rxref_example("bb_oral_both_search.rds")
}

names(bb_oral_both)
```

This returns a list because the two tables have different grains:

- `products`: one row per product RxCUI;
- `ndcs`: one row per product RxCUI/NDC pair.

```{r products-and-ndcs-products}
bb_oral_both$products
```

```{r products-and-ndcs-ndcs}
bb_oral_both$ndcs
```

## Suggested QC checks

When developing a study-specific medication list, it is good practice to
review the resulting concepts before finalizing the cohort definition.
Even if the package is working correctly, RxNorm concepts and source data
may occasionally contain surprises for a specific study use case.

The following are some examples.

### Check whether any non-oral product names remain

```{r qc-nonoral-names}
bb_oral_products |>
  filter(grepl(
    "Injection|Injectable|Ophthalmic|Topical|Transdermal|Nasal|Inhalation",
    name,
    ignore.case = TRUE
  ))
```

### Check the observed routes and dose-form groups

```{r qc-routes}
bb_oral_products |>
  count(route, dose_form_group, sort = TRUE)
```

### Check combination products

```{r qc-combinations}
bb_oral_products |>
  filter(n_ingredients > 1) |>
  count(ingredient_name, sort = TRUE)
```

### Check products without active NDCs

Some product RxCUIs may not map to active NDCs.

```{r qc-without-ndcs}
bb_oral_products |>
  anti_join(
    bb_oral_ndcs |>
      distinct(product_rxcui),
    by = "product_rxcui"
  ) |>
  select(
    ingredient_name,
    product_rxcui,
    name,
    tty,
    route,
    dose_form
  )
```

## Summary

This vignette demonstrated how to build a route-specific medication list,
with oral beta-blockers as an example, using `rxref`.

The key steps are:

1. `find_ingredients()` to resolve drug names to ingredient RxCUIs;
2. `products_for_ingredients()` to expand ingredients to product RxCUIs;
3. `filter_products_by_route()` to retain products for the route of
   interest;
4. `map_rxcui_to_ndc()` to obtain NDCs; or
5. `search_drug()` for a compact end-to-end workflow.

For drug classes with products available through multiple routes, route
filtering is an important quality-control step before mapping to NDCs or
using the resulting list in pharmacoepidemiologic analyses.

Users should also distinguish between two separate choices:

- `concept_status` controls whether active or historical RxNorm concepts
  are included.
- `status` or `ndc_status` controls which NDC status categories are
  returned.

For reproducible research, save the final product list, NDC list, package
version, and API-derived outputs used in the analytic workflow.

---
title: "Case Study: GLP-1 and related incretin therapies"
author: "Steven Smith"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{Case Study: GLP-1 and related incretin therapies}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 72
---

```{r, include = FALSE}
# Vignettes use precomputed example data by default.
# To rebuild examples with live RxNorm/RxClass API calls, set:
# Sys.setenv(RXREF_BUILD_VIGNETTES_ONLINE = "true")

online_env <- identical(
  tolower(Sys.getenv("RXREF_BUILD_VIGNETTES_ONLINE")),
  "true"
)

has_net <- tryCatch({
  requireNamespace("curl", quietly = TRUE) && curl::has_internet()
}, error = function(e) FALSE)

run_live <- online_env && has_net

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(rxref)
library(dplyr)

read_rxref_example <- function(file) {
  path <- system.file("extdata", file, package = "rxref")
  if (!nzchar(path)) {
    stop(
      "The example data file '", file, "' was not found. ",
      "Reinstall rxref or rebuild the vignette with ",
      "RXREF_BUILD_VIGNETTES_ONLINE=true."
    )
  }
  readRDS(path)
}
```

## Case

Suppose we need to identify users of GLP-1 receptor agonists and related
incretin-based therapies from EHR prescribing data, pharmacy claims data,
or both. To accomplish this, we need a medication list that includes
relevant RxNorm product concepts and, when available, corresponding NDCs.

This vignette walks through two approaches:

1. a transparent step-by-step workflow that starts with a curated list of
   ingredient names; and
2. a compact workflow using `search_drug()`.

The examples use precomputed data by default so the vignette can be built
without querying the live RxNorm API. To rebuild the examples with live
API calls, set:

```{r eval = FALSE}
Sys.setenv(RXREF_BUILD_VIGNETTES_ONLINE = "true")
```

## Defining the ingredient list

For this example, we start with a prespecified list of GLP-1 receptor
agonists and related incretin-based therapies:

- exenatide
- liraglutide
- lixisenatide
- dulaglutide
- albiglutide
- semaglutide
- tirzepatide

Tirzepatide is included here because many applied studies group it with
GLP-1-based incretin therapies, although it is a dual GIP/GLP-1 receptor
agonist rather than a GLP-1 receptor agonist alone.

```{r define-ingredients}
glp1.names <- c(
  "semaglutide",
  "exenatide",
  "liraglutide",
  "lixisenatide",
  "dulaglutide",
  "albiglutide",
  "tirzepatide"
)
```

## Option 1: Step-by-step medication list construction

### Identify ingredient RxCUIs

First, we use `find_ingredients()` to identify ingredient-level RxCUIs.
For this example, we retain concepts with TTY = `"IN"`, corresponding to
RxNorm ingredient concepts. You can see available TTY values and
descriptions with `tty_catalogue()`.

```{r ings}
if (run_live) {
  glp1.ings <- find_ingredients(glp1.names) |>
    filter(tty == "IN") |>
    distinct(
      input,
      ingredient_rxcui = rxcui,
      ingredient_name = name,
      ingredient_tty = tty
    )
} else {
  glp1.ings <- read_rxref_example("glp1_ings.rds")
}

glp1.ings
```

### Expand ingredients to product RxCUIs

Next, we use `products_for_ingredients()` to identify product concepts
related to the ingredient RxCUIs.

By default, this workflow focuses on active RxNorm concepts. This is
usually appropriate for current medication list construction. For studies
covering older calendar periods, users may want to include historical
RxNorm concepts as well.

```{r tty-sets}
product_ttys("default")
product_ttys("extended_product")
```

The default product TTY set is intended to capture product concepts that
are commonly useful for medication list construction and NDC mapping. A
broader product-related set is available with
`product_ttys("extended_product")`. Users can also supply their own
character vector of TTYs.

In this example, we include combination products. That means products
containing a GLP-1-related ingredient plus one or more other ingredients
may be retained.

```{r prods}
if (run_live) {
  glp1.prods <- products_for_ingredients(
    glp1.ings$ingredient_rxcui,
    ttys = product_ttys("default"),
    include_combos = TRUE,
    concept_status = "active"
  )
} else {
  glp1.prods <- read_rxref_example("glp1_prods.rds")
}

glp1.prods |>
  head(30)
```

Combination products may appear when `include_combos = TRUE`. Depending
on the function, ingredient fields for combination products may be
summarized across ingredients, while product-level rows remain one row
per product concept. Users should inspect ingredient fields and
`n_ingredients` before deciding whether to include or exclude fixed-dose
combination products.

### Include historical RxNorm concepts when needed

For historical studies, formulary reconstruction, or claims data spanning
older calendar periods, users may want to include both active and
historical RxNorm concepts.

```{r historical-products, eval = FALSE}
glp1.prods_historical <- products_for_ingredients(
  glp1.ings$ingredient_rxcui,
  ttys = product_ttys("default"),
  include_combos = TRUE,
  concept_status = "active_and_historical"
)
```

Historical concepts can be useful when reconstructing medication exposure
during older study periods. However, some historical concepts may have
less complete clinical attribute information than active concepts. Users
should review route, dose form, ingredient count, and NDC mappings
carefully when including historical concepts.

### Map product RxCUIs to NDCs

Next, we identify NDCs associated with the product RxCUIs. Not all RxCUIs
map to NDCs, so some product concepts may not have corresponding NDC
values.

```{r ndcs}
if (run_live) {
  glp1.ndc.map <- map_rxcui_to_ndc(
    unique(glp1.prods$product_rxcui),
    status = "ACTIVE"
  )
} else {
  glp1.ndc.map <- read_rxref_example("glp1_ndc_map.rds")
}

glp1.ndcs <- glp1.ndc.map |>
  left_join(
    glp1.prods,
    by = c("rxcui" = "product_rxcui")
  ) |>
  left_join(
    glp1.ings |>
      select(ingredient_rxcui, ingredient_name),
    by = "ingredient_rxcui"
  ) |>
  distinct(
    ingredient_rxcui,
    ingredient_name,
    product_rxcui = rxcui,
    ndc11,
    ndc_status,
    name,
    tty
  ) |>
  arrange(ingredient_name, product_rxcui, ndc11)

glp1.ndcs |>
  head(30)
```

At this point, we have a product-level and NDC-level medication list that
can be used to query EHR prescribing data, pharmacy dispensing data, or
pharmacy claims data.

## Option 2: Use `search_drug()` for a compact workflow

The same goal can often be accomplished in one step with `search_drug()`.
This function combines ingredient searching, product expansion, optional
route filtering, and optional NDC mapping.

Suppose we want NDCs for the same ingredient list, and we want to include
active, obsolete, and unspecified NDCs.

```{r search}
if (run_live) {
  alt.glp1.ndcs <- search_drug(
    term = glp1.names,
    return = "ndc",
    concept_status = "active",
    ndc_status = c("ACTIVE", "OBSOLETE", "UNSPECIFIED")
  )
} else {
  alt.glp1.ndcs <- read_rxref_example("alt_glp1_ndc.rds")
}

alt.glp1.ndcs |>
  arrange(ingredient_name, product_rxcui, ndc11) |>
  head(30)
```

Here, `concept_status` controls whether active or historical RxNorm
concepts are considered. The `ndc_status` argument controls which NDC
status categories are returned.

For example, to include historical RxNorm concepts and broader NDC status
categories, use:

```{r search-historical, eval = FALSE}
search_drug(
  term = glp1.names,
  return = "ndc",
  concept_status = "active_and_historical",
  ndc_status = c("ACTIVE", "OBSOLETE", "UNSPECIFIED")
)
```

This can be useful for studies spanning older calendar periods, but users
should carefully inspect the resulting concepts and NDCs before finalizing
an exposure definition.

## Comparing the two approaches

The step-by-step approach is more verbose, but it makes each decision
explicit:

1. identify ingredient RxCUIs;
2. expand ingredients to product concepts;
3. decide whether to include combination products;
4. decide whether to include active concepts only or active and historical
   concepts;
5. map product RxCUIs to NDCs.

The `search_drug()` approach is more compact and is useful for common
workflows where users want a product list or NDC list from one or more
drug names.

To compare the NDCs from the step-by-step workflow against the compact
workflow:

```{r check}
glp1.ndcs |>
  filter(!is.na(ndc11)) |>
  arrange(ingredient_name, product_rxcui, ndc11) |>
  head(30)
```

## Choosing active versus historical concepts

For many current medication lists, `concept_status = "active"` is a good
default. This limits the workflow to active RxNorm concepts.

Historical concepts may be appropriate when:

- the study period includes older calendar years;
- medication exposure is being reconstructed from historical claims;
- users need to capture products that may no longer be active in RxNorm;
- obsolete NDCs are intentionally included.

However, active and historical RxNorm concepts should not be confused
with NDC status. These are separate choices:

- `concept_status` controls which RxNorm concepts are considered.
- `ndc_status` controls which NDC status categories are returned.

For example:

```{r status-example, eval = FALSE}
search_drug(
  term = "semaglutide",
  return = "ndc",
  concept_status = "active_and_historical",
  ndc_status = c("ACTIVE", "OBSOLETE", "UNSPECIFIED")
)
```

## Practical considerations

Medication list construction often requires study-specific decisions.
Before using the resulting list in an analysis, users should consider:

- whether to include fixed-dose combination products;
- whether to include branded products, clinical products, packs, or all
  product-related TTYs;
- whether the study period requires historical RxNorm concepts;
- whether active NDCs only are sufficient, or obsolete/unspecified NDCs
  should also be included;
- whether route, dose form, or strength restrictions are needed;
- whether the final list should be reviewed clinically.

For strict reproducibility, users should save the final product list, NDC
list, and package/API versions used to construct the medication exposure
definition.

---
title: "rxref: Getting started"
author: "rxref authors"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{rxref: Getting started}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

## Overview

`rxref` provides tidy, vectorized helpers for working with drug identifiers and metadata from the RxNorm and RxClass APIs. It is designed for workflows that need to resolve drug names, RxCUIs, and NDCs; retrieve RxNorm concept metadata; expand ingredients to product concepts; map products to NDCs; and build medication lists for pharmacoepidemiologic or health services research.

Common tasks include:

- resolving messy drug names or NDCs to RxCUIs;
- retrieving RxNorm concept properties;
- mapping NDCs to RxCUIs, or RxCUIs to NDCs;
- finding ingredients and related product concepts;
- creating route-specific product or NDC lists;
- optionally including historical RxNorm concepts when studying older calendar periods.

The examples below are not evaluated when the vignette is built, because they query live APIs. You can run them interactively after installing rxref.

## Installation

Install the stable version of `rxref` from CRAN:

```{r install-cran}
# Install from CRAN
# install.packages("rxref")
```

Or the development version:

```{r install-dev}
# install.packages("pak")
# pak::pak("ssmithm/rxref")
```

Then load the package (the vignette will also use dplyr).

```{r pkg-load}
library(rxref)
library(dplyr)
```

## Resolve messy inputs to RxCUIs

The `resolve()` function accepts common drug identifiers, including free-text drug names, NDCs, and RxCUIs. This is often the first step when working with real-world medication data.

```{r resolve}
resolve(c(
  "metformin 500 mg tablet",
  "00093-1048-01",
  "8610"
))
```

The returned object includes the original input, the resolved RxCUI, and available concept information.

## Retrieve properties for RxCUIs

Once you have one or more RxCUIs, use `get_properties()` to retrieve RxNorm 
concept metadata. This is useful when you want to inspect concept names, term types, suppress status, or other basic RxNorm properties.

```{r get-props}
ids <- c("860975", "860976")

get_properties(ids)
```

## Map NDCs and RxCUIs

`rxref` includes helpers for mapping between NDCs and RxCUIs.

To map NDCs to RxCUIs:

```{r map-to-rxcuis}
map_ndc_to_rxcui(c(
  "00093-1048-01",
  "00093-1048-10"
))
```

To map an RxCUI to NDCs:

```{r map-to-ndcs}
map_rxcui_to_ndc("860975")
```

You can optionally filter NDC mappings by NDC status:

```{r map-filtering}
map_rxcui_to_ndc("860975", status = "ACTIVE")
```

NDC status and RxNorm concept status are related but distinct. NDC status refers to whether an NDC is active, obsolete, or otherwise categorized by RxNav. RxNorm concept status refers to whether the RxNorm concept itself is active or historical.

## Find ingredients

The `find_ingredients()` function helps identify ingredient-level RxCUIs from approximate drug names.

```{r find-ings}
find_ingredients(c("metformin", "semaglutide", "lisinopril"))
```

For many medication list workflows, users first identify a set of ingredient RxCUIs and then expand those ingredients to product concepts.

```{r ings2}
ingredients <- find_ingredients(c("metformin", "semaglutide")) |>
  filter(tty == "IN") |>
  distinct(
    ingredient_rxcui = rxcui,
    ingredient_name = name
  )

ingredients
```

## Expand ingredients to product concepts

Use `products_for_ingredients()` to identify product concepts associated with one or more ingredient RxCUIs.

```{r prods-for-ings}
products <- products_for_ingredients(
  ingredients$ingredient_rxcui,
  concept_status = "active"
)

products
```

By default, most workflows should focus on active RxNorm concepts. For studies that span older calendar periods, you may want to include *historical* concepts as well.

```{r historical-prods}
products_historical <- products_for_ingredients(
  ingredients$ingredient_rxcui,
  concept_status = "active_and_historical"
)

products_historical
```

Historical concepts can be useful when reconstructing medication exposure during older study periods. However, some historical concepts may have less complete clinical attribute information than active concepts, so users should review route, dose form, and ingredient information carefully when using historical concepts.

## Choose product term types

RxNorm includes several term types, or TTYs, for clinical drugs, branded drugs, packs, and other related concepts. `rxref` provides helpers for common product-focused TTY sets.

```{r ttys}
product_ttys("default")
product_ttys("extended_product")
product_ttys("extended")
```

For example, you can request an expanded set of product concepts:

```{r extended-product-concepts}
products_extended <- products_for_ingredients(
  ingredients$ingredient_rxcui,
  ttys = product_ttys("extended_product"),
  concept_status = "active"
)

products_extended
```

The appropriate TTY set depends on the study question. A narrower product set may be preferable for simple ingredient-to-product workflows, while an extended set may be useful when users want to capture additional product or pack concepts. More information on TTYs can be found by running `tty_catalogue()`.

## Search for drugs in one step

For many common workflows, `search_drug()` provides a compact interface that combines ingredient search, product expansion, route filtering, and optional NDC mapping.

For example, to search for oral metformin products:

```{r search}
search_drug(
  term = "metformin",
  return = "product",
  route = "ORAL"
)
```

To return NDCs instead of product concepts:

```{r search-ndcs}
search_drug(
  term = "metformin",
  return = "ndc",
  route = "ORAL",
  ndc_status = "ACTIVE"
)
```

To return both product concepts and NDCs:

```{r search-concepts-ndcs}
search_drug(
  term = "metformin",
  return = "both",
  route = "ORAL",
  ndc_status = "ACTIVE"
)
```

For historical studies, you can include historical RxNorm concepts and broader NDC status categories:

```{r search-historical}
search_drug(
  term = "metformin",
  return = "ndc",
  route = "ORAL",
  concept_status = "active_and_historical",
  ndc_status = c("ACTIVE", "OBSOLETE", "UNSPECIFIED")
)
```

Here, `concept_status` controls whether active or historical RxNorm concepts are considered. The `ndc_status` argument controls which NDC status categories are returned.

## Retrieve clinical attributes

Some workflows require route, dose form, strength, or ingredient-count information. The `get_clinical_attributes()` function retrieves product-level clinical attributes for RxCUIs.

```{r clin-attrs}
attrs <- get_clinical_attributes(products$product_rxcui)

attrs
```

These attributes can be useful for route-specific medication lists:

```{r clin-attrs-routes}
attrs |>
  count(routes, dose_form_groups, sort = TRUE)
```

Combination products may include more than one ingredient. Depending on the workflow, users may want to exclude fixed-dose combinations, retain them, or inspect them separately.

```{r clin-attrs-routes-fdcs}
attrs |>
  filter(n_ingredients > 1)
```

## Build route-specific lists

The `filter_products_by_route()` helper can be used after product expansion to retain products matching a specified route.

```{r filter-by-route}
oral_products <- products |>
  filter_products_by_route(route = "ORAL")

oral_products
```

Route filtering relies on available clinical attribute information. If a product concept lacks route or dose-form information, it may not be retained by route-specific filters. This is especially important when working with historical RxNorm concepts.

## Caching and API use

`rxref` is designed to support vectorized workflows, but large medication lists can require many API calls. Repeated requests can be reduced by enabling caching through `rxref_conf()`.

For example, to use an in-memory cache during a session:

```{r cache_mem}
rxref_conf(cache = cachem::cache_mem())
```

You can also adjust the delay between API requests:

```{r adjust-delay}
rxref_conf(rate_delay = 0.2)
```

This can be helpful for large batch workflows or when working interactively with many RxCUIs. But, be kind to the API. RxNorm specifies a [maximum request rate of 20 per second per IP](https://lhncbc.nlm.nih.gov/RxNav/TermsofService.html).

## Reproducibility

The public RxNorm and RxClass APIs reflect current source data and may change over time as RxNorm releases are updated. For strict reproducibility, especially in studies tied to a specific calendar period, users should save the medication lists, product concepts, NDC mappings, and API outputs used in the final analytic workflow.

At some point, a future version of `rxref` may support local or release-specific backends for users who need to bind analyses to specific RxNorm releases. 

## Summary

`rxref` supports common medication-identification workflows, including:

1. resolving names, NDCs, or RxCUIs with `resolve()`;
2. retrieving concept metadata with `get_properties()`;
3. mapping between NDCs and RxCUIs with `map_ndc_to_rxcui()` and `map_rxcui_to_ndc()`;
4. identifying ingredients with `find_ingredients()`;
5. expanding ingredients to products with `products_for_ingredients()`;
6. retrieving route and dose-form information with `get_clinical_attributes()`;
7. building route-specific product lists with `filter_products_by_route()`;
8. using `search_drug()` for compact end-to-end workflows.

These functions are intended to support transparent, reproducible medication list construction while keeping outputs tidy and compatible with common dplyr workflows.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{get_drug_classes}
\alias{get_drug_classes}
\title{Get class-like RxClass assertions for RxNorm drugs}
\usage{
get_drug_classes(
  x,
  by = c("rxcui", "name"),
  include_sources = c("ATC", "EPC", "SNOMEDCT"),
  collapse = TRUE,
  keep_input = TRUE,
  show_progress = interactive()
)
}
\arguments{
\item{x}{Character vector of RxCUIs or drug names.}

\item{by}{One of \code{"rxcui"} or \code{"name"}.}

\item{include_sources}{Character vector of class-like sources to include.
Defaults to \code{c("ATC", "ATCPROD", "EPC", "VA", "SNOMEDCT")}.}

\item{collapse}{Logical; if \code{TRUE}, returns one row per unique class
assertion per input and source, dropping drug-specific columns. If \code{FALSE},
returns the full source-specific rows, including matched RxCUI, drug name,
and term type.}

\item{keep_input}{Logical; if \code{TRUE}, includes the original input value.}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble of class-like RxClass assertions.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{get_drug_classes()} is an experimental convenience function that returns
RxClass assertions that are likely to behave like drug-class labels.

It combines selected assertions from ATC, ATCPROD, FDA/SPL EPC, VA, and
SNOMED CT disposition relationships. It intentionally excludes relationship
types that usually describe contraindications, indications, physiologic
effects, chemical structures, or other non-class assertions.

For a more compact and ingredient-oriented output, rely on defaults for
include_sources. For extended list of product-level sources, consider adding
"ATCPROD" and "VA".

This function is experimental because "drug class" is not a single native
RxClass concept. Different sources use different classification logic, and
this helper applies an opinionated filter to return class-like assertions.
For source-specific results, use \code{\link[=get_classes]{get_classes()}}, \code{\link[=get_atc]{get_atc()}}, \code{\link[=get_epc]{get_epc()}},
\code{\link[=get_va]{get_va()}}, or related functions directly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{rxref_conf}
\alias{rxref_conf}
\title{Configure rxref}
\usage{
rxref_conf(
  base_url = getOption("rxref.base_url"),
  rate_delay = getOption("rxref.rate_delay"),
  cache = getOption("rxref.cache")
)
}
\arguments{
\item{base_url}{Override the RxNav base URL (e.g., a local mirror)}

\item{rate_delay}{Seconds to wait between HTTP calls}

\item{cache}{A cachem cache object used by memoised calls}
}
\value{
A named list of current settings
}
\description{
Configure rxref
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tty.R
\name{tty_catalogue}
\alias{tty_catalogue}
\title{Catalogue of RxNorm TTY (Term Types)}
\usage{
tty_catalogue()
}
\value{
A tibble with metadata for key TTYs.
}
\description{
Returns a tibble describing common RxNorm TTYs you may want to use when
expanding ingredients to products. Includes whether each TTY typically maps
cleanly to NDCs and whether it's included in the package's default or
extended TTY sets.
}
\details{
Columns:
\itemize{
\item \code{tty}: RxNorm term type code.
\item \code{label}: Short, human-friendly name.
\item \code{description}: What the TTY represents in RxNorm.
\item \code{maps_to_ndc}: Logical; whether CUIs of this TTY usually map to NDCs via
\verb{/rxcui/\{rxcui\}/ndcs}.
\item \code{typical_role}: How it’s commonly used (product, component, group, name).
\item \code{include_default}: Logical; included in \code{.rxref_default_ttys}.
\item \code{include_extended}: Logical; included in \code{.rxref_extended_ttys}.
}
}
\examples{
tty_catalogue()

# TTYs that map to NDCs
subset(tty_catalogue(), maps_to_ndc)$tty

# See what your defaults and extended sets contain
subset(tty_catalogue(), include_default)$tty
subset(tty_catalogue(), include_extended)$tty

# Pick a custom set: products + components
with(tty_catalogue(),
     tty[tty \%in\% c("SCD","SBD","GPCK","BPCK","SCDC","SBDC")])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resolve.R
\name{resolve}
\alias{resolve}
\title{Resolve free text, RxCUI, or NDC to RxCUI and preferred name}
\usage{
resolve(
  x,
  type = c("auto", "name", "rxcui", "ndc"),
  max_entries = 1,
  show_progress = interactive()
)
}
\arguments{
\item{x}{Character vector: drug string, RxCUI, or NDC (10/11-digit or hyphenated)}

\item{type}{One of c("auto","name","rxcui","ndc"). Default "auto" infers.}

\item{max_entries}{Integer, passed to approximateTerm for name queries.}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble with columns: input, type, rxcui, name, tty, score (if name),
ndc11 (if ndc input), matched_term (if name input)
}
\description{
Vectorized over \code{x}. For free text, uses RxNorm approximateTerm.
For NDC, uses findRxcuiById. For RxCUI, validates and returns properties.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map.R
\name{map_ndc_to_rxcui}
\alias{map_ndc_to_rxcui}
\title{Map NDCs to RxCUIs}
\usage{
map_ndc_to_rxcui(x, show_progress = interactive())
}
\arguments{
\item{x}{Character vector of NDCs.}

\item{show_progress}{Logical; if \code{TRUE}, show a progress bar for vectorized
mapping operations. Defaults to \code{interactive()}.}
}
\value{
A tibble with columns:
\describe{
\item{input}{Original input NDC.}
\item{ndc11}{Normalized 11-digit NDC.}
\item{rxcui}{Mapped RxCUI, if found.}
}
}
\description{
Convenience wrapper around \code{\link[=map_to]{map_to()}} for mapping National Drug Codes
(NDCs) to RxNorm Concept Unique Identifiers (RxCUIs).
}
\details{
NDCs may be supplied as 10-digit, 11-digit, or hyphenated values. Input
NDCs are normalized to 11-digit format before querying RxNorm.
}
\examples{
\dontrun{
map_ndc_to_rxcui("00093-7424-56")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{package}
\name{rxref-package}
\alias{rxref}
\alias{rxref-package}
\title{rxref: Tidy RxNorm utilities}
\description{
\code{rxref} provides tidy, API-first tools for working with RxNorm and RxClass
data in R. The package helps users resolve drug names and identifiers,
inspect RxNorm concepts, map between RxCUIs and NDCs, expand ingredients to
product-level concepts, retrieve drug class information, and derive clinically
useful product attributes.
}
\details{
Common workflows include:
\itemize{
\item Resolving drug names or identifiers with \code{\link[=resolve]{resolve()}} and \code{\link[=search_drug]{search_drug()}}.
\item Retrieving concept metadata with \code{\link[=get_properties]{get_properties()}}.
\item Building product-level medication lists with \code{\link[=find_ingredients]{find_ingredients()}},
\code{\link[=ingredients_for_rxcui]{ingredients_for_rxcui()}}, and \code{\link[=products_for_ingredients]{products_for_ingredients()}}.
\item Mapping between RxCUIs and NDCs with \code{\link[=map_to]{map_to()}},
\code{\link[=map_ndc_to_rxcui]{map_ndc_to_rxcui()}}, and \code{\link[=map_rxcui_to_ndc]{map_rxcui_to_ndc()}}.
\item Retrieving therapeutic class information with \code{\link[=get_classes]{get_classes()}},
\code{\link[=find_classes]{find_classes()}}, \code{\link[=get_class_members]{get_class_members()}}, \code{\link[=get_atc]{get_atc()}}, \code{\link[=get_epc]{get_epc()}},
and \code{\link[=get_va]{get_va()}}.
\item Deriving clinical product attributes with \code{\link[=get_clinical_attributes]{get_clinical_attributes()}} and
\code{\link[=filter_products_by_route]{filter_products_by_route()}}.
}

Configure package behavior with \code{\link[=rxref_conf]{rxref_conf()}}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/ssmithm/rxref}
  \item \url{https://www.stevenmsmith.org/rxref/}
  \item Report bugs at \url{https://github.com/ssmithm/rxref/issues}
}

}
\author{
\strong{Maintainer}: Steven Smith \email{ssmith@cop.ufl.edu}

Authors:
\itemize{
  \item Steven Smith \email{ssmith@cop.ufl.edu}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{products_for_ingredients}
\alias{products_for_ingredients}
\title{Expand ingredient CUIs to product CUIs that truly contain the ingredient}
\usage{
products_for_ingredients(
  ingredient_rxcui,
  ttys = .rxref_default_ttys,
  route = NULL,
  include_combos = TRUE,
  concept_status = c("active", "active_and_historical"),
  historical_status = c("Obsolete", "Remapped", "Quantified", "NotCurrent"),
  show_progress = interactive()
)
}
\arguments{
\item{ingredient_rxcui}{Character vector of ingredient CUIs (TTY \code{IN} or \code{PIN}).}

\item{ttys}{Character vector of TTYs to include. Defaults to product-facing
TTYs returned by \code{\link[=product_ttys]{product_ttys()}}. Pass a larger set if you want groups,
components, branded concepts, or other product-related concepts, for example
\code{product_ttys("extended_product")}.}

\item{route}{Optional character vector of routes to retain. If \code{NULL}, no
route filtering is performed. Route filtering uses \code{\link[=get_clinical_attributes]{get_clinical_attributes()}}.
Route filtering is intended for product-level TTYs and may not filter well
on broader group or package TTYs.}

\item{include_combos}{Logical. If \code{FALSE}, keep only single-ingredient
products, where ingredient count is based on distinct \code{IN} concepts when
available and otherwise falls back to distinct \code{PIN} concepts.}

\item{concept_status}{Character. Which RxNorm concept universe to search.
\code{"active"} uses active-scope RxNav relationship endpoints and is the
default. \code{"active_and_historical"} also searches historical RxNorm concepts
using all-status concept retrieval and RxCUI history status metadata.
Historical searching is slower and is intended for mapping older
prescribing or dispensing data.}

\item{historical_status}{Character vector of historical RxNorm statuses to
include when \code{concept_status = "active_and_historical"}. Defaults to
\code{c("Obsolete", "Remapped", "Quantified", "NotCurrent")}. These values use
RxNorm status definitions:
\describe{
\item{\code{"Obsolete"}}{The concept is obsolete in the current RxNorm data set,
and RxNorm has not designated an active concept as equivalent.}
\item{\code{"Remapped"}}{The concept was active or obsolete at one time, is no
longer in the current data set, and has been remapped to one or more
active or obsolete concepts.}
\item{\code{"Quantified"}}{The concept has been designated as non-dispensable
because it lacks a quantity factor; related concepts with quantity
factors may be available.}
\item{\code{"NotCurrent"}}{The concept either exists in the current data set
without RxNorm vocabulary terms, or existed in a previous monthly
release but has since been removed and not remapped.}
}
See the RxNorm API documentation for concept status values:
\url{https://lhncbc.nlm.nih.gov/RxNav/APIs/api-RxNorm.getAllConceptsByStatus.html}.}

\item{show_progress}{Logical. Show progress bars for long-running API
retrieval, product matching, and optional route filtering steps. Defaults
to \code{interactive()}.}
}
\value{
A tibble with one row per matched ingredient/product concept pair.
For \code{concept_status = "active"}, columns include \code{ingredient_rxcui},
\code{product_rxcui}, \code{name}, \code{tty}, and \code{n_ingredients}. When
\code{concept_status = "active_and_historical"}, additional columns include
\code{concept_status}, \code{active_start_date}, \code{active_end_date},
\code{release_start_date}, and \code{release_end_date}.
}
\description{
Tries multiple RxNav endpoints and verifies candidates truly contain the
queried ingredient or one of its acceptable related ingredient concepts.
The function unions candidates from multiple sources, verifies ingredient
containment, and reports the number of ingredients represented by each
product concept.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map.R
\name{map_to}
\alias{map_to}
\title{Map between NDCs and RxCUIs}
\usage{
map_to(
  x,
  to = c("rxcui", "ndc"),
  status = NULL,
  history = NULL,
  show_progress = interactive()
)
}
\arguments{
\item{x}{Character vector of NDCs or RxCUIs.}

\item{to}{Direction of mapping. One of \code{"rxcui"} or \code{"ndc"}.
Use \code{"rxcui"} to map NDCs to RxCUIs and \code{"ndc"} to map RxCUIs to NDCs.}

\item{status}{For \code{to = "ndc"} only, optional character vector of NDC
statuses to retain, such as \code{"ACTIVE"}, \code{"OBSOLETE"}, \code{"ALIEN"}, or
\code{"UNKNOWN"}. If \code{NULL}, no status-based filtering is applied to the
retrieved NDCs. Note that when \code{history = "active"}, RxNorm's active NDC
endpoint is used, so returned NDCs are expected to be active even when
\code{status = NULL}.}

\item{history}{For \code{to = "ndc"} only, the NDC association history to
retrieve. One of \code{"active"}, \code{"direct"}, or \code{"all"}.
\itemize{
\item \code{"active"} retrieves currently active NDCs associated with the supplied
RxCUI.
\item \code{"direct"} retrieves NDCs ever directly associated with the supplied
RxCUI.
\item \code{"all"} retrieves NDCs ever directly or indirectly associated with the
supplied RxCUI, including associations through remapped or archived
concepts.
}

If \code{NULL} and \code{to = "ndc"}, defaults to \code{"active"}. Ignored when
\code{to = "rxcui"}.}

\item{show_progress}{Logical; if \code{TRUE}, show a progress bar for vectorized
mapping operations. Defaults to \code{interactive()}.}
}
\value{
A tibble.

For \code{to = "rxcui"}, returns one row per input NDC/RxCUI mapping with columns:
\describe{
\item{input}{Original input NDC.}
\item{ndc11}{Normalized 11-digit NDC.}
\item{rxcui}{Mapped RxCUI, if found.}
}

For \code{to = "ndc"} and \code{history = "active"}, returns one row per RxCUI/NDC
mapping with columns:
\describe{
\item{rxcui}{Input RxCUI.}
\item{ndc11}{Mapped NDC.}
\item{ndc_status}{NDC status returned by RxNorm, when available.}
}

For \code{to = "ndc"} and \code{history} equal to \code{"direct"} or \code{"all"}, the output
may also include:
\describe{
\item{related_rxcui}{RxCUI associated with the historical NDC record. This
may differ from the input RxCUI when historical or indirect associations
are retrieved.}
\item{ndc_start_date}{Start date of the NDC association, when available.}
\item{ndc_end_date}{End date of the NDC association, when available.}
}
}
\description{
\code{map_to()} maps identifiers between National Drug Codes (NDCs) and RxNorm
Concept Unique Identifiers (RxCUIs).
}
\details{
When mapping from NDC to RxCUI, NDCs may be supplied as 10-digit,
11-digit, or hyphenated values. Input NDCs are normalized to 11-digit
format before querying RxNorm.

When mapping from RxCUI to NDC, the \code{history} argument controls whether
only currently active NDCs are retrieved or whether historical NDC
associations are also included.

RxNorm distinguishes between currently active NDCs and historical NDC
associations. The default behavior, \code{history = "active"}, uses RxNorm's
active NDC endpoint and preserves the earlier behavior of \code{map_to()}.

To retrieve obsolete, discontinued, or otherwise historical NDCs, use
\code{history = "direct"} or \code{history = "all"}. Setting \code{status = NULL} does not
by itself request historical NDCs; it only means that no status filter is
applied after NDCs are retrieved.
}
\examples{
\dontrun{
# Map an NDC to RxCUI
map_to("00093-7424-56", to = "rxcui")

# Map an RxCUI to currently active NDCs
map_to("1049630", to = "ndc")

# Map an RxCUI to all directly associated historical NDCs
map_to("1049630", to = "ndc", history = "direct")

# Map an RxCUI to all historical NDCs and retain obsolete NDCs only
map_to("1049630", to = "ndc", history = "all", status = "OBSOLETE")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{search_drug}
\alias{search_drug}
\title{Search free-text drug name and return product CUIs and/or NDCs}
\usage{
search_drug(
  term,
  return = c("rxcui", "ndc", "both"),
  ndc_status = NULL,
  ttys = .rxref_default_ttys,
  route = NULL,
  show_progress = interactive(),
  ...
)
}
\arguments{
\item{term}{Character vector; free-text drug names.}

\item{return}{One of \code{c("rxcui","ndc","both")}. Note that \code{"both"} will return a list
with both an rxcui tibble and an ndc tibble.}

\item{ndc_status}{Optional character vector to filter NDCs. Options are "ACTIVE",
"OBSOLETE", "UNSPECIFIED" (the API may also return no value, which will appear as NA).}

\item{ttys}{Character vector of TTYs to include in product search.
Defaults to \code{.rxref_default_ttys}. Other prespecified option is \code{.rxref_extended_ttys}
or a character vector of explicit TTYs. Run \code{tty_catalogue()} to review options.}

\item{route}{Optional character vector of routes to retain before returning
products or mapping to NDCs. If \code{NULL}, no route filtering is performed.
Common values include \code{"ORAL"}, \code{"INJECTION"}, \code{"OPHTHALMIC"},
\code{"INHALATION"}, and \code{"TOPICAL"}.}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}

\item{...}{Passed to \code{products_for_ingredients()} (e.g., include_combos = FALSE)}
}
\value{
If \code{return="rxcui"}: tibble of products.
If \code{"ndc"}: tibble of NDCs with \code{ingredient_rxcui}, \code{product_rxcui}, \code{ndc11}, \code{ndc_status}.
If \code{"both"}: list(products=…, ndcs=…).
}
\description{
High-level convenience: free text -> ingredient(s) (IN/PIN) -> verified product
CUIs, and optionally expand to NDCs with status filtering.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{ingredients_for_rxcui}
\alias{ingredients_for_rxcui}
\title{Get ingredient concepts for RxCUIs}
\usage{
ingredients_for_rxcui(
  rxcui,
  include_pin = TRUE,
  include_min = FALSE,
  show_progress = interactive()
)
}
\arguments{
\item{rxcui}{Character vector of RxCUIs.}

\item{include_pin}{Logical. Include precise ingredient concepts (\code{PIN}).}

\item{include_min}{Logical. Include multiple ingredient concepts (\code{MIN}).}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.}
}
\value{
A tibble with columns \code{rxcui}, \code{ingredient_rxcui},
\code{ingredient_name}, and \code{ingredient_tty}.
}
\description{
Maps one or more RxCUIs to related ingredient concepts, returning ingredient
RxCUIs, names, and term types. This is useful when the input is already a
specific RxNorm product concept rather than a free-text drug name.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{rxclass_rela_sources}
\alias{rxclass_rela_sources}
\title{Get RxClass relationship sources}
\usage{
rxclass_rela_sources()
}
\value{
A tibble of RxClass relationship sources.
}
\description{
Get RxClass relationship sources
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/route.R
\name{filter_products_by_route}
\alias{filter_products_by_route}
\title{Filter RxNorm product concepts by route of administration.}
\usage{
filter_products_by_route(
  products,
  route = "ORAL",
  keep_route_info = TRUE,
  include_historical = FALSE,
  show_progress = interactive()
)
}
\arguments{
\item{products}{A tibble containing a \code{product_rxcui} column, such as the
output of \code{\link[=products_for_ingredients]{products_for_ingredients()}}.}

\item{route}{Character vector of routes to keep. Common values include
\code{"ORAL"}, \code{"INJECTION"}, \code{"OPHTHALMIC"}, \code{"OTIC"}, \code{"INHALATION"},
\code{"TOPICAL"}, \code{"RECTAL"}, \code{"VAGINAL"}, \code{"URETHRAL"}, and \code{"IMPLANT"}.}

\item{keep_route_info}{Logical. If \code{TRUE} (default), append route, dose form, and
dose-form group summaries to the returned table.}

\item{include_historical}{Logical. If \code{TRUE}, use RxCUI history status
metadata as a fallback for RxCUIs that do not return active clinical
attributes. This is useful for obsolete, remapped, quantified, or otherwise
non-current RxCUIs found in historical prescribing data. If \code{products}
contains a \code{concept_status} column with non-\code{"active"} value, historical
lookup is enabled automatically.}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble containing only rows whose \code{product_rxcui} has at least one
matching route.
}
\description{
Filters a product-level rxref table using route information from
\code{\link[=get_clinical_attributes]{get_clinical_attributes()}}. Useful when a drug or class includes
products across multiple routes, such as oral tablets, ophthalmic solutions,
injectables, patches, or topical formulations.
}
\examples{
\dontrun{
ing <- find_ingredients("metoprolol")

prods <- products_for_ingredients(
  ing$rxcui,
  include_combos = TRUE
)

filter_products_by_route(prods, route = "ORAL")

# Historical/current products can also be route-filtered
prods_hist <- products_for_ingredients(
  ing$rxcui,
  include_combos = TRUE,
  concept_status = "active_and_historical"
)

filter_products_by_route(prods_hist, route = "ORAL")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clinical.R
\name{get_clinical_attributes}
\alias{get_clinical_attributes}
\title{Clinical attributes from the concept (SCD/SBD) or related SCD/SBD}
\usage{
get_clinical_attributes(
  rxcui,
  include_historical = FALSE,
  show_progress = interactive()
)
}
\arguments{
\item{rxcui}{Character vector of RxCUIs}

\item{include_historical}{Logical. If \code{TRUE}, use RxCUI history status
metadata as a fallback for RxCUIs that do not return active clinical
attributes. This is useful for obsolete, remapped, quantified, or otherwise
non-current RxCUIs found in historical prescribing data.}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble with columns:
\describe{
\item{rxcui}{Input RxCUI}
\item{related_rxcui}{Clinical drug RxCUI (SCD/SBD) used for attributes}
\item{name}{Clinical drug name}
\item{tty}{Term type (SCD/SBD, etc.)}
\item{strength}{Parsed strength string (e.g. "500 MG")}
\item{dose_form}{Parsed dose form (e.g. "Extended Release Oral Tablet")}
\item{route}{Route parsed from dose form / DFG (e.g. "ORAL", "INJECTION")}
\item{dose_form_group}{Dose form group (DFG), if available}
\item{is_brand}{Logical; TRUE for branded clinical concepts (SBD/BPCK)}
\item{is_generic}{Logical; TRUE for generic clinical concepts (SCD/GPCK)}
\item{ingredient_count}{Number of distinct ingredients}
\item{ingredient_rxcui}{Ingredient RxCUI. For combination products,
multiple values are returned as semicolon-delimited strings.}
\item{ingredient_name}{Ingredient name. For combination products,
multiple values are returned as semicolon-delimited strings.}
\item{ingredient_tty}{Ingredient term type (IN/PIN/MIN). For combination
products, multiple values are returned as semicolon-delimited strings.}
\item{is_multi_ingredient}{Logical; TRUE if >1 ingredient}
\item{suppress}{Raw RxNorm suppress flag from properties}
\item{status}{Simple status derived from suppress: "ACTIVE" vs "INACTIVE"}
}
}
\description{
If \code{rxcui} is already a clinical drug (TTY = SCD or SBD), parse attributes directly
from its own name. Otherwise, query \code{/rxcui/{id}/related?tty=SCD,SBD} and parse.
}
\details{
In addition to strength and dose form, this returns route, dose-form group (DFG),
brand/generic flags, ingredient summaries, and a simple active/inactive status.

Note there is a fair amount of parsing of the RxNorm STR value to try to extract
relevant information (e.g., strength, dose_form), so check closely before trusting.
There may be edge cases that are not correctly parsed.

For combination products, ingredient-related columns may contain multiple
semicolon-delimited values, such as \code{"amlodipine; valsartan"}.
}
\examples{
\dontrun{
get_clinical_attributes(c("861007","860975")) |>
  dplyr::select(rxcui, related_rxcui, name, strength, dose_form, route, tty) |>
  head()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{get_epc}
\alias{get_epc}
\title{Get FDA/SPL established pharmacologic classes}
\usage{
get_epc(
  x,
  by = c("rxcui", "name"),
  keep_input = TRUE,
  show_progress = interactive()
)
}
\arguments{
\item{x}{Character vector of RxCUIs or drug names.}

\item{by}{One of \code{"rxcui"} or \code{"name"}.}

\item{keep_input}{Logical; if \code{TRUE}, includes the original input value.}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble of established pharmacologic class assertions.
}
\description{
\code{get_epc()} is a convenience wrapper for \code{get_classes()} that returns
FDA labeling class assertions for RxNorm drugs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tty.R
\name{extended_ttys}
\alias{extended_ttys}
\title{Extended RxNorm term types}
\usage{
extended_ttys()
}
\value{
A character vector of RxNorm term type abbreviations.
}
\description{
Returns an extended set of RxNorm term types that includes the default and extended
product term types plus drug component, brand name, and ingredient concepts.
This list includes essentially all term types that capture a specific ingredient,
thus it excludes dose form, dose form group, prescribable name, synonyms, and others
that are not associated with a specific ingredient.
}
\details{
This can be useful when the broadest set of RxNorm concepts that still capture an ingredient
is desired.
}
\examples{
extended_ttys()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{find_ingredients}
\alias{find_ingredients}
\title{Resolve a free-text drug name to ingredient CUIs (IN/PIN)}
\usage{
find_ingredients(
  term,
  max_entries = 10,
  include_pin = TRUE,
  show_progress = interactive()
)
}
\arguments{
\item{term}{Character vector (free text).}

\item{max_entries}{Integer. Max approximate-term candidates per input (default 10).}

\item{include_pin}{Logical. Include precise-ingredient (\code{PIN}) CUIs (default TRUE).}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble with columns: \code{input}, \code{rxcui}, \code{name}, \code{tty}, \code{score}.
}
\description{
Uses RxNav approximateTerm, then fetches properties for each candidate and
filters to ingredient-type concepts (TTY \code{IN} and, optionally, \code{PIN}).
}
\examples{
if (identical(Sys.getenv("RXREF_ONLINE"), "1")) {
  find_ingredients("metformin")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{rxclass_relas}
\alias{rxclass_relas}
\title{Get RxClass relationship types}
\usage{
rxclass_relas(rela_source)
}
\arguments{
\item{rela_source}{RxClass relationship source, such as \code{"ATC"},
\code{"DAILYMED"}, \code{"FDASPL"}, \code{"MEDRT"}, \code{"SNOMEDCT"}, or \code{"VA"}.}
}
\value{
A tibble of relationship names for the selected source.
}
\description{
\code{rxclass_relas()} is a helper function that queries the RxClass API for
generally valid relationships within the data. Note that just because a
relationship is generally valid within a rela_source, not every drug or
RxCUI will have a corresponding relationship asserted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tty.R
\name{default_product_ttys}
\alias{default_product_ttys}
\title{Default RxNorm product term types}
\usage{
default_product_ttys()
}
\value{
A character vector of RxNorm term type abbreviations.
}
\description{
Returns the default RxNorm term types used by rxref when identifying
drug products.
}
\details{
The default set is intentionally focused on product-level concepts:
semantic clinical drugs, semantic branded drugs, generic packs, and
branded packs.
}
\examples{
default_product_ttys()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{rxclass_relationships}
\alias{rxclass_relationships}
\title{Summarize RxClass relationship types for drugs}
\usage{
rxclass_relationships(
  x,
  by = c("rxcui", "name"),
  rela_source = NULL,
  show_progress = interactive()
)
}
\arguments{
\item{x}{Character vector of RxCUIs or drug names.}

\item{by}{One of \code{"rxcui"} or \code{"name"}.}

\item{rela_source}{Optional RxClass relationship source filter.}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble summarizing available relationship/class-type combinations.
}
\description{
\code{rxclass_relationships()} is a diagnostic helper that evaluates the type(s)
of relationship (rela) asserted in RxNorm MED-RT data and the number within
each relationship asserted. In some cases, \code{get_medrt()} or its various
convenience wrappers may return empty or shorter-than-expected tibbles
because some relationships expected by the user are not asserted in RxNorm
(for example, an MOA may not be asserted by MED-RT).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tty.R
\name{product_ttys}
\alias{product_ttys}
\title{RxNorm product term type sets}
\usage{
product_ttys(set = c("default", "extended_product", "extended"))
}
\arguments{
\item{set}{One of \code{"default"}, \code{"extended_product"}, or \code{"extended"}.}
}
\value{
A character vector of RxNorm term type abbreviations.
}
\description{
Returns common RxNorm term type sets used by rxref.
}
\examples{
product_ttys()
product_ttys("extended")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status.R
\name{ndc_status}
\alias{ndc_status}
\title{Get RxNav status for NDCs}
\usage{
ndc_status(ndc, show_progress = interactive())
}
\arguments{
\item{ndc}{character vector; hyphenated or digits}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble with three columns: ndc_input, ndc11, ndc_hyph, status
}
\description{
\code{ndc_status()} requests the status of an NDC entry for one or more user-supplied
NDC numbers. The returned status will be one "ACTIVE", "OBSOLETE", or "UNSPECIFIED".
On occasion, a NULL value may be pulled (resulting in an \code{NA} in the returned tibble),
when RxNorm has no status for a given NDC, or if an invalid NDC is supplied.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{get_medrt}
\alias{get_medrt}
\title{Get MED-RT assertions for RxNorm drugs}
\usage{
get_medrt(
  x,
  by = c("rxcui", "name"),
  relas = NULL,
  class_types = NULL,
  keep_input = TRUE,
  show_progress = interactive()
)
}
\arguments{
\item{x}{Character vector of RxCUIs or drug names.}

\item{by}{One of \code{"rxcui"} or \code{"name"}.}

\item{relas}{Optional MED-RT relationship filter.}

\item{class_types}{Optional MED-RT class type filter.}

\item{keep_input}{Logical; if \code{TRUE}, includes the original input value.}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble of MED-RT assertions.
}
\description{
\code{get_medrt()} is a convenience wrapper for \code{get_classes()} that returns
MED-RT assertions for RxNorm drugs. These include, for example, asserted
mechanisms of action, contraindications, physiologic effects (including
adverse side effects), etc.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{get_class_members}
\alias{get_class_members}
\title{Get RxNorm members of an RxClass class}
\usage{
get_class_members(
  class_id,
  rela_source,
  rela = NULL,
  include_indirect = TRUE,
  ttys = NULL
)
}
\arguments{
\item{class_id}{RxClass class identifier.}

\item{rela_source}{RxClass relationship source, such as \code{"ATC"},
\code{"ATCPROD"}, \code{"DAILYMED"}, \code{"FDASPL"}, \code{"MEDRT"}, \code{"SNOMEDCT"}, or \code{"VA"}.}

\item{rela}{Optional relationship filter.}

\item{include_indirect}{Logical; if \code{TRUE}, include direct and indirect
members. If \code{FALSE}, return direct members only.}

\item{ttys}{Optional RxNorm term type filter, such as \code{"IN"}, \code{"PIN"},
\code{"SCD"}, or \code{"SBD"}.}
}
\value{
A tibble of RxNorm drug members.
}
\description{
\code{get_class_members()} queries the RxClass API to identify members of a
specific drug class. Typically, this would be used to identify all
ingredient RxCUIs in a class, which can then be combined with other search
functions to extract related product RxCUIs or NDCs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/properties.R
\name{get_properties}
\alias{get_properties}
\title{Get core RxNorm properties for one or more RxCUIs}
\usage{
get_properties(rxcui, show_progress = interactive())
}
\arguments{
\item{rxcui}{Character vector of RxCUIs}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.}
}
\value{
tibble with rxcui, name, synonym, tty, language, suppress, umlscui.
}
\description{
Get core RxNorm properties for one or more RxCUIs
}
\examples{
if (identical(Sys.getenv("RXREF_ONLINE"), "1")) {
get_properties(c("860975","1049630"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map.R
\name{map_rxcui_to_ndc}
\alias{map_rxcui_to_ndc}
\title{Map RxCUIs to NDCs}
\usage{
map_rxcui_to_ndc(
  x,
  status = NULL,
  history = c("active", "direct", "all"),
  show_progress = interactive()
)
}
\arguments{
\item{x}{Character vector of RxCUIs.}

\item{status}{Optional character vector of NDC statuses to retain, such as
\code{"ACTIVE"}, \code{"OBSOLETE"}, \code{"ALIEN"}, or \code{"UNKNOWN"}. If \code{NULL}, no
status-based filtering is applied to the retrieved NDCs. Note that
\code{history = "active"} uses RxNorm's active NDC endpoint, so returned NDCs
are expected to be active even when \code{status = NULL}.}

\item{history}{For \code{to = "ndc"} only, the NDC association history to
retrieve. One of \code{"active"}, \code{"direct"}, or \code{"all"}.
\itemize{
\item \code{"active"} retrieves currently active NDCs associated with the supplied
RxCUI.
\item \code{"direct"} retrieves NDCs ever directly associated with the supplied
RxCUI.
\item \code{"all"} retrieves NDCs ever directly or indirectly associated with the
supplied RxCUI, including associations through remapped or archived
concepts.
}

If \code{NULL} and \code{to = "ndc"}, defaults to \code{"active"}. Ignored when
\code{to = "rxcui"}.}

\item{show_progress}{Logical; if \code{TRUE}, show a progress bar for vectorized
mapping operations. Defaults to \code{interactive()}.}
}
\value{
A tibble with one row per RxCUI/NDC mapping.

For \code{history = "active"}, returns columns:
\describe{
\item{rxcui}{Input RxCUI.}
\item{ndc11}{Mapped NDC.}
\item{ndc_status}{NDC status returned by RxNorm, when available.}
}

For \code{history = "direct"} or \code{"all"}, the output may also include:
\describe{
\item{related_rxcui}{RxCUI associated with the historical NDC record. This
may differ from the input RxCUI when historical or indirect associations
are retrieved.}
\item{ndc_start_date}{Start date of the NDC association, when available.}
\item{ndc_end_date}{End date of the NDC association, when available.}
}
}
\description{
Convenience wrapper around \code{\link[=map_to]{map_to()}} for mapping RxNorm Concept Unique
Identifiers (RxCUIs) to National Drug Codes (NDCs).
}
\details{
By default, this function retrieves currently active NDCs associated with
the supplied RxCUIs. Use \code{history = "direct"} or \code{history = "all"} to
retrieve historical NDC associations.

\code{status = NULL} means that no status filter is applied after NDCs are
retrieved. It does not, by itself, request historical NDCs. To retrieve
historical NDCs, set \code{history = "direct"} or \code{history = "all"}.
}
\examples{
\dontrun{
# Current active NDCs
map_rxcui_to_ndc("1049630")

# NDCs ever directly associated with this RxCUI
map_rxcui_to_ndc("1049630", history = "direct")

# All historical NDCs, retaining obsolete NDCs only
map_rxcui_to_ndc(
  "1049630",
  history = "all",
  status = "OBSOLETE"
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tty.R
\name{extended_product_ttys}
\alias{extended_product_ttys}
\title{Extended RxNorm product term types}
\usage{
extended_product_ttys()
}
\value{
A character vector of RxNorm term type abbreviations.
}
\description{
Returns an extended set of RxNorm term types that includes the default
product term types plus dose-form and dose-form-group concepts.
}
\details{
This can be useful when a broader set of product-related RxNorm concepts
is desired.
}
\examples{
extended_product_ttys()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{get_classes}
\alias{get_classes}
\title{Get RxClass assertions for RxNorm drugs}
\usage{
get_classes(
  x,
  by = c("rxcui", "name"),
  rela_source = NULL,
  relas = NULL,
  class_types = NULL,
  keep_input = TRUE,
  show_progress = interactive()
)
}
\arguments{
\item{x}{Character vector of RxCUIs or drug names.}

\item{by}{One of \code{"rxcui"} or \code{"name"}.}

\item{rela_source}{Optional RxClass relationship source filter, such as
\code{"ATC"}, \code{"ATCPROD"}, \code{"DAILYMED"}, \code{"FDASPL"}, \code{"MEDRT"}, \code{"SNOMEDCT"},
or \code{"VA"}. May be a character vector; values are queried separately.}

\item{relas}{Optional relationship filter. Examples include \code{"has_EPC"},
\code{"has_MoA"}, \code{"has_PE"}, \code{"may_treat"}, \code{"ci_with"},
\code{"has_chemical_structure"}, and others supported by RxClass.}

\item{class_types}{Optional filter on returned class types, such as \code{"EPC"},
\code{"MOA"}, \code{"PE"}, \code{"DISEASE"}, \code{"CHEM"}, \code{"VA"}, or \code{"ATC1-4"}.}

\item{keep_input}{Logical; if \code{TRUE}, includes the original input value.}

\item{show_progress}{Logical. Show a progress bar in interactive sessions.
Progress is shown only when at least 5 inputs are supplied.}
}
\value{
A tibble with one row per RxClass assertion.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

\code{get_classes()} returns RxClass assertions for RxNorm concepts or drug names.
These include drug-class relationships from sources such as ATC, FDASPL,
DAILYMED, MEDRT, SNOMEDCT, and VA. Depending on the source and relationship,
returned assertions may represent pharmacologic classes, chemical structures,
mechanisms of action, physiologic effects, indications, contraindications,
VA classes, ATC classes, or SNOMED CT dispositions/structures.

Because different sources use different classification logic, this function
preserves class type, relationship, and relationship source rather than
collapsing results into a single class label.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classes.R
\name{find_classes}
\alias{find_classes}
\title{Find RxClass drug classes by class name}
\usage{
find_classes(query, class_types = NULL)
}
\arguments{
\item{query}{Character string to search for.}

\item{class_types}{Optional class type filter, such as \code{"EPC"}, \code{"ATC1"},
\code{"ATC2"}, \code{"ATC3"}, \code{"ATC4"}, \code{"CHEM"}, \code{"DISEASE"}, or \code{"VA"}.}
}
\value{
A tibble of matching class concepts.
}
\description{
Find RxClass drug classes by class name
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-products.R
\name{exclude_products_with_ingredients}
\alias{exclude_products_with_ingredients}
\title{Exclude products containing specified ingredients}
\usage{
exclude_products_with_ingredients(
  data,
  ingredients = NULL,
  ingredient_rxcuis = NULL,
  product_id_col = "product_rxcui",
  ingredient_name_col = "ingredient_name",
  ingredient_rxcui_col = "ingredient_rxcui",
  return_excluded = FALSE
)
}
\arguments{
\item{data}{A data frame containing product RxCUIs and ingredient information,
such as that produced by \code{\link[=products_for_ingredients]{products_for_ingredients()}}.}

\item{ingredients}{Optional character vector of ingredient names to exclude.
These are resolved to ingredient RxCUIs using \code{\link[=find_ingredients]{find_ingredients()}}.}

\item{ingredient_rxcuis}{Optional character vector of ingredient RxCUIs to
exclude. These are used directly and do not require name resolution.}

\item{product_id_col}{Name of the column containing product RxCUIs. Default
is \code{product_rxcui}.}

\item{ingredient_name_col}{Name of the column containing ingredient names.
Used for reporting and, when needed, fallback matching. Default is
\code{ingredient_name}.}

\item{ingredient_rxcui_col}{Name of the column containing ingredient RxCUIs.
Default is \code{ingredient_rxcui}.}

\item{return_excluded}{Logical. If \code{FALSE}, return the filtered data frame.
If \code{TRUE}, return a list with \code{data}, \code{excluded}, and
\code{resolved_ingredients}.}
}
\value{
If \code{return_excluded = FALSE}, a data frame of the same general shape as
\code{data}, with excluded product rows removed.

If \code{return_excluded = TRUE}, a list with:
\describe{
\item{data}{The filtered data frame.}
\item{excluded}{Rows from the input data belonging to excluded products.}
\item{resolved_ingredients}{The ingredient names/RxCUIs used for exclusion.}
}
}
\description{
Remove all rows for products that contain one or more user-specified
ingredients. This is useful when broad ingredient-based product searches
return fixed-dose combination products that contain an ingredient of interest
but are not clinically relevant to a specific study definition.
}
\details{
Exclusions are applied at the product level. If any ingredient in a product
matches the exclusion list, all rows for that product are removed.
}
\examples{
\dontrun{
antihtn_clean <- antihtn_products |>
  exclude_products_with_ingredients(ingredients = "sacubitril")

antihtn_audit <- antihtn_products |>
  exclude_products_with_ingredients(
    ingredients = "sacubitril",
    return_excluded = TRUE
  )

antihtn_audit$data
antihtn_audit$excluded
antihtn_audit$resolved_ingredients
}

}
